/*
  cxproc - Configurable Xml PROCessor

  Copyright (C) 2006,2007,2008 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <assert.h>
#include <errno.h>
#include <limits.h>

#include <libxml/tree.h>

/* 
 */
#include "basics.h"
#include "tools.h"
#include "dom_x.h"
#include "plain_x.h"
#include "make_x.h"
#include "calendar_x.h"

#include "cxp_dtd.h"

/* Global constants
 */
const xmlChar *dow[] = {
  BAD_CAST "sun",
  BAD_CAST "mon",
  BAD_CAST "tue",
  BAD_CAST "wed",
  BAD_CAST "thu",
  BAD_CAST "fri",
  BAD_CAST "sat",
  NULL
};

const xmlChar *dow_de[] = {
  BAD_CAST "So",
  BAD_CAST "Mo",
  BAD_CAST "Di",
  BAD_CAST "Mi",
  BAD_CAST "Do",
  BAD_CAST "Fr",
  BAD_CAST "Sa",
  NULL
};

const xmlChar *moy[] = {
  BAD_CAST "Januar",
  BAD_CAST "Februar",
  BAD_CAST "Maerz",
  BAD_CAST "April",
  BAD_CAST "Mai",
  BAD_CAST "Juni",
  BAD_CAST "Juli",
  BAD_CAST "August",
  BAD_CAST "September",
  BAD_CAST "Oktober",
  BAD_CAST "November",
  BAD_CAST "Dezember",
  NULL
};


static BOOL_T
substitute_format(xmlNodePtr node);

static long int
get_day_abs_str(xmlChar *str_gcal);

static int
get_dow_int(xmlChar *str_dow);

static int *
get_year_arr(const xmlNodePtr node_calendar);


/* Global variables
 */
long int offset_days = -1;		/* difference between get_day_abs() */
struct tm today;

/*! Copy all date related nodes into the calendar DOM.

\param node_parent in calendar DOM
\param col_id \@id of col element
\param a_node node and his siblings to insert
*/
BOOL_T
copy_todo_to_calendar(xmlNodePtr node_parent, xmlChar *col_id, xmlNodePtr a_node)
{
  xmlNodePtr node_current;
  xmlNodePtr node_attr_id;
  xmlChar *ptr_date;
  xmlChar *ptr_header;

  for (node_current = a_node; node_current; node_current = node_current->next) {
    if (node_current->type == XML_ELEMENT_NODE
	&& (!xmlStrcasecmp(node_current->name, BAD_CAST "task")
	    || !xmlStrcasecmp(node_current->name, BAD_CAST "target")
	    || !xmlStrcasecmp(node_current->name, BAD_CAST "link"))
	&& ((ptr_date = attribute_str(node_current,BAD_CAST "done"))
	    || (ptr_date = attribute_str(node_current,BAD_CAST "date")))
	) {
      xmlNodePtr n;
      if (!xmlStrcasecmp(node_current->name, BAD_CAST "link")) {
	  if (xmlHasProp(node_current, BAD_CAST "href")) {
	  n = xmlNewNode(NULL,BAD_CAST "p");
          xmlAddChild(n,xmlCopyNodeList(node_current));
	  }
	  else {
	/* rename element if link */
	  n = xmlCopyNode(node_current,1);
	xmlNodeSetName(n, BAD_CAST "p");
	  }
      }
      else {
	  n = xmlCopyNode(node_current,1);
      }
      if ((node_attr_id = attribute_node(n,BAD_CAST"id"))) {
        /* rename id attribute to idref in copy of n */
	xmlNodeSetName(node_attr_id, BAD_CAST "idref");
      }
      
      /*! add ancestor axis to n */
      if ((ptr_header = get_parent_str(node_current))) {
	xmlSetProp(n, BAD_CAST "hstr", ptr_header);
	xmlFree(ptr_header);
      }
      insert_node_cal(node_parent,
		      col_id,
		      ptr_date,
		      n);
      xmlFreeNodeList(n);
    }
    else if (IS_NODE(node_current,BAD_CAST"section")
	&& ((ptr_date = attribute_str(node_current,BAD_CAST "done"))
	    || (ptr_date = attribute_str(node_current,BAD_CAST "date")))
	) {
      xmlChar *ptr_d = NULL;
      /*!\todo SECTION/\@DATE: check if any child::TAKS/\@DATE is later than parent DEADLINE -> XSL??*/
      xmlNodePtr n = xmlNewNode(NULL,BAD_CAST "deadline");
      /*! add ancestor axis to n */
      if ((ptr_header = get_parent_str(node_current))) {
	xmlSetProp(n, BAD_CAST "hstr", ptr_header);
	xmlFree(ptr_header);
      }
      if ((ptr_d = attribute_str(node_current,BAD_CAST "done"))) {
	xmlSetProp(n, BAD_CAST "done", ptr_d);
      }
      if ((ptr_d = attribute_str(node_current,BAD_CAST "date"))) {
	xmlSetProp(n, BAD_CAST "date", ptr_d);
      }

      xmlAddChild(n,xmlCopyNode(child_element(node_current,BAD_CAST "h"),1));
      insert_node_cal(node_parent,
		      col_id,
		      ptr_date,
		      n);
      xmlFreeNodeList(n);
      copy_todo_to_calendar(node_parent, col_id, node_current->children);
    }
    else if (node_current->children) {
      copy_todo_to_calendar(node_parent, col_id, node_current->children);
    }
  }
  return TRUE;
}
/* End of copy_todo_to_calendar() */


/*! Decodes the given string and returns the according result of get_day_abs()

\param str_gcal date string
*/
long int
get_day_abs_str(xmlChar *str_gcal)
{
  int year, mon, mday, week, wday;

  if (str_gcal && sscanf_datum(str_gcal,&year, &mon, &mday, &week, &wday))
    return get_day_abs(year, mon, mday, week, wday);

  return -1;
}
/* end of get_day_abs_str() */



/*! returns the index number 
*/
int
get_dow_int(xmlChar *str_dow)
{
  int i;

  for (i=0; dow[i]; i++) {
    if (xmlStrcasecmp(str_dow,dow[i]) == 0
	/*\todo || xmlStrcasecmp(str_dow,dow_de[i]) == 0 */
	) {
      return i;
    }
  }

  return -1;
}
/* end of get_dow_int() */


/*! computes the number of days since 1970-01-01 as an absolute value
*/
long int
get_day_abs(int year, int mon, int mday, int week, int wday)
{
  long int result = -1;

  if (year>1970) {
    struct tm t;
    t.tm_year = year - 1900; 

    if (mon > 0 && mday > 0 && mday < 32) {
      t.tm_yday = 0;
      t.tm_mon  = mon - 1;
      t.tm_mday = mday;
      t.tm_wday = 0;
      t.tm_hour = 0;
      t.tm_min  = 0;
      t.tm_sec  = 0;
      t.tm_isdst = 0;
      result = mktime(&t) / (60 * 60 * 24);
    }
    else if (week > -1 && week < 54 && wday > -1 && wday < 7) {
      int week_of_year;

      /* starting with the very first day of year */
      t.tm_yday = 0;
      t.tm_mon  = 0;
      t.tm_mday = 1;
      t.tm_wday = 0;
      t.tm_hour = 0;
      t.tm_min  = 0;
      t.tm_sec  = 0;
      t.tm_isdst = 0;
      result = mktime(&t) / (60 * 60 * 24);
      week_of_year = (t.tm_wday > 3 || t.tm_wday < 1) ? 0 : 1;
      /*!\todo not ready */
      result += (week - week_of_year) * 7 + (wday - t.tm_wday);
    }
  }
  return result;
}
/* end of get_day_abs() */


/*! Scans 'str_gcal' for a datum description and write the results to
  the given addresses.

\bug "Rosenmontag" holidays on Win32/apache 

\param str_gcal date string
\param ptr_r_y pointer to result year
\param ptr_r_m pointer to result month
\param ptr_r_d pointer to result day of month
\param ptr_r_w pointer to result week of year
\param ptr_r_wday pointer to result day of week
 */
xmlChar *
sscanf_datum(xmlChar *str_gcal, int *ptr_r_y, int *ptr_r_m, int *ptr_r_d, int *ptr_r_w, int *ptr_r_wday)
{
  int y, m, w, d, d_w;
  int l_c = xmlStrlen(str_gcal);
  char b[BUFFER_LENGTH];
  xmlChar *ptr_dow = NULL;
  xmlChar *ptr_sep = NULL;

  y = m = w = d = d_w = -1;

  if (l_c < 4) {
    printf_log(2,"Ignore '%s', too short", str_gcal);
    return NULL;
  }

  /*!\todo: ISO-Date "2006-10-22" or german "22.10.2006" */

  /*
    eval year string
  */
  strncpy(b,(const char *)&str_gcal[0],4);
  b[4]='\0';
  y = atoi(b);
  if (y==0 || (y>1969 && y<2032)) {
  }
  else {
    printf_log(2,"Ignore '%s': invalid year '%s'", str_gcal, b);
    return NULL;
  }

  if (l_c > 5) {
    if (str_gcal[4] == 'W' || str_gcal[4] == 'w'
	|| (str_gcal[4] == '*'
	    && (str_gcal[5] == 'W' || str_gcal[5] == 'w'))) {
      /*
	eval year/week string
      */
      xmlChar *ptr_e;

      ptr_e = (str_gcal[4] == '*') ? &str_gcal[6] : &str_gcal[5];

      w = (int) strtol((char *)ptr_e,(char **)&ptr_e,10);
      if (w>-1 && w<54) {
      }
      else {
	printf_log(2,"Ignore '%s': invalid week '%s'", str_gcal, ptr_e);
	return NULL;
      }

      for (ptr_sep=ptr_e; isalpha(*ptr_sep); ptr_sep++) {}
      if (ptr_sep != ptr_e) {
	/* entry for a specified day of week found
	 */
	ptr_dow = xmlStrndup(BAD_CAST ptr_e, (ptr_sep - BAD_CAST ptr_e));
	d_w = get_dow_int(ptr_dow);
	xmlFree(ptr_dow);
      }
    }	/* end of week processing */
    else if (isdigit(str_gcal[4])) {
      /* 
	 eval year/month/day string
      */
      strncpy(b,(const char *)&str_gcal[4],2);
      b[2]='\0';
      m = atoi(b);

      if (m==0 || (m>0 && m<13)) {
      }
      else {
	printf_log(2,"Ignore '%s': invalid month '%s'", str_gcal, b);
	return NULL;
      }
      
      if (l_c > 7 && isdigit(str_gcal[6])) {
	strncpy(b,(const char *)&str_gcal[6],2);
	b[2]='\0';
	d = atoi(b);
	if (d==0 || (d>0 && d<367)) {
	  ptr_sep = &str_gcal[8];
	}
	else {
	  printf_log(2,"Ignore '%s': invalid day '%s'", str_gcal, b);
	  return NULL;
	}
      }
      else {
	ptr_sep = &str_gcal[l_c]; /* set the result to end of str_gcal */
      }
    }
    else if (str_gcal[4]=='@') {
#if 0
      /* 
	 eval datum variable
      */
      if (str_gcal[5] == 'e') {
	flag_easter = TRUE;
	/*!\todo handle year '0000' */
	get_easter(y,&m,&d);
	ptr_sep = &str_gcal[6];
      }
      else {
	printf_log(2,"Ignore other '%s': variables yet '%s'", str_gcal, &str_gcal[4]);
	return NULL; 
      }
#else
	/*!\todo handle variables */
	ptr_sep = &str_gcal[4];
#endif
    }
  }
  else if (l_c > 3) {
    /* year only */
    printf_log(2,"Whole year '%s'", str_gcal);
    ptr_sep = &str_gcal[4];
  }
  else {
    printf_log(2,"Ignore '%s'", str_gcal);
    return NULL;
  }
  /* Values for {y, (w,ptr_dow) | (m,d)} are set and checked */

  /* set results */
  *ptr_r_y = y;
  *ptr_r_m = m;
  *ptr_r_d = d;
  *ptr_r_w = w;
  *ptr_r_wday = d_w;

  return ptr_sep;
}
/* end of sscanf_datum() */


/*! Returns true if c is a valid char for a date string. The separator chars '#' and ',' are handled separately
*/
BOOL_T
iscal(xmlChar c)
{

  if (isalnum((int)c) || (c=='*') || (c==':') || (c=='.') || (c=='+') || (c=='-') || (c=='@')) {
    return TRUE;
  }
  return FALSE;
}
/* end of iscal() */


/*! Assembles the next date part.

\param str_gcal pointer to comma separated list of date strings
*/
xmlChar *
assemble_next_date(xmlChar *str_gcal)
{
  xmlChar *ptr_e;
  xmlChar *ptr_b;		/* begin of next string */
  xmlChar *str_tmp;
  int l_new;
  int l_old;
  BOOL_T flag_week = FALSE;

  for (ptr_b=str_gcal, l_old=0; iscal(*ptr_b); ptr_b++, l_old++) {
    /*
      skip all chars in first calendar string
    */
    if (*ptr_b=='*')
      flag_week = TRUE;
  }

  if (*ptr_b == ',' || *ptr_b == '#') {

    for (ptr_b++, ptr_e=ptr_b; iscal(*ptr_e); ptr_e++) {
      /* skip all chars in second calendar string */
    }

    l_new = ptr_e - ptr_b;

    if (l_new > 0) {
      if (flag_week) {
	int c_b = tolower((int)*ptr_b);	/* speedup */
	if (*ptr_b == 'w' && isdigit(*(ptr_b+1))) {
	  /*
	    next date descriptor is a week number beginning with '*w'
	   */
	  str_tmp = xmlStrndup(str_gcal, 5);
	}
	else if (c_b == tolower((int)dow[0][0])
		 || c_b == tolower((int)dow[1][0])
		 || c_b == tolower((int)dow[2][0])
		 || c_b == tolower((int)dow[3][0])
		 || c_b == tolower((int)dow[4][0])
		 || c_b == tolower((int)dow[5][0])
		 || c_b == tolower((int)dow[6][0]) ) {
	  /*
	    next date descriptor is a day of week string
	   */
	  str_tmp = xmlStrndup(str_gcal, l_old - l_new);
	}
	else if (isdigit(c_b)) {
	  /*
	    next date descriptor is a week number
	  */
	  char *ptr_e;
	  int w = (int) strtol((char *)ptr_b,(char **)&ptr_e,10);
	  if (w>-1 && w<54 && isalpha((int)*ptr_e)) {
	    str_tmp = xmlStrndup(str_gcal, l_old - l_new);
	  }
	  else {
	    printf_log(2,"No valid pattern '%s' in '%s'", ptr_b, str_gcal);
	    return NULL;
	  }
	}
	else {
	  printf_log(2,"No valid pattern '%s' in '%s'", ptr_b, str_gcal);
	  return NULL;
	}
      }
      else if (isdigit(*ptr_b)) {
	  if (l_new < 3) {
	    str_tmp = xmlStrndup(str_gcal, 6);
	    if (l_new < 2)  /* insert leading zero */
	      str_tmp = xmlStrcat(str_tmp, BAD_CAST "0");
	  }
	  else if (l_new < 5) {
	    str_tmp = xmlStrndup(str_gcal, 4);
	    if (l_new < 4)  /* insert leading zero */
	      str_tmp = xmlStrcat(str_tmp, BAD_CAST "0");
	  }
	  else if (l_new > 7) {
	    /* this is a complete new 'yyyymmdd' */
	    /* 	memmove(str_gcal,ptr_b,xmlStrlen(ptr_b)+1); */
	    str_tmp = xmlStrndup(str_gcal, 0);
	  }
	  else {
	    printf_log(2,"no valid extension '%s' in '%s'", ptr_b, str_gcal);
	    return NULL;
	  }
      }
      else {
	printf_log(2,"Different types between '%s' in '%s'", ptr_b, str_gcal);
	return NULL;
      }
      str_tmp = xmlStrcat(str_tmp, ptr_b);
      memcpy(str_gcal,str_tmp,xmlStrlen(str_tmp)+1);
      xmlFree(str_tmp);
    }
    return str_gcal;
  }

  return NULL;
}
/* end of assemble_next_date() */


/*! Searches in a CALENDAR DOM for the next node with same name like node_i.

\param node_i start node
\param distance between result and node_i

Both directions are supported. Distance values lower than zero means
backward searching.
*/
xmlNodePtr
next_element_cal(const xmlNodePtr node_i, int distance)
{
  BOOL_T flag_forward = TRUE;
  xmlNodePtr node_n;

  /*!\todo problems when a date definition starts before begin of
    calendar DOM (maybe projection from past to present)

    2005*w3mon:7 nor result in calender starting 2006

  */

  if (node_i && node_i->type == XML_ELEMENT_NODE) {

    if (xmlStrcmp(node_i->name,BAD_CAST"week")==0) {
      /* node_i is a WEEK element 
       */
      if (distance == 0) {
	/*! nearest distance is reached */
	return node_i;
      }
      else if (distance > 0) {  /* search forward */
	distance--;
      }
      else {  /*!\todo search backward */
	flag_forward = FALSE;
	distance++;
      }
      if (flag_forward) {
	return next_element_cal(next_element(node_i,NULL),distance);
      }
      else {
	assert(FALSE);		/*!\todo backward movement */
      }
    }
    else if (xmlStrcmp(node_i->name,BAD_CAST"day")==0) {
      /* node_i is a DAY element 
       */
      if (distance == 0) {
	/*! nearest distance is reached */
	return node_i;
      }
      else if (distance > 0) {  /* search forward */
	distance--;
      }
      else {  /*!\todo search backward */
	flag_forward = FALSE;
	distance++;
      }

      if (flag_forward) {
	if ((node_n = next_element(node_i,NULL))) {
	  return next_element_cal(node_n,distance);
	}
      }
      else {
	if (node_i->prev) {	/*!\todo use prev_element(node_i,NULL) */
	  /*!\bug when element in previuos period, node_i->prev==NULL */
	  return next_element_cal(node_i->prev,distance);
 	}
      }
      /*
	no next/prev DAY node in this MONTH found
       */
      if (node_i->parent) {
	xmlNodePtr node_pn;
	xmlNodePtr node_pnc;
	if (flag_forward) {
	  /* search for a node_i->parent->next->children */
	  for (node_pn=node_i->parent->next; node_pn; node_pn=node_pn->next) {
	    if ((node_pnc = child_element(node_pn,node_i->name))) {
	      /* use first DAY in next MONTH */
	      return next_element_cal(node_pnc,distance);
	    }
	  }
	}
	else {
	  for (node_pn=node_i->parent->prev; node_pn; node_pn=node_pn->prev) {
	    if (node_pn->last) {
	      /* use last DAY in prev MONTH */
	      return next_element_cal(node_pn->last,distance);
	    }
	  }
	}
      }
      /*
	no next/prev MONTH node in this YEAR found
      */
      if (node_i->parent->parent) {
	xmlNodePtr node_ppn;
	if (flag_forward) {
	  /* else search for a node_i->parent->parent->next->children->children */
	  for (node_ppn=node_i->parent->parent->next; node_ppn; node_ppn=node_ppn->next) {
	    if (node_ppn->children && node_ppn->children->children) {
	      /* use first DAY in next MONTH */
	      return next_element_cal(node_ppn->children->children,distance);
	    }
	  }
	}
	else {
	  for (node_ppn=node_i->parent->parent->prev; node_ppn; node_ppn=node_ppn->prev) {
	    if (node_ppn->last && node_ppn->last->last) {
	      /* use last DAY in last MONTH in prev YEAR */
	      return next_element_cal(node_ppn->last->last,distance);
	    }
	  }
	}
      }
    }
    else {
      /* other elements from recursive calls */
    }
  }
  return NULL;
}
/* End of next_element_cal() */


/*! sets the lower and upper values for year, according to date and done attributes
 */
BOOL_T
get_calendar_year_min_max(const xmlNodePtr node,int *year_min, int *year_max)
{
  if (IS_NODE(node,NULL)) {
    xmlChar *str_date;
    xmlNodePtr c;

    if ((str_date = attribute_str(node,BAD_CAST "done")) || (str_date = attribute_str(node,BAD_CAST "date"))) {
      printf_log(3,"DATE '%s'",str_date);
      if (isdigit(str_date[0]) && isdigit(str_date[1]) && isdigit(str_date[2]) && isdigit(str_date[3])) {
	int year = (str_date[0] - '0') * 1e3 + (str_date[1] - '0') * 1e2 + (str_date[2] - '0') * 1e1 + (str_date[3] - '0');
	if (year > 1900 && year < 3000) {
	  if (year < *year_min) {
	    *year_min = year;
	    if (*year_max == 0) {
	      *year_max = *year_min;
	    }
	  }
	  else if (year > *year_max) {
	    *year_max = year;
	    if (*year_min == 2999) {
	      *year_min = *year_max;
	    }
	  }
	}
      }
    }
    for (c = node->children; c; c=c->next) {
      get_calendar_year_min_max(c,year_min,year_max);
    }
    return TRUE;
  }
  else {
    return FALSE;
  }
}


/*! returns an array of year numbers

syntax of \@YEAR

"2007"
"2005..2007"

*/
int *
get_year_arr(const xmlNodePtr node_calendar)
{
  int *result = NULL;
  xmlChar *ptr_y_1;

  xmlChar *str_y = attribute_str(node_calendar,BAD_CAST "year");
  if (str_y != NULL && xmlStrEqual(str_y,BAD_CAST "content")) {
    /*! detect the calendar years from content */
    int year_min = 2999;
    int year_max = 0;
    int i;
    xmlNodePtr node_col;

    for (node_col = child_element(node_calendar,NAME_COL); 
	 node_col;
	 node_col = next_element(node_col,NAME_COL)) {
      get_calendar_year_min_max(node_col,&year_min,&year_max);
    }

    if (year_min == 2999) {
      /* no defined year found */
      year_min = today.tm_year + 1900;
    }
    if (year_max == 0) {
      /* no defined year found */
      year_max = today.tm_year + 1900;
    }
    /*!\todo check year_max-year_min < 10 */

    result = (int *)xmlMemMalloc((year_max-year_min+2)*sizeof(int));
    for (i=0; year_min+i <= year_max; i++) {
      result[i] = year_min + i;
    }
    result[i] = -1;
  }
  else if (str_y != NULL && (ptr_y_1 = BAD_CAST xmlStrstr(str_y, BAD_CAST ".."))) {
    /*
      string at str_y contains a year range
    */
    int y_0, y_1;
    y_0 = atoi((const char *)str_y);
    if (errno != ERANGE && y_0 > 1900) {
      y_1 = atoi((const char *)(ptr_y_1 + 2));
      if (errno != ERANGE) {
	if (y_0 <= y_1 || (y_1 - y_0) < 100) {
	  int i;
	  result = (int *)xmlMemMalloc((y_1 - y_0 + 2) * sizeof(int));
	  for (i=0; y_0 + i <= y_1; i++) {
	    result[i] = y_0 + i;
	  }
	  result[i] = -1;
	}
      }
    }
  }
  else {
    /*
      string at str_y contains a single year or nothing
    */
    result = (int *)xmlMemMalloc(2*sizeof(int));
    if (str_y==NULL) {
      result[0] = today.tm_year + 1900;
    }
    else {
      int y;
      y = atoi((const char *)str_y);
      if (errno == ERANGE) {
	result[0] = today.tm_year + 1900;
      }
      else {
	result[0] = y;
      }
    }
    result[1] = -1;
  }

  return result;
}
/* end of get_year_arr() */


/*! process the required calendar files
 */
xmlDocPtr
process_calendar(xmlNodePtr node_make_calendar)
{
  int *arr_y;
  int *ptr_y;
  xmlChar buffer[BUFFER_LENGTH];
  xmlDocPtr doc_result;
  xmlNodePtr node_calendar;
  xmlNodePtr node_col;
  xmlNodePtr node_meta;
  xmlChar *attr_year;

  printf_log(1,"Processing CALENDAR");

  /* mit Systemzeit initialisieren */
  time(&system_zeit);
  memcpy(&today, localtime(&system_zeit), sizeof(struct tm));

  /*! create DOM
   */
  doc_result = xmlNewDoc(BAD_CAST "1.0");
  node_calendar = xmlNewDocNode(doc_result, NULL, BAD_CAST "calendar", NULL); 
  xmlDocSetRootElement(doc_result, node_calendar);

  if ((attr_year = attribute_str(node_make_calendar,BAD_CAST "year"))) {
    xmlSetProp(node_calendar, BAD_CAST "year",attr_year);
  }

  node_meta = xmlNewChild(node_calendar, NULL, BAD_CAST "meta", NULL);
  xmlAddChild(node_meta,xmlCopyNode(node_make_calendar,1));
  /* Get the current time. */
  xmlStrPrintf(buffer,BUFFER_LENGTH,BAD_CAST "%i", (int)time(NULL));
  xmlSetProp(node_meta, BAD_CAST "mtime", BAD_CAST buffer);

  /*! preprocess COL entities */
  for (node_col = child_element(node_make_calendar,NAME_COL); 
       node_col;
       node_col = next_element(node_col,NAME_COL)) {
    xmlChar *id_col = attribute_str(node_col,BAD_CAST "id");
    xmlNodePtr node_colchild;

    /* copy col element */
    xmlNodePtr node_col_new = xmlNewChild(node_calendar,NULL,NAME_COL,NULL);
    xmlSetProp(node_col_new, BAD_CAST "id", id_col);

    for (node_colchild = node_col->children; node_colchild; node_colchild = node_colchild->next) {
      if (IS_NODE_XML(node_colchild)) {
	/*  */
	xmlDocPtr doc_pie;
	/*!\todo test call of process_xml() or process_make() */
	doc_pie = process_xml(node_colchild);	  /*  call */
	if (doc_pie) {
	  xmlNodePtr node_r = xmlDocGetRootElement(doc_pie);
	  if (IS_NODE_PIE(node_r)) {
	    printf_log(2,"Insert PIE DOM into calendar");
	    /*!\todo avoid copy of DOM, replace instead */
#if 0
	    xmlUnlinkNode(node_r);
	    xmlSetTreeDoc(node_r, doc_result);
	    xmlAddChild(node_col_new,node_r);
#else
	    xmlAddChild(node_col_new,xmlCopyNodeList(node_r));
#endif
	  }
	  else {
	    printf_log(1,"This is no valid PIE xml");
	  }
	  xmlFreeDoc(doc_pie);
	}
      }
      else if (IS_NODE_PIE(node_colchild)) {
	/*  */
	xmlAddChild(node_col_new,xmlCopyNodeList(node_colchild));
      }
    }
  }
  arr_y = get_year_arr(node_calendar);
  if (arr_y) {
    for (ptr_y=arr_y; ptr_y && *ptr_y>1900; ptr_y++) {
      printf_log(1,"Calendar for '%i'", *ptr_y);
      make_calendar(node_calendar, *ptr_y);
    }
    xmlMemFree(arr_y);

  /*! insert COL entities */
  for (node_col = child_element(node_calendar,NAME_COL); 
       node_col;
       node_col = next_element(node_col,NAME_COL)) {
    printf_log(2,"Copy PIE into calendar");
    copy_todo_to_calendar(node_calendar,attribute_str(node_col,BAD_CAST "id"),child_element(node_col,NAME_PIE));
    /*!\todo xmlUnlinkNode(node_col); */
  }

#if 1
  substitute_format(node_calendar);
#endif

/*   xmlSaveFormatFileEnc((const char *)"/tmp/a.pie", doc_result, NULL, 1); */
  }
  else {
  }

  return doc_result;
}


/*! Returns the WEEK node according to the other arguments.

\param node_parent YEAR node of CALENDAR
\param ptr_y pointer to YEAR/\@AD
\param ptr_woy pointer to YEAR/WEEK/\@OY
\param ptr_dow pointer to YEAR/MONTH/DAY/\@OWN
*/
xmlNodePtr
get_node_of_week(xmlNodePtr node_parent, xmlChar *ptr_y, xmlChar *ptr_woy, const xmlChar *ptr_dow)
{
  xmlNodePtr node_y;

  for (node_y = child_element(node_parent,BAD_CAST "year"); node_y; node_y = next_element(node_y,BAD_CAST "year")) {
    if (!xmlStrcasecmp(attribute_str(node_y,BAD_CAST "ad"),ptr_y)) {
      /* found the year */
      if (ptr_woy) {
	if (ptr_dow) {
	  /* look for the right week day in MONTH/DAY */
	  xmlNodePtr node_m;
	  xmlNodePtr node_d;
	  for (node_m = child_element(node_y,BAD_CAST "month"); node_m; node_m = next_element(node_m,BAD_CAST "month")) {
	    for (node_d = child_element(node_m,BAD_CAST "day"); node_d; node_d = next_element(node_d,BAD_CAST "day")) {
	      if (!xmlStrcasecmp(ptr_woy,attribute_str(node_d,BAD_CAST "cw"))
		  || !xmlStrcasecmp(ptr_woy,BAD_CAST "00")) {
		if (!ptr_dow || xmlStrcasecmp(ptr_dow,attribute_str(node_d,BAD_CAST "own")) == 0) {
		  return node_d;
		}
	      }
	    }
	  }
	}
	else {
	  /* look for the right WEEK */
	  xmlNodePtr node_w;
	  for (node_w = child_element(node_y,BAD_CAST "week"); node_w; node_w = next_element(node_w,BAD_CAST "week")) {
	    if (!xmlStrcasecmp(ptr_woy,attribute_str(node_w,BAD_CAST "nr"))
		|| !xmlStrcasecmp(ptr_woy,BAD_CAST "00")) {
	      return node_w;
	    }
	  }
	}
      }
    }
  }
  return NULL;
}


/*! Substitutes the format strings in childs of node.

\param node parent node
*/
BOOL_T
substitute_format(xmlNodePtr node)
{
  xmlNodePtr
    node_year,
    node_month,
    node_week,
    node_day,
    node_col,
    node_p,
    node_text;
  xmlChar
    *b,
    *c,
    *e,
    *d;
  xmlChar buffer[BUFFER_LENGTH];

  if (node==NULL) {
    return FALSE;
  }

  if (node->type == XML_ELEMENT_NODE && !xmlStrcasecmp(node->name, BAD_CAST "col")) {
    /*
      all COL children nodes
    */
    node_col = node;

    if (node_col->parent && !xmlStrcasecmp(node_col->parent->name, BAD_CAST "year")) {
      /* node for whole year */
      node_day   = NULL;
      node_week  = NULL;
      node_month = NULL;
      node_year  = node_col->parent;
    }
    else if (node_col->parent && !xmlStrcasecmp(node_col->parent->name, BAD_CAST "month")) {
      node_day   = NULL;
      node_week  = NULL;
      node_month = node_col->parent;
      node_year  = node_col->parent->parent;
    }
    else if (node_col->parent && !xmlStrcasecmp(node_col->parent->name, BAD_CAST "week")) {
      node_day   = NULL;
      node_week  = node_col->parent;
      node_month = NULL;
      node_year  = node_col->parent->parent;
    }
    else if (node_col->parent && !xmlStrcasecmp(node_col->parent->name, BAD_CAST "day")) {
      node_day   = node_col->parent;
      node_week  = NULL;
      node_month = node_col->parent->parent;
      node_year  = node_col->parent->parent->parent;
    }
    else {
    }

    for (node_p = node_col->children; node_p; node_p = node_p->next) {
      if (node_p->type == XML_ELEMENT_NODE 
	    && (!xmlStrcasecmp(node_p->name, BAD_CAST "link") || !xmlStrcasecmp(node_p->name, BAD_CAST "p"))) {
	/* all P children nodes */
	if (node_p->children && node_p->children->type == XML_TEXT_NODE && node_p->children->content) {
	  xmlChar *ptr_new;

	  node_text = node_p->children;
	  /* non-@free day of year */
	  b = node_text->content;
	  if ((c = BAD_CAST xmlStrstr(b,BAD_CAST "%N"))) {
	    /* search for all preceeding siblings in same COL in same
	       YEAR
	    */
	    xmlNodePtr node_d;
	    xmlNodePtr node_c;
	    xmlNodePtr node_pi;
	    int n;
	    xmlChar *str_free, *str_holiday, *str_ow, *str_id_col;

	    str_id_col = attribute_str(node_col,BAD_CAST "name");

	    if (node_day && node_year) {
	      for (n=0, node_d = child_element(child_element(node_year,BAD_CAST "month"),BAD_CAST "day");
		   node_d;
		   node_d = next_element_cal(node_d,+1)) {
		/* step over all day nodes */
		/*!\todo set n for every col separately, perfomance? */
		for (node_c = child_element(node_d,NAME_COL);
		     node_c;
		     node_c = node_c->next) {
		  /* step over all col nodes */
		  for (node_pi = child_element(node_c,BAD_CAST "p");
		       node_pi;
		       node_pi = node_pi->next) {
		    /* step over all p nodes */
		    if (((str_free = attribute_str(node_pi,BAD_CAST "free"))
			 && !xmlStrcasecmp(str_free,BAD_CAST "yes")
			 && !xmlStrcasecmp(attribute_str(node_c,BAD_CAST "name"),str_id_col))
			|| 
			((str_holiday = attribute_str(node_pi,BAD_CAST "holiday"))
			 && !xmlStrcasecmp(str_holiday,BAD_CAST "yes"))) {
		      /* a p[@free='yes'] in a col[@id=str_id_col] found */
		      /* or */
		      /* a p[@holiday='yes'] found */
		      /* dont count this day */
		      goto skip_day;
		    }
		  }
		}
		n++;
	      skip_day:
		if (node_d == node_day) {
		  /* day node itself found */
		  break;
		}
	      }
	    }
	    xmlStrPrintf(buffer,BUFFER_LENGTH,BAD_CAST "%i", n);
	    ptr_new = str_replace(b,BAD_CAST "%N",buffer);
	    xmlNodeSetContent(node_text,ptr_new);
	    xmlFree(ptr_new);
	  }
	  /* day of year */
	  b = node_text->content;
	  if ((c = BAD_CAST xmlStrstr(b,BAD_CAST "%j"))) {
	    if ((d = attribute_str(node_day,BAD_CAST "oy"))) {
	      ptr_new = str_replace(b,BAD_CAST "%j",d);
	      xmlNodeSetContent(node_text,ptr_new);
	      xmlFree(ptr_new);
	    }
	  }
	  /* day of week */
	  b = node_text->content;
	  if ((c = BAD_CAST xmlStrstr(b,BAD_CAST "%a"))) {
	    if ((d = attribute_str(node_day,BAD_CAST "own"))) {
	      ptr_new = str_replace(b,BAD_CAST "%a",d);
	      xmlNodeSetContent(node_text,ptr_new);
	      xmlFree(ptr_new);
	    }
	  }
	  /* day of month */
	  b = node_text->content;
	  if ((c = BAD_CAST xmlStrstr(b,BAD_CAST "%d"))) {
	    if ((d = attribute_str(node_day,BAD_CAST "om"))) {
	      ptr_new = str_replace(b,BAD_CAST "%d",d);
	      xmlNodeSetContent(node_text,ptr_new);
	      xmlFree(ptr_new);
	    }
	  }
	  /* month of year */
	  b = node_text->content;
	  if ((c = BAD_CAST xmlStrstr(b,BAD_CAST "%m"))) {
	    if ((d = attribute_str(node_month,BAD_CAST "nr"))) {
	      ptr_new = str_replace(b,BAD_CAST "%m",d);
	      xmlNodeSetContent(node_text,ptr_new);
	      xmlFree(ptr_new);
	    }
	  }
	  /* ISO week of year */
	  b = node_text->content;
	  if ((c = BAD_CAST xmlStrstr(b,BAD_CAST "%V"))) {
	    if ((node_week && (d = attribute_str(node_week,BAD_CAST "nr")))
		|| (d = attribute_str(node_day,BAD_CAST "cw"))) {
	      ptr_new = str_replace(b,BAD_CAST "%V",d);
	      xmlNodeSetContent(node_text,ptr_new);
	      xmlFree(ptr_new);
	    }
	  }
	  /* year */
	  b = node_text->content;
	  if ((c = BAD_CAST xmlStrstr(b,BAD_CAST "%Y"))) {
	    if ((d = attribute_str(node_year,BAD_CAST "ad"))) {
	      ptr_new = str_replace(b,BAD_CAST "%Y",d);
	      xmlNodeSetContent(node_text,ptr_new);
	      xmlFree(ptr_new);
	    }
	  }
	  /* difference in years */
	  b = node_text->content;
	  if ((c = BAD_CAST xmlStrstr(b,BAD_CAST "%B"))) {
	    if ((d = attribute_str(node_year,BAD_CAST "ad"))) {
	      xmlChar *c_e;
	      xmlChar c_buffer[BUFFER_LENGTH];
	      int y_p = (int) strtol((char *)d,NULL,10);
	      int y_c = (int) strtol((char *)c+2,(char **)&c_e,10);
	      xmlStrPrintf(c_buffer,BUFFER_LENGTH,BAD_CAST "%i", y_p - y_c);

	      if (c-b > 0) {
		e = xmlStrndup(b,c-b);
		e = xmlStrcat(e,c_buffer);
	      }
	      else {
		e = xmlStrdup(c_buffer);
	      }
	      e = xmlStrcat(e,c_e);
	      xmlNodeSetContent(node_text,e);
	      xmlFree(e);
	    }
	  }
	}
      }
    }
    substitute_format(node->next);
  }
  else {
    /*!\todo other elements than p ? */
    substitute_format(node->children);
    substitute_format(node->next);
  }
  return TRUE;
}
/* end of substitute_format() */


/*! Inserts the node at the according COL element of 'node'

\param node in calendar DOM
\param col_id \@ID of COL element
\param node_insert node to insert
*/
BOOL_T
insert_node_cal_col(xmlNodePtr node, xmlChar *col_id, xmlNodePtr node_insert)
{
  xmlNodePtr node_new;
  xmlNodePtr node_col;
  xmlNodePtr node_child;
xmlAttrPtr attr_del;

  if (node == NULL || node_insert == NULL)
    return FALSE;

  /* check COL */
  if (col_id) {
    for (node_col = child_element(node,NAME_COL);
	 node_col && xmlStrcasecmp(col_id,attribute_str(node_col,BAD_CAST "name"));
	 node_col = next_element(node_col,NAME_COL)) {}
    if (node_col == NULL) {
      /* create one before */
      node_col = xmlNewChild(node, NULL, NAME_COL, NULL);
      xmlSetProp(node_col, BAD_CAST "name", col_id);
    }
  }
  else {
    node_col = node;
  }
  /* check if there is a sibling with same content 
   */
  for (node_child = node_col->children;
       node_child;
       node_child = node_child->next) {
    if (node_is_equal(node_child,node_insert)) {
      printf_log(3,"Ignore redundant entry '%s'", (unsigned char*) node_insert->name);
      return FALSE;
    }
  }
  node_new = xmlCopyNodeList(node_insert);
  printf_exit(node_new==NULL,23,"xmlCopyNodeList() failed");
#if 0
  if ((attr_del = xmlHasProp(node_new, BAD_CAST "date"))) {
      xmlRemoveProp(attr_del);
  }
  if ((attr_del = xmlHasProp(node_new, BAD_CAST "done"))) {
      xmlRemoveProp(attr_del);
  }
#endif
  /* append to the according COL element */
  return (xmlAddChildList(node_col,node_new) != NULL);
}
/* end of insert_node_cal_col() */


/*! detects the node for 'cal_str' and calls insert_node_cal_col()
 */
BOOL_T
insert_node_cal(xmlNodePtr node_parent, xmlChar *col_id, xmlChar *cal_str, xmlNodePtr node_insert)
{
  int y, m, w, d, d_w;
  xmlNodePtr node_y;
  xmlNodePtr node_m;
  xmlNodePtr node_d;
  xmlChar buffer_y[BUFFER_LENGTH];
  xmlChar buffer_m[BUFFER_LENGTH];
  xmlChar buffer_d[BUFFER_LENGTH];
  xmlChar buffer_woy[BUFFER_LENGTH];
  xmlChar *ptr_sep = NULL;
  BOOL_T flag_easter = FALSE;	/* marks easter date dependency */
  BOOL_T flag_eternity = FALSE;
  BOOL_T flag_done = FALSE;

  /*!\bug if before easter dates \@e-52 */

  if (node_parent==NULL || cal_str==NULL || *cal_str=='\0') {
    return FALSE;
  }

  ptr_sep = sscanf_datum(cal_str,&y,&m,&d,&w,&d_w);
  if (ptr_sep==NULL) {
    return FALSE;
  }

  /* its a valid date -> process 
   */
  xmlStrPrintf(buffer_y,BUFFER_LENGTH,BAD_CAST "%04i",y);
  if (ptr_sep && ptr_sep[0]=='@' && ptr_sep[1]=='e') {
    flag_easter = TRUE;
    ptr_sep += 2;
  }
  if (y == 0) {
    flag_eternity = TRUE;
  }
  for (node_y = child_element(node_parent,BAD_CAST "year");
       node_y;
       node_y = next_element(node_y,BAD_CAST "year")) {
    if (flag_eternity || !xmlStrcasecmp(buffer_y,attribute_str(node_y,BAD_CAST "ad"))) {
      if (flag_easter) {
	/* detect easter references */
	int y_n = atoi((char *)attribute_str(node_y,BAD_CAST "ad"));
	easter1(y_n,&d,&m);
	printf_log(2,"Insert Easter '%04i%02i%02i' in '%s'", y_n, m, d, cal_str);
      }
      if (w == 0) {
	if (d_w < 0) {
	  /* eternal week */
	  xmlNodePtr node_n;
	  int i;
	  printf_log(2,"Eternal Week");
	  node_d = get_node_of_week(node_parent,attribute_str(node_y,BAD_CAST "ad"),BAD_CAST "00", NULL);
	  for (node_n=node_d, i=0; i<1000; i++) {
	    insert_node_cal_col(node_n,col_id,node_insert);
	    node_n = next_element_cal(node_n,+1);
	  }
	}
	else {
	  /* eternal day of week */
	  xmlNodePtr node_n;
	  int i;
	  node_d = get_node_of_week(node_parent,attribute_str(node_y,BAD_CAST "ad"),BAD_CAST "00", dow[d_w]);
	  for (node_n=node_d, i=0; i<1000; i++) {
	    insert_node_cal_col(node_n,col_id,node_insert);		    
	    node_n = next_element_cal(node_n,+7);
	  }
	}
	insert_node_cal(node_parent, col_id, assemble_next_date(cal_str), node_insert);
      }
      else if (w < 0 && m < 0 && d < 0) {
	/* whole year */
	insert_node_cal_col(node_y,col_id,node_insert);		    
      }
      else if (w > 0) {
	xmlStrPrintf(buffer_woy,BUFFER_LENGTH,BAD_CAST "%02i",w);
	if (d_w > -1) {
	  /* entry for a specified day of week */
	  node_d = get_node_of_week(node_parent,attribute_str(node_y,BAD_CAST "ad"),buffer_woy, dow[d_w]);
	  /* begin of copy */
	  if (*ptr_sep == '+' || *ptr_sep == '-') {
	    /* find the '+-n' DAY */
	    int n = atoi((char *)ptr_sep);
	    if (n < 1000)
	      insert_node_cal_col(next_element_cal(node_d,n),col_id,node_insert);		    
	  }
	  else if (*ptr_sep == ':') {
	    /* find the next ':n' DAY */
	    xmlNodePtr node_n;
	    int i;
	    int n;
	    /*!\todo remove following workaround */
	    if (ptr_sep[0] == ':' && ptr_sep[1] == '9' && ptr_sep[2] == '9' && ptr_sep[3] == '9') {
	      /* skip ':999' of gcal */
	      printf_log(2,"Modifying '%s'", cal_str);
	      memmove(ptr_sep,&ptr_sep[4],xmlStrlen(ptr_sep)+1);
	      /* ptr_sep += 4;  */
	      insert_node_cal(node_parent, col_id, cal_str, node_insert);			
	    }
	    else {
	      n = atoi((char *)ptr_sep+1);
	      for (node_n=node_d, i=0; i<n && i<1000; i++) {
		insert_node_cal_col(node_n,col_id,node_insert);		    
		node_n = next_element_cal(node_n,+1);
	      }
	    }
	  }
#if 0
	  else if (*ptr_sep == '#') {
	    /* find the DAYs to #... */
	    long int n, m;
	    insert_node_cal_col(node_d,col_id,node_insert);
	    m = get_day_abs_str(assemble_next_date(cal_str));
	    n = m - get_day_abs(y,-1,-1,w,d_w);;
	    if (n<1) {
	      printf_log(2,"Datum is a backward distance '%s'", cal_str);
	    }
	    else {
	      xmlNodePtr node_n;
	      for (node_n=node_d; n>0; n--) {
		node_n = next_element_cal(node_n,+1);
		insert_node_cal_col(node_n,col_id,node_insert);		    
	      }
	    }
	  }
#else
	  else if (*ptr_sep == '#') {
	    /* find the DAYs to #... */
	    xmlChar *str_next;
	    insert_node_cal_col(node_d,col_id,node_insert);
	    str_next = assemble_next_date(cal_str);
	    if (str_next && xmlStrlen(str_next)) {
	      /* 			  int n = get_day_abs_str(str_next) - get_day_abs(y,m,d,-1,-1); */
	      int n = get_day_abs_str(str_next) - get_day_abs(y,-1,-1,w,d_w);
	      if (n<1) {
		printf_log(2,"Datum is a backward distance '%s'", cal_str);
	      }
	      else {
		xmlNodePtr node_n;
		for (node_n=node_d; n>0; n--) {
		  node_n = next_element_cal(node_n,+1);
		  insert_node_cal_col(node_n,col_id,node_insert);		    
		}
	      }
	    }
	    else {
	      /* s. message from assemble_next_date() */
	    }
	  }
#endif
	  else if (*ptr_sep == '.') {
	    /* find the every 'n'th DAY */
	    xmlNodePtr node_next;
	    int i;
	    int n = (int) strtol((char *)ptr_sep+1,(char **)&ptr_sep,10);
	    int m = (int) ((*ptr_sep == 'x') ? strtol((char *)ptr_sep+1,(char **)&ptr_sep,10) : -1);
	    for (node_next=node_d, i=0;
		 ((m>-1 && i<m) || m<0) && node_next;
		 node_next = next_element_cal(node_next,n), i++) {
	      insert_node_cal_col(node_next,col_id,node_insert);		    
	    }
	  }
	  else if (*ptr_sep == ',') {
	    /* recursion */
	    insert_node_cal_col(node_d,col_id,node_insert);
	    while (assemble_next_date(cal_str)) {
	      insert_node_cal(node_parent, col_id, cal_str, node_insert);			
	    }
	  }
	  else {
	    insert_node_cal_col(node_d,col_id,node_insert);
	  }
	  /* end of copy */
	}
	else {
	  /* entry for a week */
	  node_d = get_node_of_week(node_parent,attribute_str(node_y,BAD_CAST "ad"),buffer_woy,NULL);
	  insert_node_cal_col(node_d,col_id,node_insert);
	}
	insert_node_cal(node_parent, col_id, assemble_next_date(cal_str), node_insert);
      } /* end of week processing */
      else if (m > -1) {
	xmlStrPrintf(buffer_m,BUFFER_LENGTH,BAD_CAST "%02i", m);
	xmlStrPrintf(buffer_d,BUFFER_LENGTH,BAD_CAST "%02i", d);
	for (node_m = child_element(node_y,BAD_CAST "month"); node_m; node_m = next_element(node_m,BAD_CAST "month")) {
	  if (m==0 || !xmlStrcasecmp(buffer_m,attribute_str(node_m,BAD_CAST "nr"))) {
	    if (d > -1) {
	      for (node_d = child_element(node_m,BAD_CAST "day"); node_d; node_d = next_element(node_d,BAD_CAST "day")) {
		if (d==0 || !xmlStrcasecmp(buffer_d,attribute_str(node_d,BAD_CAST "om"))) {
		  /* begin: copy to above */
		  if (*ptr_sep == '+' || *ptr_sep == '-') {
		    /* find the '+-n' DAY */
		    int n = atoi((char *)ptr_sep);
		    if (n < 1000)
		      insert_node_cal_col(next_element_cal(node_d,n),col_id,node_insert);		    
		  }
		  else if (*ptr_sep == ':') {
		    /* find the next ':n' DAY */
		    xmlNodePtr node_n;
		    int i;
		    int n;
		    /*!\todo remove following workaround */
		    if (ptr_sep[0] == ':' && ptr_sep[1] == '9' && ptr_sep[2] == '9' && ptr_sep[3] == '9') {
		      /* skip ':999' of gcal */
		      printf_log(2,"Modifying '%s'", cal_str);
		      memmove(ptr_sep,&ptr_sep[4],xmlStrlen(ptr_sep)+1);
		      /* ptr_sep += 4;  */
		      insert_node_cal(node_parent, col_id, cal_str, node_insert);			
		    }
		    else {
		      n = atoi((char *)ptr_sep+1);
		      for (node_n=node_d, i=0; i<n && i<1000; i++) {
			insert_node_cal_col(node_n,col_id,node_insert);		    
			node_n = next_element_cal(node_n,+1);
		      }
		    }
		  }
		  else if (*ptr_sep == '#') {
		    /* find the DAYs to #... */
		    xmlChar *str_next;
		    insert_node_cal_col(node_d,col_id,node_insert);
		    str_next = assemble_next_date(cal_str);
		    if (str_next && xmlStrlen(str_next)) {
		      int n = get_day_abs_str(str_next) - get_day_abs(y,m,d,-1,-1);
		      /* 			  int n = get_day_abs_str(str_next) - get_day_abs(y,-1,-1,w,d_w); */
		      if (n<1) {
			printf_log(2,"Datum is a backward distance '%s'", cal_str);
		      }
		      else {
			xmlNodePtr node_n;
			for (node_n=node_d; n>0; n--) {
			  node_n = next_element_cal(node_n,+1);
			  insert_node_cal_col(node_n,col_id,node_insert);		    
			}
		      }
		    }
		    else {
		      /* s. message from assemble_next_date() */
		    }
		  }
		  else if (*ptr_sep == '.') {
		    /* find the every 'n'th DAY */
		    xmlNodePtr node_next;
		    int i;
		    int n = (int) strtol((char *)ptr_sep+1,(char **)&ptr_sep,10);
		    int m = (int) ((*ptr_sep == 'x') ? strtol((char *)ptr_sep+1,(char **)&ptr_sep,10) : -1);
		    for (node_next=node_d, i=0;
			 ((m>-1 && i<m) || m<0) && node_next;
			 node_next = next_element_cal(node_next,n), i++) {
		      insert_node_cal_col(node_next,col_id,node_insert);		    
		    }
		  }
		  else if (*ptr_sep == ',') {
		    /* recursion */
		    insert_node_cal_col(node_d,col_id,node_insert);
		    while (assemble_next_date(cal_str)) {
		      insert_node_cal(node_parent, col_id, cal_str, node_insert);			
		    }
		  }
		  else {
		    insert_node_cal_col(node_d,col_id,node_insert);
		  }
		  /* end: copy to above */
		}
	      }
	    }
	    else {
	      /* entry for this MONTH */
	      insert_node_cal_col(node_m,col_id,node_insert);
	    }
	  }
	}
	insert_node_cal(node_parent, col_id, assemble_next_date(cal_str), node_insert);
      } /* end of month processing */
      else {
#if 0
	/* entry for this YEAR */
	insert_node_cal_col(node_y,col_id,node_insert);
#else
	assert(TRUE);
#endif
      }
      flag_done = TRUE;
    }
  }
  return FALSE;
}
/* end of insert_node_cal() */


/*! creates a DOM for a whole 'year'
*/
BOOL_T
make_calendar(xmlNodePtr node_parent, int year)
{
  int month_current;
  int week_of_year;
  struct tm t, t_1_1;
  xmlChar buffer[BUFFER_LENGTH];
  xmlNodePtr node_year;

  xmlStrPrintf(buffer,BUFFER_LENGTH,
	       BAD_CAST "%04i%02i%02i",
	       today.tm_year + 1900, today.tm_mon + 1, today.tm_mday);
  xmlSetProp(node_parent, BAD_CAST "today", buffer);

  if (year > 999)
    year -= 1900;

  t.tm_year = year; 
  t.tm_yday = 0;
  t.tm_mon  = 0;
  t.tm_mday = 0;
  t.tm_wday = 0;
  t.tm_hour = 0;
  t.tm_min  = 0;
  t.tm_sec  = 0;
  t.tm_isdst  = 0;

  memcpy(&t_1_1, &t, sizeof(struct tm));
  t_1_1.tm_mday++;
  mktime(&t_1_1);
  week_of_year = (t_1_1.tm_wday > 4 || t_1_1.tm_wday < 1) ? 0 : 1;

  node_year = xmlNewChild(node_parent, NULL, BAD_CAST "year", NULL);
  xmlStrPrintf( buffer,BUFFER_LENGTH,BAD_CAST "%i", year +  1900);
  xmlSetProp(node_year, BAD_CAST "ad", buffer);

  month_current = -1;
  for( ; ; )
    {
      xmlNodePtr node_month;
      xmlNodePtr node_day;

      t.tm_mday++;
      if (mktime(&t) == -1 || t.tm_year != year) {
	break;
      }

      if (t.tm_mon > month_current) {
	node_month = xmlNewChild(node_year, NULL, BAD_CAST "month", NULL);
	xmlStrPrintf( buffer,BUFFER_LENGTH,BAD_CAST "%02i", t.tm_mon + 1);
	xmlSetProp(node_month, BAD_CAST "nr", buffer);
	xmlSetProp(node_month, BAD_CAST "name", moy[t.tm_mon]);
	month_current = t.tm_mon;
      }

      node_day = xmlNewChild(node_month, NULL, BAD_CAST "day", NULL);

      xmlStrPrintf( buffer,BUFFER_LENGTH,BAD_CAST "%02i", t.tm_mday);
      xmlSetProp(node_day, BAD_CAST "om", buffer);

      xmlStrPrintf( buffer,BUFFER_LENGTH,BAD_CAST "%i", t.tm_yday + 1);
      xmlSetProp(node_day, BAD_CAST "oy", buffer);

      xmlStrPrintf( buffer,BUFFER_LENGTH,BAD_CAST "%i", t.tm_wday);
      xmlSetProp(node_day, BAD_CAST "ow", buffer);
      /* attribute is used in get_node_of_week() */
      xmlSetProp(node_day, BAD_CAST "own", dow[t.tm_wday]);

      xmlStrPrintf( buffer,BUFFER_LENGTH,BAD_CAST "%02i", week_of_year);
      xmlSetProp(node_day, BAD_CAST "cw", buffer);

      if (t.tm_mday == today.tm_mday
	  && t.tm_mon == today.tm_mon
	  && t.tm_year == today.tm_year ) {
	xmlSetProp(node_day, BAD_CAST "today", BAD_CAST "yes");
      }
      if( t.tm_wday == 0 ) { /* am Sonntag */
	xmlNodePtr node_week;
	node_week = xmlNewChild(node_year, NULL, BAD_CAST "week", NULL);
	xmlStrPrintf( buffer,BUFFER_LENGTH,BAD_CAST "%02i", week_of_year);
	xmlSetProp(node_week, BAD_CAST "nr", buffer);
	week_of_year++;
      }
    }
  return TRUE;
}
/* end of make_calendar() */

