/* 
   cxproc - Configurable Xml PROCessor

   Copyright (C) 2006,2007,2008 by Alexander Tenbusch

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

*/

/*! pie main frontend
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <assert.h>
#include <limits.h>

/* 
 */
#include <libxml/HTMLtree.h>
#include <libxml/parser.h>

/* 
 */
#include <libxslt/xslt.h>
#include <libxslt/xsltutils.h>
#include <libxslt/transform.h>
#include <libxslt/variables.h>

#include <iconv.h>

/* 
 */
#include "cgi-util.h"

/* 
 */
#include "basics.h"
#include "tools.h"
#include "iconv_x.h"
#include "calendar_x.h"
#include "make_x.h"
#include "plain_x.h"
#include "sql_x.h"
#include "dom_x.h"


int
main (int argc, char **argv, char **envp)
{
  xmlDocPtr doc_argv, doc_cgi;
  xmlChar *str_method;
  char *ptr;

  if ((ptr = getenv("CXP_LOG"))) {
    set_level_log(atoi((*ptr=='\"') ? (ptr + 1) : ptr));
  }
  else {
    set_level_log(1);
  }

  printf_log(1,"**************************************************************");

  printf_log(2,"Based on $HeadURL: file:///media/TENBUSCH_A/repos/cxproc/tags/0.7/src/cxproc.c $");

  LIBXML_TEST_VERSION

  if (xmlInitMemory()==0) {
    exit(EXIT_FAILURE);
  }

  if (!set_encodingsUTF8(
#ifdef _MSC_VER
      BAD_CAST "ISO-8859-1",BAD_CAST "UTF-8",BAD_CAST "ISO-8859-1",BAD_CAST "ISO-8859-1",BAD_CAST "ISO-8859-1"
#else
      BAD_CAST "UTF-8",BAD_CAST "UTF-8",BAD_CAST "ISO-8859-1",BAD_CAST "ISO-8859-1",BAD_CAST "ISO-8859-1"
#endif
      )) {
    exit(EXIT_FAILURE);
  }

  /* register for exit() */
  if( atexit(exit_debug) != 0
      || atexit(free_make) != 0
      || atexit(free_iconv) != 0
      || atexit(free_subst) != 0
      || atexit(free_dom) != 0
      || atexit(xsltCleanupGlobals) != 0
      || atexit(xmlCleanupParser) != 0
      || atexit(xmlMemoryDump) != 0) {
    exit(EXIT_FAILURE);
  }

  if ((str_method = getenvUTF8(BAD_CAST "REQUEST_METHOD"))!=NULL
      && (xmlStrcmp(str_method,BAD_CAST "GET")==0 || xmlStrcmp(str_method,BAD_CAST "POST")==0)) {
    /*
      this is the CGI mode
    */

    /* Initialize the CGI */
    int res = cgi_init();

    /* Was there an error initializing the CGI??? */
    printf_exit((res != CGIERR_NONE), 1, "%s", cgi_strerror(res));

    runmode = cgi;

    if( atexit(cgi_quit) != 0)
      exit(EXIT_FAILURE);
    
    doc_cgi = process_cgi(argc,argv,envp);
    if (doc_cgi) {
#if 1
/*       printf("Content-type: text/plain\n\n"); */
      process_make(xmlDocGetRootElement(doc_cgi));
#else
      printf("Content-type: text/plain\n\n");
      printf("Content-type: text/xml\n\n");
      xmlSaveFormatFileEnc((const char *) "-", doc_cgi, NULL, 1);
#endif
      xmlFreeDoc(doc_cgi);
    }
  }
  else {
    /*
      this is the command line mode (not CGI mode!)
    */
    runmode = command;
    doc_argv = process_argv(argc,argv,envp);
    if (doc_argv) {
#if 1
      process_make(xmlDocGetRootElement(doc_argv));
#else
      xmlSaveFormatFileEnc((const char *) "-", doc_argv, NULL, 1);
#endif
      xmlFreeDoc(doc_argv);
    }
  }

  exit(EXIT_SUCCESS);
}

