/*
  cxproc - Configurable Xml PROCessor

  Copyright (C) 2006,2007,2008 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/types.h>
#ifdef _MSC_VER
#include <direct.h>
#include <io.h>
#include <time.h>
#else
#include <sys/dir.h>
#include <unistd.h>
#include <dirent.h>
#endif

/* 
 */
#include <libxml/parser.h>

#include <iconv.h>

#include <pcre.h>

/* 
 */
#ifdef _MSC_VER
#include "config-win32.h"
#else
#include "../config.h"
#endif

#if HAVE_LIBID3TAG
#include <id3tag.h>
#endif

#ifdef HAVE_LIBMAGICK
#include <magick/ImageMagick.h>
#endif

#include "basics.h"
#include "iconv_x.h"
#include "cxp_dtd.h"
#include "dom_x.h"
#include "dir_x.h"
#include "plain_x.h"
#include "make_x.h"
#include "tools.h"


static xmlDocPtr doc_search = NULL;	/*! pointer to a file search DOM */

static BOOL_T
add_content_type(xmlNodePtr node_file, xmlChar *file_name, int level_verbose);

#if HAVE_LIBID3TAG

static xmlChar *
input_id3_get_string(struct id3_tag *tag, char *frame_name);

static BOOL_T
add_id3_tags(xmlNodePtr node_file, char *file_name);

#endif

#ifdef HAVE_LIBMAGICK

static BOOL_T
add_image_info(xmlNodePtr node_file, char *file_name);

/*!

 */
BOOL_T
add_image_info(xmlNodePtr node_file, char *file_name)
{
  xmlNodePtr node_info;
  xmlNodePtr node_info_child;
  ExceptionInfo *exception;
  ImageInfo *image_info;
  Image *image;

  const char *property;

  exception=AcquireExceptionInfo();
  image_info=AcquireImageInfo();

  /*
    Read an image.
  */
  printf_log(3,"Read IMAGE info from file '%s'",file_name);

  CopyMagickString(image_info->filename,file_name,MaxTextExtent);
  image=ReadImage(image_info,exception);
  CatchException(exception);
  DestroyExceptionInfo(exception);

  if (image == NULL) {
    printf_log(1,"Read Error");
    xmlSetProp(node_file, BAD_CAST "type", BAD_CAST"error/format");
    DestroyImageInfo(image_info);
    return FALSE;
  }
  else {
    xmlChar buffer_value[BUFFER_LENGTH];
    char *value;

    xmlStrPrintf(buffer_value,sizeof(buffer_value),BAD_CAST"image/%s",image->magick);
    xmlSetProp(node_file, BAD_CAST "type", buffer_value);

    node_info = xmlNewChild(node_file, NULL, BAD_CAST"image", NULL);
    node_info_child = xmlNewChild(node_info, NULL, BAD_CAST"size",NULL);
    /*  */
    xmlStrPrintf(buffer_value,sizeof(buffer_value),BAD_CAST"%i",image->columns);
    xmlSetProp(node_info_child, BAD_CAST "col",buffer_value);
    /*  */
    xmlStrPrintf(buffer_value,sizeof(buffer_value),BAD_CAST"%i",image->rows);
    xmlSetProp(node_info_child, BAD_CAST "row",buffer_value);

    node_info_child = xmlNewChild(node_info, NULL, BAD_CAST"color",NULL);
    /*  */
    xmlStrPrintf(buffer_value,sizeof(buffer_value),BAD_CAST"%i",image->depth);
    xmlSetProp(node_info_child, BAD_CAST "depth",buffer_value);
    /*  */
    xmlStrPrintf(buffer_value,sizeof(buffer_value),BAD_CAST"%i",image->colors);
    xmlSetProp(node_info_child, BAD_CAST "colors",buffer_value);

    node_info_child = xmlNewChild(node_info, NULL, BAD_CAST"resolution",NULL);
    /*  */
    xmlStrPrintf(buffer_value,sizeof(buffer_value),BAD_CAST"%.3f",image->x_resolution);
    xmlSetProp(node_info_child, BAD_CAST "x",buffer_value);
    /*  */
    xmlStrPrintf(buffer_value,sizeof(buffer_value),BAD_CAST"%.3f",image->y_resolution);
    xmlSetProp(node_info_child, BAD_CAST "y",buffer_value);

    /* default is DPI */
    xmlSetProp(node_info_child,
	       BAD_CAST "dotper",
	       BAD_CAST ((image->units == PixelsPerCentimeterResolution) ? "cm" : "inch"));

    /*
      code derived from ImageMagick-6.4.3/magick/identify.c
    */
    GetImageProperty(image,"exif:*");
    ResetImagePropertyIterator(image);
    property = GetNextImageProperty(image);
    if (property) {
      /* add image properties. */
      /* xmlNodePtr node_info_exif; */
      xmlNodePtr node_info_prop;

      /* node_info_exif = xmlNewChild(node_info, NULL, BAD_CAST"exif",NULL); */
      while (property != (const char *) NULL) {
	if (strlen(property) > 1 && xmlStrstr(property,BAD_CAST"xmlns")==NULL) {
	  value=GetImageProperty(image,property);
	  if (value) {
	    node_info_prop = xmlNewChild(node_info, NULL, BAD_CAST property,NULL);
	    xmlSetProp(node_info_prop, BAD_CAST "value",BAD_CAST value);
	  }
	}
	property=GetNextImageProperty(image);
      }
    }
    DestroyImage(image);
  }
  DestroyImageInfo(image_info);

  return TRUE;
}
/* end of add_image_info() */

#endif


#if HAVE_LIBID3TAG

/* from xmms-mad
*/
xmlChar *
input_id3_get_string(struct id3_tag *tag, char *frame_name)
{
  const id3_ucs4_t *string;
  struct id3_frame *frame;
  union id3_field *field;

  frame = id3_tag_findframe (tag, frame_name, 0);
  if (!frame) return NULL;

  if (frame_name == ID3_FRAME_COMMENT)
      field = id3_frame_field (frame, 3);
  else
      field = id3_frame_field (frame, 1);

  if (!field) return NULL;

  if (frame_name == ID3_FRAME_COMMENT)
      string = id3_field_getfullstring (field);
  else
      string = id3_field_getstrings (field, 0);

  if (!string) return NULL;

  if (frame_name == ID3_FRAME_GENRE)
     string = id3_genre_name (string);

  if (strlen(string) > 0) {
    return id3_ucs4_utf8duplicate(string);
  }
  return NULL;
}


/*! Tags in audio files

  s. http://www.fireflymediaserver.org/ mt-daapd-0.2.3.tar.gz 
  or libid3tag http://www.underbit.com/products/mad/
*/
BOOL_T
add_id3_tags(xmlNodePtr node_file, char *file_name)
{
  xmlNodePtr node_id3;
  xmlNodePtr node_id3_tag;
  struct id3_file *id3file = NULL;

  node_id3 = xmlNewChild(node_file, NULL, BAD_CAST"id3", NULL);

  id3file = id3_file_open(file_name, ID3_FILE_MODE_READONLY);
  if (id3file) {
    struct id3_tag *tag = NULL;

    tag = id3_file_tag(id3file);
    if (tag) {
      xmlChar *ptr_release;

      ptr_release = input_id3_get_string(tag, ID3_FRAME_TRACK);
      if (ptr_release) {
	node_id3_tag = xmlNewChild(node_id3, NULL, BAD_CAST"track",ptr_release);
	xmlFree(ptr_release);
      }

      ptr_release = input_id3_get_string(tag, ID3_FRAME_ARTIST);
      if (ptr_release) {
	node_id3_tag = xmlNewChild(node_id3, NULL, BAD_CAST"artist",ptr_release);
	xmlFree(ptr_release);
      }

      ptr_release = input_id3_get_string(tag, ID3_FRAME_ALBUM);
      if (ptr_release) {
	node_id3_tag = xmlNewChild(node_id3, NULL, BAD_CAST"album",ptr_release);
	xmlFree(ptr_release);
      }

      ptr_release = input_id3_get_string(tag, ID3_FRAME_TITLE);
      if (ptr_release) {
	node_id3_tag = xmlNewChild(node_id3, NULL, BAD_CAST"title",ptr_release);
	xmlFree(ptr_release);
      }

      ptr_release = input_id3_get_string(tag, "TLEN");
      if (ptr_release) {
	node_id3_tag = xmlNewChild(node_id3, NULL, BAD_CAST"length",ptr_release);
	xmlFree(ptr_release);
      }

      ptr_release = input_id3_get_string(tag, ID3_FRAME_COMMENT);
      if (ptr_release) {
	node_id3_tag = xmlNewChild(node_id3, NULL, BAD_CAST"comment",ptr_release);
	xmlFree(ptr_release);
      }
    }
    id3_file_close(id3file);
  }
  else {
    /* no id3tag */
  }

  return TRUE;
}
/* end of add_id3_tags() */

#endif


/*! returns a pointer to a buffer containing the extension of filename at c or NULL

 A sequence of digits at the end of a filename is no extension
 (version stamps). Problem with man page filenames.

 */
xmlChar *
file_extension(xmlChar *c)
{
  xmlChar *c_tmp, *result, *c_tail;
  int l_result;

  if (c == NULL || *c == (xmlChar)'\0') {
    return NULL;
  }

  for (c_tmp=c, result=NULL;
       (c_tmp = BAD_CAST xmlStrchr(c_tmp,(xmlChar)'.'));
       c_tmp++, result=c_tmp) {

    /* check if the tail consists of numerical chars only */
    for (c_tail = c_tmp+1; isdigit(*c_tail); c_tail++) { /* skip all digits */ }

    if (*c_tail==(xmlChar)'\0') {
      /* case (4) */
      if (result==NULL) {
	c_tmp++;
	l_result = c_tail - c_tmp;
	if (l_result > 0 && l_result < 5) {
	  return xmlStrdup(c_tail);
	}
      }
      else {
	l_result = c_tmp - result;
	if (l_result > 0 && l_result < 5) {
	  return xmlStrndup(result,l_result);
	}
      }
      return NULL;
    }
  }

   /*
    (1) simple filename without extension:

    [a|b|c|\0]
    NULL <- c_tmp
    NULL <- result

    (2) simple filename with extension:

    [a|b|c|.|d|e|f|\0]
             ^ 
             |
           result
	     |
           c_tail
    NULL <- c_tmp

    (3) simple filename with dot char inside and extension:

    [a|b|c|.|d|e|f|.|h|i|j|\0]
                     ^ 
                     |
                   result
		     |
                   c_tail
    NULL <- c_tmp

    (4) filename with extension and numeric version:

    [d|e|f|.|1|2|\0]
             ^   ^
             |   |
           c_tmp |
               c_tail
    NULL <- result

    [a|b|c|.|d|e|f|.|1|2|\0]
             ^     ^     ^
             result|     |
                   c_tmp |
                         c_tail
   */

  if (result != NULL && *result != (xmlChar)'\0') {
    l_result = xmlStrlen(result);
    if (l_result > 0 && l_result < 5) {
      /* case (2) */
      /* case (3) */
      xmlChar *ptr;
      xmlChar *r = xmlStrdup(result);

      for (ptr=r; *ptr != (xmlChar)'\0'; ptr++) {
	if (isalpha(*ptr)) {
	  *ptr = (xmlChar)tolower(*ptr);
	}
      }

      return r;
    }
  }
  /* case (1) */
  return NULL;
}
/* end of file_extension() */


/*! Append the directory listing to the given node_parent

thread safe

verbosity

- 0 -- directories only

- 1 -- simple files (default)

- 2 -- including file system informations

- 3 -- including MIME attribute

- 4 -- additional internal informations (structure, id3 tags, image informations)

?one thread per directory, test of performance

 */
BOOL_T
append_dir(xmlNodePtr node_parent, const xmlChar *name_dir, const int i_depth, const int level_verbose, pcre *re_match, pcre *re_grep)
{
  char buffer_in[BUFFER_LENGTH];
  int err;
  int inlen;
  int l;
  int outlen = BUFFER_LENGTH-1;
  long int long_filesize;
  int int_in;
  int int_out;
  xmlChar buffer_conv[BUFFER_LENGTH];
  xmlChar buffer_out[BUFFER_LENGTH];
  xmlChar buffer_pattern[BUFFER_LENGTH];
  xmlChar *ext;
  xmlChar *ptr_0;
  xmlChar *ptr_index;
  xmlNodePtr node_dir;
  xmlNodePtr node_file;

#ifdef _MSC_VER
  char *ptr_out, *ptr_in;
  size_t nconv;
  struct _finddata_t c_file;
  intptr_t hFile;
  struct _stat stbuf; 
#else
  DIR *dp;
  struct dirent *d;
  struct stat stbuf;
#endif

  if (node_parent==NULL) {
    return FALSE;
  }

  if (name_dir==NULL) {
    /* the working path has to be inititialized with current directory */
    name_dir = getcwdUTF8();
  }
  /*!\todo this is a relative path */

  /* set the according directory separators */
  strncpy((char *)buffer_conv,(const char *)name_dir,BUFFER_LENGTH-1);

  /* ptr_0 = file_normalize(buffer_conv,NULL); */
  ptr_0 = buffer_conv;

  printf_log(3,"%s '%s'", NAME_DIR, ptr_0);
  node_dir = xmlNewChild(node_parent, NULL, NAME_DIR, NULL);
  /* handle possible problems with ptr_0 as ISO8859-1 or UTF-8 
   */

  if (cd.to.env) {
#if 0
    strncpy(buffer_in,(char *)ptr_0,BUFFER_LENGTH - 1);

    int_in = strlen(buffer_in) + 2;
    ptr_in  = buffer_in;
    int_out = BUFFER_LENGTH - 1;
    ptr_out = (char *)buffer_out;

    nconv = iconv(cd.to.env,&ptr_in,&int_in,&ptr_out,&int_out);
    iconv_error(errno,nconv);
#else
   /*\todo why fails the iconv() ???*/
    int_in = strlen((const char*)ptr_0) + 2;
    int_out = BUFFER_LENGTH - 1;
    isolat1ToUTF8((unsigned char *)buffer_out,&int_out,(unsigned char *)ptr_0,&int_in);
#endif
  }
  else {
    /* no conversion necessary */
    strncpy((char *)buffer_out,(const char *)ptr_0,(size_t)BUFFER_LENGTH - 1);
  }
   
  /* basename as relative path, because there is a DIR parent OR absolute path, because its the first level*/
  if  (IS_NODE_DIR(node_parent)) {
    xmlChar *ptr_basename = file_basename(buffer_out);
    if (ptr_basename != NULL && xmlStrlen(ptr_basename) > 0) {
      xmlSetProp(node_dir, BAD_CAST "name", ptr_basename);
      xmlFree(ptr_basename);
    }
  }
  else {
    xmlSetProp(node_dir, BAD_CAST "name", buffer_out);
  }

#if 1
  /* try to open an accelaration index file */
  ptr_index = file_normalize(ptr_0,".index.pie",NULL);
  if (ptr_index) {
    if (type_is_readableUTF8(ptr_index,FALSE,&long_filesize)==TRUE && long_filesize > 30) {
      xmlDocPtr doc_index;
      printf_log(1,"Read index from '%s'", ptr_index);
      doc_index = xmlParseFile((char *)ptr_index);
      xmlFree(ptr_index);
      if (doc_index) {
        xmlNodePtr node_dir_index;
        xmlNodePtr node_dir_index_child;
	node_dir_index = xmlDocGetRootElement(doc_index);
	if (IS_NODE_PIE(node_dir_index)
	    && (node_dir_index_child = child_element(node_dir_index,BAD_CAST "dir"))!=NULL) {
          xmlNodePtr node_release;
	  node_release = xmlReplaceNode(node_dir,node_dir_index_child);
  	  xmlSetProp(node_dir_index_child, BAD_CAST "name", IS_NODE_DIR(node_parent) ? file_basename(ptr_0) : ptr_0);
	  xmlSetProp(node_dir_index_child, BAD_CAST "source", BAD_CAST "index");
	  xmlFreeNode(node_release);
	  xmlFreeDoc(doc_index);
	  return TRUE;
	}
	else {
	  xmlFreeDoc(doc_index);
	  return FALSE;
	}
      }
    }
    xmlFree(ptr_index);
    /*!\todo confusing if there are redundant index files in substructures */
  }
#endif

#ifdef _MSC_VER
  /* 
     Windows Code for MS VC++
  */
  err = _stat((const char*)ptr_0, &stbuf);
  if (err == -1) {
    printf_log(1,"Cant access");
    xmlSetProp(node_dir, BAD_CAST "error", xmlStrdup(BAD_CAST "perm"));
    return TRUE;
  }

  if (level_verbose > 1) {
    xmlStrPrintf(buffer_out,BUFFER_LENGTH,BAD_CAST "%i",stbuf.st_mtime);
    xmlSetProp(node_dir, BAD_CAST "mtime", buffer_out);

      xmlSetProp(node_dir, BAD_CAST "read",
		 BAD_CAST ((stbuf.st_mode & _S_IREAD) ? "yes" : "no"));
      xmlSetProp(node_dir, BAD_CAST "write",
		 BAD_CAST ((stbuf.st_mode & _S_IWRITE) ? "yes" : "no"));
      /*
      xmlSetProp(node_dir, BAD_CAST "execute",
		 BAD_CAST ((stbuf.st_mode & _S_IXUSR) ? "yes" : "no"));
      */ 
  }

  if (isalpha(ptr_0[0])
      && ptr_0[1] == (xmlChar)':'
      && (ptr_0[2] == (xmlChar)'/' || ptr_0[2] == (xmlChar)'\\')
      && ptr_0[3] == (xmlChar)'\0') {
    /* cut trailing separator */
    ptr_0[2] = (xmlChar)'\0';
  }
  xmlStrPrintf(buffer_pattern,sizeof(buffer_pattern),BAD_CAST"%s\\*.*", ptr_0);
  /*  Find first file in current directory */
  if( (hFile = _findfirst( (const char*)buffer_pattern, &c_file )) == -1L ) {
    printf_log(1,"Cant open '%s'", ptr_0);
    if (errno==EACCES) {
      xmlSetProp(node_dir, BAD_CAST "error", BAD_CAST "noread");
    }
    else {
      xmlSetProp(node_dir, BAD_CAST "error", BAD_CAST "perm");
    }
    return FALSE;
  }

  if (i_depth==0) {
    /* recursion depth reached */
    _findclose( hFile );
    return TRUE;
  }

  do {
    if (strcmp(c_file.name,".")==0 || strcmp(c_file.name,"..")==0) {
      /* ignore */
      continue;
    }

    xmlStrPrintf(buffer_pattern,sizeof(buffer_pattern),BAD_CAST"%s\\%s", ptr_0,c_file.name);
    if (c_file.attrib & _A_SUBDIR) {
      /*!\todo create one new thread per subdir */
      append_dir(node_dir,buffer_pattern,i_depth-1, level_verbose, re_match, re_grep);
    }
    else if (level_verbose > 0) {
      /* file matching */
      if (re_match) {
	int rc;
	int ovector[30];
	rc = pcre_exec(re_match,NULL,c_file.name,strlen(c_file.name),0,0,ovector,30);
	if (rc < 0) {
	  printf_log(4,"FILE ignore '%s'", c_file.name);
	  continue;
	}
      }
      printf_log(4,"FILE '%s'", c_file.name);
      node_file = xmlNewChild(node_dir, NULL, NAME_FILE, NULL);

      if (cd.from.env) {
	strncpy(buffer_in,c_file.name,BUFFER_LENGTH - 1);
	int_in = strlen(buffer_in) + 1;
	ptr_in  = buffer_in;
	int_out = BUFFER_LENGTH - 1;
	ptr_out = (char *)buffer_out;

	nconv = iconv(cd.from.env,&ptr_in,(size_t *)&int_in,&ptr_out,(size_t *)&int_out);
	iconv_error(errno,nconv);
	xmlSetProp(node_file, BAD_CAST "name", BAD_CAST buffer_out);
      }
      else {
	/* no conversion necessary */
	xmlSetProp(node_file, BAD_CAST "name", BAD_CAST c_file.name);
      }

      /* attribute EXT from attribute NAME */
      if ((ext = file_extension(attribute_str(node_file,BAD_CAST "name")))) {
	/*!\todo make uppercase to ext */
	xmlSetProp(node_file, BAD_CAST "ext", ext);
	xmlFree(ext);
      }

      err = _stat((const char*)buffer_pattern, &stbuf); 
      if (err == -1) {
	printf_log(1,"Cant access");
	xmlSetProp(node_file, BAD_CAST "error", xmlStrdup(BAD_CAST "perm"));
	continue;
      }

      if (level_verbose > 1) {
	if (stbuf.st_size > 0) {
	  xmlStrPrintf(buffer_out,BUFFER_LENGTH,BAD_CAST "%i",stbuf.st_size);
	  xmlSetProp(node_file, BAD_CAST "size", buffer_out);
	}
	xmlStrPrintf(buffer_out,BUFFER_LENGTH,BAD_CAST "%i",stbuf.st_mtime);
	xmlSetProp(node_file, BAD_CAST "mtime", buffer_out);

      xmlSetProp(node_file, BAD_CAST "read",
		 BAD_CAST ((stbuf.st_mode & _S_IREAD) ? "yes" : "no"));
      xmlSetProp(node_file, BAD_CAST "write",
		 BAD_CAST ((stbuf.st_mode & _S_IWRITE) ? "yes" : "no"));
#if 0
      xmlSetProp(node_file, BAD_CAST "execute",
		 BAD_CAST ((stbuf.st_mode & S_IXUSR) ? "yes" : "no"));

	xmlStrPrintf(buffer_out,BUFFER_LENGTH,"%i",stbuf.st_atime);
	xmlSetProp(node_file, BAD_CAST "atime", buffer_out);
#endif
      }

      add_content_type(node_file,buffer_pattern,level_verbose);

      if (re_grep) {
	xmlNodePtr node_result;
	int l;
	int mode = 1;
	int mode_type = 1;
	FILE *in;
	xmlChar buffer_in[BUFFER_LENGTH];

	/* comfortable string matching using regexps */
	printf_log(2,"Grep file '%s'", c_file.name);
	if ((in = fopen((const char*)buffer_pattern,"r"))) {

	  for (l=1; fgets((char *)buffer_in,BUFFER_LENGTH-1,in);l++) {
	    int rc;
	    int ovector[30];

	    chomp(buffer_in);
	    if (buffer_in[0] == (xmlChar)'\0') {
	      /* empty string */
	      continue;
	    }
	    /*\todo encoding before matching or when appending to DOM */
	    rc = pcre_exec(
			   re_grep,             /* result of pcre_compile() */
			   NULL,           /* we didn't study the pattern */
			   (const char*)buffer_in,  /* the subject string */
			   strlen((const char*)buffer_in),             /* the length of the subject string */
			   0,              /* start at offset 0 in the subject */
			   0,              /* default options */
			   ovector,        /* vector of integers for substring information */
			   30);            /* number of elements (NOT size in bytes) */

	    if (rc > -1) {
	      node_result = xmlNewChild(node_file, NULL, BAD_CAST "result", BAD_CAST buffer_in);
	      xmlStrPrintf(buffer_in,sizeof(buffer_in),BAD_CAST "%i",l);
	      xmlSetProp(node_result, BAD_CAST "line", BAD_CAST buffer_in);
	    }
	    else {
	      /* pcre_free(rc); */
	    }
	  }
	  fclose(in);
	  xmlStrPrintf(buffer_in,sizeof(buffer_in),BAD_CAST "%i",l);
	  xmlSetProp(node_file, BAD_CAST "lines", BAD_CAST buffer_in);
	}
	else {
	  printf_log(1,"Cant open file '%s'", c_file.name);
	}
      }
    }
  } while( _findnext( hFile, &c_file ) == 0 );
  _findclose( hFile );
#else
  /*
    Code for GCC
  */
  err = stat((const char *)ptr_0,&stbuf); 
  if (err == -1) {
    printf_log(1,"Cant access");
    xmlSetProp(node_dir, BAD_CAST "error", BAD_CAST "perm");
    return TRUE;
  }

  if (level_verbose > 1) {
    xmlStrPrintf(buffer_out,BUFFER_LENGTH,BAD_CAST "%i",(int)stbuf.st_mtime);
    xmlSetProp(node_dir, BAD_CAST "mtime", buffer_out);

    /*!\bug read permissions for current user/group, not general */

    xmlSetProp(node_dir, BAD_CAST "read",
	       BAD_CAST ((stbuf.st_mode & S_IRUSR) ? "yes" : "no"));
    xmlSetProp(node_dir, BAD_CAST "write",
	       BAD_CAST ((stbuf.st_mode & S_IWUSR) ? "yes" : "no"));
    xmlSetProp(node_dir, BAD_CAST "execute",
	       BAD_CAST ((stbuf.st_mode & S_IXUSR) ? "yes" : "no"));
  }
  if ((dp = opendir((const char *)ptr_0))==NULL) {
    printf_log(1,"Cant open '%s'", ptr_0);
    if (errno==EACCES) {
      xmlSetProp(node_dir, BAD_CAST "error", BAD_CAST "noread");
    }
    else {
      xmlSetProp(node_dir, BAD_CAST "error", BAD_CAST "perm");
    }
    return FALSE;
  }

  if (i_depth==0) {
    /* recursion depth reached */
    closedir(dp);
    return TRUE;
  }

  while ((d = readdir(dp))) {
    if (d->d_ino == 0 || strcmp(d->d_name,".")==0 || strcmp(d->d_name,"..")==0) {
      /* ignore */
      continue;
    }

    xmlStrPrintf(buffer_pattern,sizeof(buffer_pattern),BAD_CAST "%s/%s", ptr_0,d->d_name);
    /*!\todo detect symbolic links etc */
    if (d->d_type == DT_DIR) {
      append_dir(node_dir,buffer_pattern,i_depth-1, level_verbose, re_match, re_grep);
    }
    else if (level_verbose > 0) {
      /* file matching */
      if (re_match) {
	int rc;
	int ovector[30];
	rc = pcre_exec(re_match,NULL,d->d_name,(int)strlen(d->d_name),0,0,ovector,30);
	if (rc < 0) {
	  printf_log(4,"FILE ignore '%s'", d->d_name);
	  continue;
	}
      }
      printf_log(4,"FILE '%s'", d->d_name);
      node_file = xmlNewChild(node_dir, NULL, NAME_FILE, NULL);
      /* handle possible problems with ptr_0 as ISO8859-1 or UTF-8 
       */
      if (xmlCheckUTF8(BAD_CAST d->d_name)) {
	xmlSetProp(node_file, BAD_CAST "name", BAD_CAST d->d_name);
      }
      else {
	inlen = (int)strlen(d->d_name);
	l = isolat1ToUTF8(buffer_out,&outlen,BAD_CAST d->d_name,&inlen);
	if (l < 1) {
	  printf_log(1,"Problem with name");
	  xmlSetProp(node_file, BAD_CAST "error", BAD_CAST "name");
	  continue;
	}
	buffer_out[l] = (xmlChar)'\0';
	xmlSetProp(node_file, BAD_CAST "name", buffer_out);
      }

      /* attribute EXT from attribute NAME */
      if ((ext = file_extension(attribute_str(node_file,BAD_CAST "name")))) {
	/*!\todo make uppercase to ext */
	xmlSetProp(node_file, BAD_CAST "ext", ext);
	xmlFree(ext);
      }

      err = stat((const char *)buffer_pattern,&stbuf); 
      if (err == -1) {
	printf_log(1,"Cant access");
	xmlSetProp(node_file, BAD_CAST "error", BAD_CAST "perm");
	continue;
      }

      if (level_verbose > 1) {
	if (stbuf.st_size > 0) {
	  xmlStrPrintf(buffer_out,BUFFER_LENGTH,BAD_CAST "%i",stbuf.st_size);
	  xmlSetProp(node_file, BAD_CAST "size", buffer_out);
	}
	xmlStrPrintf(buffer_out,BUFFER_LENGTH,BAD_CAST "%i",(int)stbuf.st_mtime);
	xmlSetProp(node_file, BAD_CAST "mtime", buffer_out);

	xmlSetProp(node_file, BAD_CAST "read",
		   BAD_CAST ((stbuf.st_mode & S_IRUSR) ? "yes" : "no"));
	xmlSetProp(node_file, BAD_CAST "write",
		   BAD_CAST ((stbuf.st_mode & S_IWUSR) ? "yes" : "no"));
	xmlSetProp(node_file, BAD_CAST "execute",
		   BAD_CAST ((stbuf.st_mode & S_IXUSR) ? "yes" : "no"));

#if 0
	xmlStrPrintf(buffer_out,BUFFER_LENGTH,BAD_CAST "%i",stbuf.st_atime);
	xmlSetProp(node_file, BAD_CAST "atime", buffer_out);
	xmlStrPrintf(buffer_out,BUFFER_LENGTH,BAD_CAST "%u",d->d_type);
	xmlSetProp(node_file, BAD_CAST "type", buffer_out);
#endif
      }

      add_content_type(node_file,buffer_pattern,level_verbose);

      if (re_grep) {
	xmlNodePtr node_result;
	int j;
	FILE *in;
	xmlChar buffer_line[BUFFER_LENGTH];

	/* comfortable string matching using regexps */
	printf_log(2,"Grep file '%s'", d->d_name);
	if ((in = fopen((const char *)buffer_pattern,"r"))) {

	  for (j=1;fgets((char *)buffer_line,BUFFER_LENGTH-1,in);j++) {
	    int rc;
	    int ovector[30];

	    chomp(buffer_line);
	    if (buffer_line[0] == (xmlChar)'\0') {
	      /* empty string */
	      continue;
	    }
	    /*\todo encoding before matching or when appending to DOM */
	    rc = pcre_exec(
			   re_grep,             /* result of pcre_compile() */
			   NULL,           /* we didn't study the pattern */
			   (const char*)buffer_line,  /* the subject string */
			   strlen((const char*)buffer_line),             /* the length of the subject string */
			   0,              /* start at offset 0 in the subject */
			   0,              /* default options */
			   ovector,        /* vector of integers for substring information */
			   30);            /* number of elements (NOT size in bytes) */

	    if (rc > -1) {
	      node_result = xmlNewChild(node_file, NULL, BAD_CAST "result", BAD_CAST buffer_line);
	      xmlStrPrintf(buffer_line,sizeof(buffer_line),BAD_CAST"%i",j);
	      xmlSetProp(node_result, BAD_CAST "line", BAD_CAST buffer_line);
	    }
	    else {
	      /*\todo ?? pcre_free(rc); */
	    }
	  }
	  fclose(in);
	  xmlStrPrintf(buffer_out,sizeof(buffer_in),BAD_CAST "%i",j);
	  xmlSetProp(node_file, BAD_CAST "lines", buffer_out);
	}
	else {
	  printf_log(1,"Cant open file '%s'", d->d_name);
	}
      }
    }
  }
  closedir(dp);
#endif
  return TRUE;
}
/* end of append_dir() */


/*! process the DIR child instructions of node_make_pie

\todo additional argument depth_to_go for recursion limit
 */
xmlDocPtr
process_dir(xmlNodePtr node_make_dir)
{
  xmlChar *attr_grep;
  xmlChar *attr_depth;
  xmlChar *attr_match;
  xmlChar *ptr_path;
  int n_depth;
  int level_verbose;

  xmlChar buffer[BUFFER_LENGTH];
  xmlDocPtr result;
  xmlNodePtr n_i;
  xmlNodePtr node_pie;
  xmlNodePtr node_meta;
  pcre *re_grep  = NULL;
  pcre *re_match = NULL;
  int erroffset;
  int opt_grep_pcre = 0;
  int opt_match_pcre = 0;

  const char *error;

  assert(node_make_dir);

  /*!\todo set depth attribute per single dir element, instead of global */
  if ((attr_depth = attribute_str(node_make_dir,BAD_CAST "depth")) 
      && ((n_depth = atoi((char *)attr_depth)) > 0)) {
  }
  else {
    n_depth = 9999;
  }
  printf_log(2,"Set DIR depth to '%i'", n_depth);

  if ((attr_depth = attribute_str(node_make_dir,BAD_CAST "verbosity")) 
      && ((level_verbose = atoi((char *)attr_depth)) > 0)) {
  }
  else {
    level_verbose = 1;
  }
  printf_log(2,"Set DIR verbosity to '%i'", level_verbose);

  if ((attr_grep = attribute_str(node_make_dir,BAD_CAST "grep"))) {
    printf_log(2,"Use grep '%s' in %s with depth '%i'", attr_grep, NAME_DIR, n_depth);
    re_grep = pcre_compile((const char*)attr_grep,opt_grep_pcre,&error,&erroffset,NULL);
  }
  else if ((attr_grep = attribute_str(node_make_dir,BAD_CAST "igrep"))) {
    printf_log(2,"Use caseless grep '%s' in %s with depth '%i'", attr_grep, NAME_DIR, n_depth);
    opt_grep_pcre |= PCRE_CASELESS;
    re_grep = pcre_compile((const char*)attr_grep,opt_grep_pcre,&error,&erroffset,NULL);
  }

  if ((attr_match = attribute_str(node_make_dir,BAD_CAST "match"))) {
    printf_log(2,"Use file match '%s' in %s with depth '%i'", attr_match, NAME_DIR, n_depth);
#ifdef _MSC_VER
    opt_match_pcre |= PCRE_CASELESS;
#endif
    re_match = pcre_compile((const char*)attr_match,opt_match_pcre,&error,&erroffset,NULL);
  }
  else if ((attr_match = attribute_str(node_make_dir,BAD_CAST "imatch"))) {
    printf_log(2,"Use caseless file match '%s' in %s with depth '%i'", attr_match, NAME_DIR, n_depth);
    opt_match_pcre |= PCRE_CASELESS;
    re_match = pcre_compile((const char*)attr_match,opt_match_pcre,&error,&erroffset,NULL);
  }

  /*!\todo use exclude="(~|)" (or: mach="" type="exclude") */

  result = xmlNewDoc(BAD_CAST "1.0");
  node_pie = xmlNewDocNode(result, NULL, NAME_PIE, NULL); 
  xmlSetProp(node_pie, BAD_CAST "class", BAD_CAST "directory");
  xmlDocSetRootElement(result, node_pie);
  /* this is default namespace */
  /*!\todo xmlNewNs(node_pie, "http://", NULL); */ /* no prefix string for this ns */

#if HAVE_LIBMAGICK
  /* these are namespace declarations only */
  xmlSetProp(node_pie, BAD_CAST "xmlns:exif", BAD_CAST "http://www.tenbusch.info/cxproc/exif");
  xmlSetProp(node_pie, BAD_CAST "xmlns:jpeg", BAD_CAST "http://www.tenbusch.info/cxproc/jpeg");
  xmlSetProp(node_pie, BAD_CAST "xmlns:tiff", BAD_CAST "http://www.tenbusch.info/cxproc/tiff");
  /*!\todo add other namespaces TIFF etc. */
#endif

  node_meta = xmlNewChild(node_pie, NULL, BAD_CAST "meta", NULL);
  xmlAddChild(node_meta,xmlCopyNode(node_make_dir,1));
  /* Get the current time. */
  xmlStrPrintf(buffer,BUFFER_LENGTH,BAD_CAST "%i", (int)time(NULL));
  xmlSetProp(node_meta, BAD_CAST "mtime", BAD_CAST buffer);

  if ((ptr_path = attribute_str(node_make_dir,BAD_CAST "pathxsl"))) {
#if 1
    xmlChar *ptr_path_sub;
    xmlChar *ptr_0;
    xmlChar *ptr_sep;
    xmlChar *ptr_dir;

    /* multiple dir names? */
    ptr_path_sub = xmlStrdup(ptr_path);
    printf_log(2,"Processing argument PATHXSL '%s'", ptr_path_sub);
    for (ptr_0=ptr_path_sub; ptr_0; ptr_0=ptr_sep+1) {
      /*!\todo support for URLs? reading of directories in UNC notation "file:///server/share/dir/" */
      /*!\todo dir name="/home/abc" as non recursive and name="/home/abc//" as recursive (like kpathsea) */
      ptr_sep = (xmlChar *)xmlStrchr((const xmlChar *)ptr_0,(xmlChar)';');
      if (ptr_sep) {
	xmlChar *ptr_release = xmlStrndup(ptr_0,ptr_sep - ptr_0);
	ptr_dir = file_normalize(ptr_release,NULL);
	append_dir(node_pie,ptr_dir,n_depth, level_verbose, re_match, re_grep);
	xmlFree(ptr_release);
	xmlFree(ptr_dir);
      }
      else {
	/* no separator char found */
	ptr_dir = file_normalize(ptr_0,NULL);
	append_dir(node_pie,ptr_dir,n_depth, level_verbose, re_match, re_grep);
	xmlFree(ptr_dir);
	break;
      }
    }
    xmlFree(ptr_path_sub);
#else
    xmlChar *ptr_dir = file_normalize(ptr_path,NULL);
    printf_log(2,"Processing argument PATHXSL '%s'", ptr_dir);
    append_dir(node_pie,ptr_dir,n_depth, level_verbose, NULL,NULL);
    xmlFree(ptr_dir);
#endif
  }
  else if (node_make_dir->children) {
    /* handle multiple DIR elements in one XML */
    for (n_i=node_make_dir->children;
	 n_i;
	 n_i = n_i->next) {
      if (IS_NODE_DIR(n_i) && (ptr_path = attribute_str(n_i,BAD_CAST "name"))) {
	printf_log(2,"Processing %s '%s' with depth '%i'", NAME_DIR, ptr_path, n_depth);
	append_dir(node_pie,ptr_path,n_depth,level_verbose,re_match,re_grep);
      }
    }
  }
  else if ((ptr_path = attribute_str(node_make_dir,BAD_CAST "name"))) {
    printf_log(2,"Processing %s '%s' with depth '%i'", NAME_DIR, ptr_path, n_depth);
    append_dir(node_pie,ptr_path,n_depth,level_verbose,re_match,re_grep);
  }
  else {
    printf_log(2,"Processing current dir with depth '%i'", n_depth);
    append_dir(node_pie,NULL,n_depth, level_verbose,re_match,re_grep);
  }

  if (re_match)
    pcre_free((void *)re_match);
  if (re_grep)
    pcre_free((void *)re_grep);

  return result;
}


/*! returns the whole filename with path of the given basename in search DOM

\todo use attribute case="yes|no"
*/
xmlChar *
file_search(xmlNodePtr node_search, const xmlChar *str_filename)
{
  xmlChar buffer_path[BUFFER_LENGTH];
  xmlChar *ptr_attr;
  xmlChar *ptr_begin;
  xmlChar *ptr_end;
  xmlChar *ptr_release;
  xmlChar *str_name;
  xmlNodePtr p;
  xmlNodePtr node_search_new = NULL;
  BOOL_T flag_recursion;

  if (node_search == NULL
      && (node_search = xmlDocGetRootElement(doc_search)) == NULL) {
    /* neither search node nor global search DOM found */
    return NULL;
  }

  assert(str_filename!=NULL);
  assert(*str_filename != (xmlChar)'\0');

  printf_log(3,"Search for file '%s'",str_filename);

#if 1
  if ((str_name = get_dom_file(str_filename))) {
    printf_log(2,"Found cached filename '%s'",str_name);
    return xmlStrdup(str_name);
  }
#endif

  for (flag_recursion=TRUE, ptr_begin=(xmlChar *)str_filename, ptr_end=(xmlChar *)str_filename, str_name=NULL; ; ptr_end++) {

    if (*ptr_end == (xmlChar)'/' && (ptr_end - ptr_begin) > 0) {
      /* separator found */

      str_name = xmlStrndup(ptr_begin, ptr_end - ptr_begin);

      if ((node_search_new = filenode_search(node_search,str_name,FALSE))) {
	/* node_search itself is the result */
	node_search = node_search_new;
	ptr_begin = ptr_end + 1;
	flag_recursion = FALSE;
	break;
      }
      else {
	/* search for node over all childs */
	for (node_search = (node_search ? node_search->children : NULL); node_search; node_search=node_search->next) {
	  if ((node_search_new = filenode_search(node_search,str_name,flag_recursion))) {
	    /* node with str_name found */
	    node_search = node_search_new;
	    ptr_begin = ptr_end+1;
	    flag_recursion = FALSE;
	    break;
	  }
	}
      }
      /*!\todo xmlFree(str_name) in recursive call */
    }
    else if (*ptr_end == (xmlChar)'\0') {
      if ((ptr_end - ptr_begin) > 0) {
	str_name = xmlStrndup(ptr_begin, ptr_end - ptr_begin);
	node_search_new = filenode_search(node_search,str_name,TRUE);
      }
      break;
    }
  }

  if (str_name) xmlFree(str_name);

  if (node_search_new) {
    ptr_attr = attribute_str(node_search_new,BAD_CAST "name");
    xmlStrPrintf(buffer_path,sizeof(buffer_path),BAD_CAST"%s",ptr_attr);
    /* build path from parent nodes */
    for (p=node_search_new->parent; p; p=p->parent) {
      if ((IS_NODE_FILE(p) || IS_NODE_DIR(p)) && (ptr_attr = attribute_str(p,BAD_CAST "name"))!=NULL) {
	ptr_release = xmlStrdup(buffer_path);
	xmlStrPrintf(buffer_path,sizeof(buffer_path),BAD_CAST"%s/%s",ptr_attr,ptr_release);
	xmlFree(ptr_release); /* release old string */
      }
    }

    printf_log(2,"Found path '%s'",buffer_path);
    return file_normalize(buffer_path,NULL);
  }
  return NULL;
}


/*! returns the pointer to the first node with same filename
*/
xmlNodePtr
filenode_search(xmlNodePtr node_search, const xmlChar *f, BOOL_T flag_recursive)
{
  xmlChar *attr_name;
  xmlNodePtr c;
  xmlNodePtr node_result;

  if (doc_search!=NULL
      && (node_search!=NULL || (node_search = xmlDocGetRootElement(doc_search))!=NULL)) {
    if (node_search->type == XML_ELEMENT_NODE) {
      if ((attr_name = attribute_str(node_search,BAD_CAST "name")) != NULL
	  && xmlStrcasecmp(attr_name,f)==0) {
	return node_search;
      }
      else if (flag_recursive==TRUE) {
	for (c=node_search->children; c; c=c->next) {
#if 1
	  if ((attr_name = attribute_str(c,BAD_CAST "name")) != NULL
	      && (attr_name[0] == (xmlChar)'.' && attr_name[1] != (xmlChar)'.')) {
	    /* ignore hidden entries */
	  }
	  else if ((node_result = filenode_search(c,f,flag_recursive))) { /* recursion */
	    return node_result;
	  }
#else
	  if ((node_result = filenode_search(c,f,flag_recursive))) { /* recursion */
	    return node_result;
	  }
#endif			
	}
      }
      else {
	/* ignore child element */
      }
    }
  }
  return NULL;
}


/*! sets the mime attribute

  \todo recognize more MIME types using libmime (like >>file<< command)
*/
BOOL_T
add_content_type(xmlNodePtr node_file, xmlChar *file_name, int level_verbose)
{
  BOOL_T flag_xml = FALSE;

  /*
    set a MIME conformant attribute
  */
  xmlChar *attr_ext = attribute_str(node_file,BAD_CAST"ext");

  if (attr_ext != NULL && level_verbose > 2) {
    if (xmlStrEqual(attr_ext,BAD_CAST"txt")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "text/plain");

      if (level_verbose > 3) {
	FILE *in;
	xmlNodePtr node_tmp;

	printf_log(1,"Read pie file '%s'", file_name);

	node_tmp = xmlNewNode(NULL, NAME_PLAIN);
	xmlSetProp(node_tmp, BAD_CAST "name", BAD_CAST file_name);

	if ((in = fopen((char *)file_name,"r"))) {
	  /* FILE is a txt2x file */
	  xmlNodePtr node_pie;

	  node_pie = xmlNewNode(NULL, NAME_PIE);
	  append_input(node_pie,node_tmp,in);
	  fclose(in);
	  xmlAddChildList(node_file,node_pie);
	}
	else {
	  printf_log(1,"Cant open pie file '%s'", file_name);
	  xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "error/fopen");
	}
	xmlFreeNodeList(node_tmp);
      }
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"pie")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "application/pie+xml");
      if (level_verbose > 3) {
	flag_xml = TRUE;      	
      }
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"cxp")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "application/cxp+xml");
      flag_xml = TRUE;
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"xspf")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "application/xspf+xml");
      flag_xml = TRUE;
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"xsl")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "text/xsl");
      flag_xml = TRUE;
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"xml")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "text/xml");
      flag_xml = TRUE;
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"mm")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "application/mm+xml");
      flag_xml = TRUE;
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"css")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "text/css");
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"html") || xmlStrEqual(attr_ext,BAD_CAST"htm")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "text/html");
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"mp3")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "audio/mp3");
#if HAVE_LIBID3TAG
      if (level_verbose > 3) {
	add_id3_tags(node_file, file_name);
      }
#endif
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"ogg")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "audio/ogg");
      if (level_verbose > 3) {
	/*!\todo Tags in OGG-Vorbis format */
      }
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"jpg") || xmlStrEqual(attr_ext,BAD_CAST"jpeg")) {
#if HAVE_LIBMAGICK
      /* get image information details via ImageMagick */
      if (level_verbose > 3) {
	add_image_info(node_file, file_name);
      }
      else {
#endif
	/*!\todo use minimal filesize to set attribute to error */
	xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "image/jpeg");
#if HAVE_LIBMAGICK
      }
#endif
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"gif")) {
#if HAVE_LIBMAGICK
      if (level_verbose > 3) {
	add_image_info(node_file, file_name);
      }
      else {
#endif
	xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "image/gif");
#if HAVE_LIBMAGICK
      }
#endif
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"tif")) {
#if HAVE_LIBMAGICK
      if (level_verbose > 3) {
	add_image_info(node_file, file_name);
      }
      else {
#endif
	xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "image/tiff");
#if HAVE_LIBMAGICK
      }
#endif
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"png")) {
#if HAVE_LIBMAGICK
      if (level_verbose > 3) {
	add_image_info(node_file, file_name);
      }
      else {
#endif
	xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "image/png");
#if HAVE_LIBMAGICK
      }
#endif
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"prt") || xmlStrEqual(attr_ext,BAD_CAST"asm") || xmlStrEqual(attr_ext,BAD_CAST"drw")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "application/proengineer");
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"bat") || xmlStrEqual(attr_ext,BAD_CAST"cmd")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "application/command");
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"dot")
	     || xmlStrEqual(attr_ext,BAD_CAST"log")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "text/plain");
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"dll")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "binary/dll");
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"exe")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "binary/exe");
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"o")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "binary/object");
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"obj")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "binary/object");
    }
    else if (xmlStrEqual(attr_ext,BAD_CAST"c")) {
      xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "text/c");
    }
    else {
      /* no mime informations */
      return FALSE;
    }

    if (flag_xml && level_verbose > 3) {
      xmlDocPtr doc_result = NULL;
      xmlNodePtr node_root_result;

      doc_result = xmlParseFile((const char *)file_name);
      if (doc_result) {
	if ((node_root_result = xmlDocGetRootElement(doc_result)) != NULL
	    &&
	    (IS_NODE_PIE(node_root_result))) {
	  xmlNodePtr node_insert;
	  /* xmlUnlinkNodeList(node_root_result); */
	  node_insert = xmlCopyNodeList(node_root_result);
	  xmlAddChildList(node_file,node_insert);
	  /*!\todo xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "plain/root+xml"); */
	}
	else {
	}
	xmlFreeDoc(doc_result);
      }
      else {
	/* handle parser errors */
	xmlSetProp(node_file, BAD_CAST "type", BAD_CAST "error/parse");
      }
    }
  }
  return TRUE;
}


/*! returns the whole directory path of the given filename
*/
xmlChar *
file_basedir(const xmlChar *str_filename)
{
  xmlChar *ptr_sep;
  xmlChar *ptr;
  xmlChar *ptr_s;
  xmlChar *ptr_path;

  if (str_filename==NULL) {
    return NULL;
  }
    
  assert(*str_filename != (xmlChar)'\0');

  for (ptr=ptr_path=(xmlChar *)str_filename, ptr_sep=NULL;
       ptr!=NULL
	 && *ptr!= (xmlChar)'\0'
	 && ((ptr_s=(xmlChar *)xmlStrchr(ptr,(xmlChar)'/'))!=NULL || (ptr_s=(xmlChar *)xmlStrchr(ptr,(xmlChar)'\\'))!=NULL);
       ptr=ptr_s, ptr_sep=ptr, ptr++) {
    /* search for the last path separator */
  }

  if (ptr_sep != NULL 
      && xmlStrlen(ptr_sep) > 3 
      && (ptr_sep - ptr_path > 1 || *ptr_path != (xmlChar)'.')) {
    /* there is a directory path on argv[1] */
    return xmlStrndup(BAD_CAST ptr_path, ptr_sep - ptr_path);
  }
  return NULL;
}


/*! returns a dynamically allocated string with the basename of the given filename
*/
xmlChar *
file_basename(const xmlChar *str_filename)
{
  xmlChar *ptr_sep;
  xmlChar *ptr;
  xmlChar *ptr_s;

  if (str_filename==NULL) {
    return NULL;
  }
    
  assert(*str_filename != (xmlChar)'\0');

  for (ptr=(xmlChar *)str_filename, ptr_sep=NULL;
       ptr!=NULL
	 && *ptr!= (xmlChar)'\0'
	 && ((ptr_s=(xmlChar *)xmlStrchr((const xmlChar *)ptr,(xmlChar)'/'))!=NULL || (ptr_s=(xmlChar *)xmlStrchr((const xmlChar *)ptr,(xmlChar)'\\'))!=NULL);
       ptr=ptr_s, ptr_sep=ptr, ptr++) {
    /* search for the last path separator */
  }

  if (ptr_sep != NULL && xmlStrlen(ptr_sep) > 1) {
    /* there is a basename */
    return xmlStrdup(ptr_sep+1);
  }
  return NULL;
}


/*! 
 */
BOOL_T
make_set_searchpath(xmlNodePtr n)
{
  xmlChar *ptr_path;

  if (n==NULL) {
    return FALSE;
  }

  ptr_path = attribute_str(n,BAD_CAST "pathxsl");
  if (ptr_path==NULL) {
    return FALSE;
  }

  if (doc_search) {
    xmlNodePtr node_root;
    xmlNodePtr node_dir;
    xmlChar *ptr_name;
	
    if ((node_root = xmlDocGetRootElement(doc_search)) != NULL 
	&& (node_dir = child_element(node_root,BAD_CAST"dir")) != NULL 
	&& (ptr_name = attribute_str(node_dir,BAD_CAST "name")) != NULL
	&& xmlStrcmp(ptr_path,ptr_name)==0) {
      printf_log(2,"Reuse former search DOM '%s'", ptr_name);
      return TRUE;
    }

    printf_log(2,"Release former search DOM");
    xmlFreeDoc(doc_search);
  }
  printf_log(1,"Use DOM from '%s' for searching",ptr_path);
  doc_search = process_dir(n);

  return (doc_search) ? TRUE : FALSE;
}
/* end of make_set_searchpath() */


/*! normalize a given NULL-terminated argument list of string pointers to path pieces (similar to Tcl: [file normalize ...])

  \param NULL is the "end of arguments marker"
  \returns a pointer to a new allocated UTF-8 string
*/
xmlChar *
file_normalize(const xmlChar *ptr_first, ...)
{
    va_list args;    
    xmlChar *str_result;
    xmlChar *ptr_argv;
    xmlChar *ptr_a, *ptr_b;
    xmlChar *ptr_release;
    xmlChar *ptr_prefix;
    BOOL_T flag_url;

    if(ptr_first==NULL) {
	return NULL;
    }

    va_start(args,ptr_first);

    if (xmlStrcmp(ptr_first,BAD_CAST"-")==0) {	/* stdin */
	return xmlStrdup(ptr_first);
    }

    if (xmlStrcasestr(ptr_first,BAD_CAST"http://")==ptr_first || xmlStrcasestr(ptr_first,BAD_CAST"ftp://")==ptr_first) { /* web access */
	flag_url = TRUE;
    }
    else {
	flag_url = FALSE;
    }
    if (xmlStrcasestr(ptr_first,BAD_CAST"file://")==ptr_first) {
	if (xmlStrlen(ptr_first) > 7) {
	    ptr_prefix = ptr_first + 7;
	}
	else {
	    ptr_prefix = va_arg(args,xmlChar *);
	}
    }
    else {
	ptr_prefix = ptr_first;
    }

    if (ptr_prefix[0]== (xmlChar)'/' ||  ptr_prefix[0]== (xmlChar)'\\') {
	/* absolute path */
#ifdef _MSC_VER
	str_result = getcwdUTF8();
	ptr_release = str_result;
	str_result = xmlStrndup(str_result,3);
	xmlFree(ptr_release);
	ptr_release = str_result;
	str_result = xmlStrncatNew(str_result,ptr_prefix,xmlStrlen(ptr_prefix));
	xmlFree(ptr_release);
#else
	str_result = xmlStrdup(ptr_prefix);
#endif
    }
    else if (isalpha((int)ptr_prefix[0])!=0 && ptr_prefix[1]== (xmlChar)':' && (ptr_prefix[2]== (xmlChar)'/' || ptr_prefix[2]== (xmlChar)'\\')) {
	/* DOS drive path */
#ifdef _MSC_VER
	str_result = xmlStrdup(ptr_prefix);
#else
	str_result = xmlStrdup(ptr_prefix + 2);
#endif
    }
    else if (flag_url==TRUE) {
	/* prefix complete */
	str_result = xmlStrdup(ptr_prefix);
    }
    else {
	/* relative path */
	str_result = getcwdUTF8();
	ptr_release = str_result;
	str_result = xmlStrncatNew(str_result,BAD_CAST"/",2);
	xmlFree(ptr_release);
	ptr_release = str_result;
	str_result = xmlStrncatNew(str_result,ptr_prefix,xmlStrlen(ptr_prefix));
	xmlFree(ptr_release);
    }

    while ((ptr_argv = va_arg(args,xmlChar *))) {
	ptr_release = str_result;
	str_result = xmlStrncatNew(str_result,BAD_CAST"/",2);
	xmlFree(ptr_release);
	ptr_release = str_result;
	str_result = xmlStrncatNew(str_result,ptr_argv,xmlStrlen(ptr_argv));
	xmlFree(ptr_release);
	/*!\todo use xmlFree() */
    }
    va_end(args);

    printf_log(4,"Pre  normalize '%s'",str_result);

    /*! set path separators according to OS */
    for ( ptr_a=ptr_b=((flag_url==TRUE) ? str_result+6 : str_result); ptr_a!=NULL && *ptr_a!=(xmlChar)'\0'; ) {
	if (*ptr_a== (xmlChar)';') {
	    break;
	}
	else if (*ptr_a== (xmlChar)'/' || *ptr_a== (xmlChar)'\\') {
#ifdef _MSC_VER
	    *ptr_b = (xmlChar)((flag_url) ? '/' : '\\');
#else
	    *ptr_b = (xmlChar)'/';
#endif
	    ptr_b++;
	    while (*ptr_a== (xmlChar)'/' || *ptr_a== (xmlChar)'\\') ptr_a++;

	    if (ptr_a[0]== (xmlChar)'.' && ptr_a[1]== (xmlChar)'.' && (ptr_a[2]== (xmlChar)'\0' || ptr_a[2]== (xmlChar)'/' || ptr_a[2]== (xmlChar)'\\')) {
		/* one dir up */
		while (ptr_b > str_result && !(*ptr_b== (xmlChar)'/' || *ptr_b== (xmlChar)'\\')) ptr_b--;
		while (ptr_b > str_result &&  (*ptr_b== (xmlChar)'/' || *ptr_b== (xmlChar)'\\')) ptr_b--;
		while (ptr_b > str_result && !(*ptr_b== (xmlChar)'/' || *ptr_b== (xmlChar)'\\')) ptr_b--;
		ptr_a += 2;
	    }
	    else if (ptr_a[0]== (xmlChar)'.' && (ptr_a[1]== (xmlChar)'\0' || ptr_a[1]== (xmlChar)'/' || ptr_a[1]== (xmlChar)'\\')) {
		/* dir self */
		ptr_a += 2;
	    }
	}
	else {
	    *ptr_b = *ptr_a;
	    ptr_a++;
	    ptr_b++;
	}
    }
    /* cut all trailing path separators, but not MS DOS drive root directory */
    for (ptr_b--; ptr_b > str_result && (*ptr_b== (xmlChar)'/' || *ptr_b== (xmlChar)'\\') && *(ptr_b-1) != (xmlChar)':'; ptr_b--) ;
    ptr_b[1]= (xmlChar)'\0';
    printf_log(4,"Post normalize '%s'",str_result);
    return str_result;
}


#ifdef TESTCODE

/*! test code for this module 
*/
int
test_module_dir(void)
{
    xmlChar *ptr;
    xmlNodePtr node_ptr;
    int n_ok;
    int i;

    n_ok=0;
    i=0;

    /*
    */
    i++;
    ptr=file_normalize("..","..","temp","abc/Def",NULL);
    if (xmlStrcasecmp(ptr,BAD_CAST"C:\\home\\alex\\develop\\cxproc\\temp\\abc\\Def")==0) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_normalize(): %i",i);
    }
    if (ptr) xmlFree(ptr);

    /*
    */
    i++;
    ptr=file_normalize("c:\\temp","abc/Def",NULL);
    if (xmlStrcasecmp(ptr,BAD_CAST"c:\\temp\\abc\\Def")==0) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_normalize(): %i",i);
    }
    if (ptr) xmlFree(ptr);

    /*
    */
    i++;
    ptr=file_normalize("file://","c:/temp","abc","Def",NULL);
    if (xmlStrcmp(ptr,BAD_CAST"c:\\temp\\abc\\Def")==0) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_normalize(): %i",i);
    }
    if (ptr) xmlFree(ptr);

    /*
    */
    i++;
    ptr=file_normalize("../../","contrib/pie","REadme.txt",NULL);
    if (xmlStrcasecmp(ptr,BAD_CAST"C:\\home\\alex\\develop\\cxproc\\contrib\\pie\\REadme.txt")==0) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_normalize(): %i",i);
    }
    if (ptr) xmlFree(ptr);

    /*
    */
    i++;
    ptr=file_normalize("/tmp","abc/Def","sjhkf/fslj fslfjl",NULL);
    if (xmlStrcasecmp(ptr,BAD_CAST"c:\\tmp\\abc\\Def\\sjhkf\\fslj fslfjl")==0) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_normalize(): %i",i);
    }
    if (ptr) xmlFree(ptr);

    /*
    */
    i++;
    ptr=file_normalize("ftp://ftp.test.de/","qwert",NULL);
    if (xmlStrcasecmp(ptr,BAD_CAST"ftp://ftp.test.de/qwert")==0) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_normalize(): %i",i);
    }
    if (ptr) xmlFree(ptr);

    /*
    */
    i++;
    ptr=file_normalize("http://www.test.de/abc.html","qwert",NULL);
    if (xmlStrcasecmp(ptr,BAD_CAST"http://www.test.de/abc.html/qwert")==0) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_normalize(): %i",i);
    }
    if (ptr) xmlFree(ptr);

    /*
    */
    i++;
    node_ptr = filenode_search(xmlDocGetRootElement(doc_search),BAD_CAST"pie.css",TRUE);
    if (node_ptr != NULL) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_normalize(): %i = find partial path",i);
    }

    /*
    */
    i++;
    node_ptr = filenode_search(xmlDocGetRootElement(doc_search),BAD_CAST"blue_bright",TRUE);
    if (node_ptr != NULL) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_normalize(): %i = find partial path",i);
    }

    /*
    */
    i++;
    node_ptr = file_search(NULL,BAD_CAST"blue_bright");
    if (node_ptr != NULL) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_normalize(): %i = find partial path",i);
    }

    /*
    */
    i++;
    node_ptr = file_search(NULL,BAD_CAST"blue_bright/presentation.css");
    if (node_ptr != NULL) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_normalize(): %i = find partial path",i);
    }

    /*
    */
    i++;
    ptr=file_normalize(file_normalize(file_search(NULL,BAD_CAST"config.cxp"),NULL),NULL);
    if (xmlStrcasecmp(ptr,BAD_CAST"C:\\home\\alex\\develop\\cxproc\\contrib\\pie\\config.cxp")==0) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_normalize(): %i",i);
    }
    if (ptr) xmlFree(ptr);

    /*
    */
    i++;
    ptr=file_normalize(file_search(NULL,BAD_CAST"Validate.xsl"),NULL);
    if (xmlStrcasecmp(ptr,BAD_CAST"C:\\home\\alex\\develop\\cxproc\\xml\\xsl\\Validate.xsl")==0) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_normalize(): %i",i);
    }
    /*
    */
    i++;
    if (xmlStrcasecmp(file_extension(ptr),BAD_CAST"xsl")==0) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_normalize(): %i",i);
    }
    if (ptr) xmlFree(ptr);

    /*
    */
    i++;
    ptr=file_normalize(file_search(NULL,BAD_CAST"pie/html/blue_bright/presentation.css"),NULL);
    if (xmlStrcasecmp(ptr,BAD_CAST"C:\\home\\alex\\develop\\cxproc\\contrib\\pie\\html\\blue_bright\\presentation.css")==0) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_normalize(): %i = find partial path",i);
    }
    if (ptr) xmlFree(ptr);

    /*
    */
    i++;
    ptr=file_basedir(BAD_CAST"Validate.xsl");
    if (ptr==NULL) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_basedir(): %i",i);
    }
    if (ptr) xmlFree(ptr);

    /*
    */
    i++;
    ptr=file_basedir(file_search(NULL,BAD_CAST"MarkDateLinks.xsl"));
    if (xmlStrcasecmp(ptr,BAD_CAST"C:\\home\\alex\\develop\\cxproc\\contrib\\pie\\xsl")==0) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_basedir(): %i",i);
    }
    if (ptr) xmlFree(ptr);

    /*
    */
    i++;
    ptr=file_basedir(file_search(NULL,BAD_CAST"entries"));
    if (ptr==NULL) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_basedir(): %i",i);
    }
    if (ptr) xmlFree(ptr);

    printf_log(1,"Test: %i/%i OK in '%s'", n_ok,i,__FILE__);

    return (n_ok - i);
}
#endif
