/*
  cxproc - Configurable Xml PROCessor

  Copyright (C) 2006,2007,2008 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <assert.h>
#include <limits.h>

/* 
 */
#include <libxml/HTMLtree.h>
#include <libxml/parser.h>

/* 
 */
#include <libxslt/xslt.h>
#include <libxslt/xsltutils.h>
#include <libxslt/transform.h>
#include <libxslt/variables.h>

#include <pcre.h>
/* 
 */
#include "basics.h"
#include "tools.h"
#include "dir_x.h"
#include "dom_x.h"
#include "make_x.h"

#include "cxp_dtd.h"

struct {
  xmlDocPtr dom;
  xmlChar *name;
  int uses;
} arr_dom[BUFFER_LENGTH];

/*!\todo arr_dom[0].dom = NULL; resize arr_dom[] dynamically 

*/


/*! returns a xmlDocPtr to a cached DOM with same name

*/
xmlDocPtr
get_dom(const xmlChar *str_name)
{
  int i;
  xmlChar *str_name_norm;

  if (str_name && (str_name_norm = file_normalize(str_name,NULL))!=NULL) {
    for (i=0; arr_dom[i].dom && i < BUFFER_LENGTH; i++) {
      if (xmlStrEqual(arr_dom[i].name,str_name_norm)) {
	/*!\todo verify cached DOM against file modification */
	printf_log(2,"Found cached DOM of '%s'", (char *)str_name_norm);
	arr_dom[i].uses++;
	return arr_dom[i].dom;
      }
    }
    printf_log(3,"No cached DOM of '%s' found", (char *)str_name_norm);
    xmlFree(str_name_norm);
  }
  return NULL;
}
/* end of get_dom() */


/*! returns a xmlDocPtr to a cached DOM with same name

*/
xmlChar *
get_dom_file(const xmlChar *str_name)
{
  int i;
  /* search in DOM cache */
    for (i=0; arr_dom[i].dom && i < BUFFER_LENGTH; i++) {
      if (xmlStrstr(arr_dom[i].name,str_name)) {
        return arr_dom[i].name;
      }
    }
  return NULL;
}
/* end of get_dom_file() */


/*! register a complete DOM copy doc fo later re-use, identification with str_name
*/
BOOL_T
set_dom(const xmlDocPtr doc, const xmlChar *str_name)
    {
    int i;
    xmlNodePtr r;
    xmlChar *str_name_norm;

    if (doc!=NULL 
	&& (r = xmlDocGetRootElement(doc)) && xmlStrcmp(r->name,BAD_CAST "html") != 0
	&& str_name != NULL && xmlStrlen(str_name) > 0 && xmlStrEqual(str_name,BAD_CAST "-")==0
	&& (str_name_norm = file_normalize(str_name,NULL))!=NULL) {
	    for (i=0; i < BUFFER_LENGTH; i++) {
		if (xmlStrEqual(str_name_norm,arr_dom[i].name)) {
		    /* xmlFree() old versions */
		    xmlFreeDoc(arr_dom[i].dom);
		    printf_log(1,"Replacing DOM of '%s'", (char *)str_name_norm);
		    arr_dom[i].uses = 0;
		    arr_dom[i].dom = doc;
		    xmlFree(str_name_norm);
		    return TRUE;
		    }
		else if (arr_dom[i].dom==NULL) {
		    /* register new DOM at end of array */
		    printf_log(1,"Caching DOM of '%s'", (char *)str_name_norm);
		    arr_dom[i].uses = 0;
		    arr_dom[i].dom = doc;
		    arr_dom[i].name = str_name_norm;
		    arr_dom[i+1].dom = NULL;
		    return TRUE;
		    }
		}
	}
    return FALSE;
    }
/* end of set_dom() */


/*! 
*/
void
free_dom(void)
{
  int i;

  if (!arr_dom[0].dom) {
    printf_log(1,"There are no cached DOMs");
    return;
  }

  printf_log(1,"Freeing the cached DOMs");
  for (i=0; arr_dom[i].dom && i < BUFFER_LENGTH; i++) {
    printf_log(1,"%3i. %3i '%s'", i, arr_dom[i].uses, arr_dom[i].name);
#if 0
    /* save DOM for debug help */
    xmlStrPrintf(buffer,BUFFER_LENGTH,BAD_CAST "dom-%i.xml",i);
    xmlSaveFormatFileEnc((const char *)buffer, arr_dom[i].dom, NULL, 1);
#endif
    xmlFreeDoc(arr_dom[i].dom);
    arr_dom[i].dom = NULL;
    xmlFree(arr_dom[i].name);
  }
}
/* end of free_dom() */


/*! returns a string of descant H
*/
xmlChar *
get_parent_str(xmlNodePtr node_n)
{
  xmlChar *result;

  for ( result=NULL; node_n; node_n = node_n->parent) {
    if (IS_NODE(node_n,BAD_CAST"section")) {
      xmlChar *ptr_cur = result;
#if 1
      /*!\todo use recursive all XML_TEXT_NODE */
      result = xmlStrdup(child_content(child_element(node_n,BAD_CAST "h")));
#else
      xmlNodePtr node_h_text;
      result = xmlStrdup(BAD_CAST " BB ");
      for (node_h_text = child_element(node_n,BAD_CAST "H"); node_h_text; node_h_text=node_h_text->next) {
	  if (node_h_text->type == XML_ELEMENT_NODE) {
	    result = xmlStrcat(result,child_content(node_h_text));
	  }
	  else if (node_h_text->type == XML_TEXT_NODE) {
	    result = xmlStrcat(result,node_h_text->content);
	  }
      }
#endif
      result = xmlStrcat(result, BAD_CAST " :: ");
      result = xmlStrcat(result, ptr_cur);
      xmlFree(ptr_cur);
    }
  }

  if (result) {
    if (xmlStrlen(result) > 0) {
      return result;
    }
    else {
      xmlFree(result);
    }
  }
  return NULL;
}
/* end of get_parent_str() */


/*! searchs under the nexts of 'a_node' for an element named 'element_name'

\param a_node is the current node (for recursion)
\param element_name string for name element

\return the next node with name 'element_name'
*/
xmlNodePtr
next_element(xmlNodePtr a_node, xmlChar *element_name)
{
  xmlNodePtr cur_node;

  if (a_node) {
    if (element_name==NULL) {
      element_name = BAD_CAST a_node->name;
    }
    for (cur_node = a_node->next; cur_node; cur_node = cur_node->next) {
      if (cur_node->type == XML_ELEMENT_NODE
	  && !xmlStrcasecmp(cur_node->name, element_name)) {
	return cur_node;
      }
    }
  }
  return NULL;
}
/* End of next_element() */


/*! 
*/
BOOL_T
dump_dom(char *name_file, xmlNodePtr node, int depth)
{
  FILE *out;
  if (node && (out = fopen(name_file,"w"))) {
    printf_log(1,"Write Dump to '%s'", name_file);
    fprintf(out,"digraph pkg2 {\nnode [shape=record];\n");
    fprintf(out,"rankdir=\"LR\";\n");
    dump_dom_recursive(out,node,depth);
    fprintf(out,"}\n\n");
    fclose(out);
    return TRUE;
  } 
  return FALSE;
}
/* end of dump_dom() */


/*! 
*/
void
dump_dom_recursive(FILE *out, xmlNodePtr node, int depth)
{
  if (node && depth > 0) {
#if 1
    if (node->type == XML_ELEMENT_NODE) {
      fprintf(out,
	      "\"%x\" [label = \"ELEMENT|\\\"%s:%s\\\"\"];\n",
	      (unsigned int) node,
	      ((node->ns == NULL) ? "null" : (char *)node->ns->prefix),
	      (char *)node->name);
    }
    else if (node->type == XML_TEXT_NODE) {
      xmlChar *ptr_a = BAD_CAST xmlStrchr(node->content,'\n');
      int l_max = (ptr_a ? ptr_a - node->content : 8);
      fprintf(out,
	      "\"%x\" [label = \"TEXT|\\\"%s\\\"\"];\n",
	      (unsigned int) node,
	      (char *)xmlStrsub(node->content,0,l_max)
	      );
    }
    else if (node->type == XML_ATTRIBUTE_NODE) {
      fprintf(out,"\"%x\" [label = \"ATTRIBUTE|\\\"%s\\\"\"];\n", (unsigned int) node, (char *)node->name);
    }
    else if (node->type == XML_ENTITY_REF_NODE) {
      fprintf(out,"\"%x\" [label = \"ENTITY_REF|\\\"%s\\\"\"];\n", (unsigned int) node, (char *)node->name);
    }
    if (node->parent)
      fprintf(out,"\"%x\" -> \"%x\" [label = \"%s\"];\n", (unsigned int) node, (unsigned int) (node->parent ? node->parent : node), "parent");
    if (node->next)
      fprintf(out,"\"%x\" -> \"%x\" [label = \"%s\"];\n", (unsigned int) node, (unsigned int) (node->next ? node->next : node), "next");
    if (node->children)
      fprintf(out,"\"%x\" -> \"%x\" [label = \"%s\"];\n", (unsigned int) node, (unsigned int) (node->children ? node->children : node), "children");
    if (node->last)
      fprintf(out,"\"%x\" -> \"%x\" [label = \"%s\"];\n", (unsigned int) node, (unsigned int) (node->last ? node->last : node), "last");
    if (node->properties)
      fprintf(out,"\"%x\" -> \"%x\" [label = \"%s\"];\n", (unsigned int) node, (unsigned int) (node->properties ? (unsigned int)node->properties : (unsigned int)node), "properties");
#else
    if (node->type == XML_ELEMENT_NODE) {
      fprintf(out,"ELEMENT|%s", (char *)node->name);
    }
    else if (node->type == XML_TEXT_NODE) {
      fprintf(out,"TEXT|%s", (char *)xmlStrsub(node->content,0,8));
    }
    else if (node->type == XML_ATTRIBUTE_NODE) {
      fprintf(out,"ATTRIBUTE|%s", (char *)node->name);
    }
    else if (node->type == XML_ENTITY_REF_NODE) {
      fprintf(out,"ENTITY_REF|%s", (char *)node->name);
    }
#endif
    fprintf(out,"\n");
    dump_dom_recursive(out,(xmlNodePtr)node->properties,depth); 
    dump_dom_recursive(out,node->children,depth-1);
    dump_dom_recursive(out,node->next,depth-1);
  }
  return;
}
/* end of dump_dom_recursive() */


/*! counts the number of elements with right name

  \param a_node is the current node
  \param element_name string for name element
  \param n counter for recursive call

  \return the number of elements with name 'element_name'
*/
index_t
number_of(xmlNode * a_node, const xmlChar *element_name, index_t n)
{
  while (a_node) {
    if (a_node->type == XML_ELEMENT_NODE && !xmlStrcasecmp(a_node->name, element_name))
      n++;

    n = number_of(a_node->children, element_name, n);

    a_node = a_node->next;
  }
  return n;
}


/*! counts the number of child elements with right name

  \param a_node is the current node
  \param element_name string for name element (with element_name==NULL count childs only)

  \return the number of child elements with name 'element_name'
*/
index_t
number_of_child(xmlNode * a_node, const xmlChar *element_name)
{
  xmlNodePtr c_node;
  index_t result;

  for (result=0, c_node=a_node->children; c_node; c_node=c_node->next) {
    if (IS_NODE(c_node,NULL)
	&& (element_name==NULL || xmlStrEqual(c_node->name, element_name)))
      result++;
  }
  return result;
}


/*! searchs under the childs of 'a_node' for an element named 'element_name'

  \param a_node is the current node (for recursion)
  \param element_name string for name element

  \return the child node with name 'element_name'
*/
xmlNodePtr
child_element(xmlNode * a_node, const xmlChar *element_name)
{
  xmlNode *cur_node;

  assert(a_node);
  for (cur_node = a_node->children; cur_node; cur_node = cur_node->next) {
    if (IS_NODE(cur_node,element_name)) {
      return cur_node;
    }
  }
  return NULL;
}
/* End of child_element() */


/*!

\return the text (the child content) of a node
*/
xmlChar *
child_content(xmlNode * a_node)
{
  if (a_node && a_node->children && a_node->children->type == XML_TEXT_NODE)
    return a_node->children->content;

  return NULL;
}
/* End of child_content() */


/*!
\param a_node parent node for attributes
\param attribute_name name of wanted attribute
\return the content pointer of the attribute named 'attribute_name'
of 'a_node' OR NULL if no attribute found
*/
xmlChar *
attribute_str(xmlNode * a_node, const xmlChar *attribute_name)
{
  xmlNodePtr curr_attr = attribute_node(a_node,attribute_name);

if (curr_attr!=NULL && curr_attr->children!=NULL) {
  return curr_attr->children->content;
}

  return NULL;
}
/* End of attribute_str() */


/*!
\param a_node parent node for attributes
\param attribute_name name of wanted attribute
\return the node pointer of the attribute named 'attribute_name'
of 'a_node' OR NULL if no attribute found
*/
xmlNodePtr
attribute_node(xmlNode * a_node, const xmlChar *attribute_name)
{
  xmlAttr *curr_attr;

  if (a_node && attribute_name) {
    for (curr_attr = a_node->properties; curr_attr; curr_attr = curr_attr->next)
      if (curr_attr->type == XML_ATTRIBUTE_NODE
	  && !xmlStrcasecmp(curr_attr->name, attribute_name))
	return (xmlNodePtr) curr_attr; /*!\todo is there a difference between xmlNodePtr and (xmlAttr *) */
  }
  return NULL;
}
/* End of attribute_node() */


/*! Save a copy of the result DOM or string to a file. Register the
  copy for caching and leave doc_result untouched.
*/
BOOL_T
save_file(const xmlDocPtr doc_result, const xmlChar *result_str, const xmlNodePtr n)
{
  xmlChar *attr_target_file;
  xmlChar *attr_mode;
  xmlChar *attr_cache;
  xmlDocPtr doc_save;
  BOOL_T mode_cache;
  BOOL_T mode_append = FALSE;

  if (doc_result == NULL && result_str == NULL) {
    printf_log(1,"Empty result");
    return FALSE;
  }

  if (n == NULL) {
    return FALSE;
  }

  if (isreadonly()==TRUE) {
    /* we are running in readonly mode */
    mode_cache  = TRUE;
  }
  else {
    mode_cache  = FALSE;
  }

  if ((attr_mode = attribute_str(n,BAD_CAST "mode"))
      && xmlStrcasecmp(attr_mode,BAD_CAST "append")==0) {
    mode_append = TRUE;
  }
  /*!\todo respect the other \@mode values (s. DTD) */

  if ((attr_cache = attribute_str(n,BAD_CAST "cache"))
      && !xmlStrcasecmp(attr_cache,BAD_CAST "yes")) {
    mode_cache = TRUE;
  }

  attr_target_file = file_normalize(attribute_str(n,BAD_CAST "name"),NULL);
  if (attr_target_file) {
    /* Save the result only if a \@NAME is given.
     */
    printf_log(1,"Saving result to '%s'", attr_target_file);
    if (doc_result) {
      if (xmlStrEqual(attr_target_file,BAD_CAST "-")) {
	/* dont cache stdout */
	if (runmode == cgi) {
	  /* s. RFC 1806 */
	  xmlChar *attr_type = attribute_str(n,BAD_CAST "type");
	  if (attr_type && xmlStrlen(attr_type)>5) {
	    printf("Content-Type: %s; charset=utf-8\n", attr_type);
	    puts("Content-Disposition: attachment; filename=cxproc");
	    puts("Content-Description: Dynamic cxproc content");
	  }
	  else {
	    if (IS_NODE_XML(n)) {
	      puts("Content-type: text/xml; charset=utf-8");
	    }
	    else if (IS_NODE_XHTML(n)) {
	      puts("Content-type: text/html; charset=utf-8");
	    }
	    else {
	      puts("Content-type: text/plain; charset=utf-8");
	    }
	  }
	  puts("");
	  fflush(stdout); /*! because problems with VC++ (reverse order in stdout) */
	}

	if (IS_NODE_XHTML(n)) {
	  htmlSaveFileFormat((const char *)attr_target_file, doc_result, "UTF-8", 0);
	}
	else {
	  xmlSaveFormatFileEnc((const char *)attr_target_file, doc_result, "UTF-8", 1);
	}
      }
      else {
	/* normal file output */

	if (mode_cache) {
	  /* make a copy of doc_result, save this to named file and cache DOM
	     (because later transformations)
	  */
	  doc_save = xmlCopyDoc(doc_result,1);
	  if (doc_save) {
	    set_dom(doc_save, attr_target_file);
	  }
	  else {
	    printf_log(1,"Cache error");
	  }
	}

	if (isreadonly()) {
	  /* dont write any file!! */
	  printf_log(1,"Cant write files in readonly mode");
	}
	else {
	  /*!\todo use encoding settings from DOM (s. doc_save->encoding ) */
	  /* 	xmlSaveFormatFileEnc((const char *)attr_target_file, doc_save, "ISO-8859-1", 1); */

	  if (IS_NODE_XHTML(n)) {
	    htmlSaveFileFormat((const char *)attr_target_file, doc_result, NULL, 1);
	  }
	  else {
	    xmlSaveFormatFileEnc((const char *)attr_target_file, doc_result, NULL, 1);
	  }
	}
	/*!\todo change file permissions to value in \@PERM */
      }
    }
    else {
      if (xmlStrEqual(attr_target_file,BAD_CAST "-")) {
	if (runmode == cgi) {
	  xmlChar *attr_type = attribute_str(n,BAD_CAST "type");
	  if (attr_type && xmlStrlen(attr_type)>5) {
	    printf("Content-type: %s; charset=utf-8\n\n", attr_type);
	  }
	  else {
	    printf("Content-type: text/plain; charset=utf-8\n\n");
	  }
	  fflush(stdout); /*! because problems with VC++ (reverse order in stdout) */
	}
	puts((char *)result_str);
      }
      else if (isreadonly()) {
	/* dont write any file!! */
	printf_log(1,"Cant write files in readonly mode");
      }
      else {
	FILE *out = fopen((char *)attr_target_file, (mode_append ? "w+" : "w"));
#if 1
	fputs((char *)result_str,out);
#else
	fwrite(result_str,strlen(result_str),1,out);
#endif	
	fclose(out);
      }
    }
    xmlFree(attr_target_file);
  }
  else {
    printf_log(3,"Anonymous result");
  }

  return TRUE;
}
/* end of save_file() */


/*! compare two node with all childrens

\param node_a first candidate
\param node_b second candidate
*/
BOOL_T
node_is_equal(xmlNodePtr node_a, xmlNodePtr node_b)
{
  xmlNodePtr node_a_child;
  xmlNodePtr node_b_child;

  if ( node_a==NULL || node_b==NULL) {
    /* node_a or node_b is NULL */
    return FALSE;
  }

  if (node_a == node_b) {
    /* pointers to same nodes */
    return TRUE;
  }

  /* pointers to different nodes, compare node element properties
   */
  if ( node_a->type != node_b->type
       || !xmlStrEqual(node_a->name,node_b->name)
       || !xmlStrEqual(node_a->content,node_b->content)) {
    return FALSE;
  }

  /* compare all childrens */
  for (node_a_child = node_a->children, node_b_child = node_b->children;
       node_a_child && node_b_child;
       node_a_child = node_a_child->next, node_b_child = node_b_child->next ) {
    if (!node_is_equal(node_a_child, node_b_child)) {
      return FALSE;
    }
  }
  /* compare all properties */
  for (node_a_child = (xmlNodePtr)node_a->properties, node_b_child = (xmlNodePtr)node_b->properties;
       node_a_child && node_b_child;
       node_a_child = node_a_child->next, node_b_child = node_b_child->next ) {
    if (!node_is_equal(node_a_child, node_b_child)) {
      return FALSE;
    }
  }
  return TRUE;
}
/* end of node_is_equal() */


#if 0
/* derived from from xmlUnlinkNode() in libxml2-2.6.26/tree.c */
/**
 * xmlUnlinkNodeList:
 * @cur:  the node
 *
 * Unlink a node from it's current context, the node is not freed
 */
void
xmlUnlinkNodeList(xmlNodePtr cur) {
    if (cur == NULL) {
#ifdef DEBUG_TREE
        xmlGenericError(xmlGenericErrorContext,
		"xmlUnlinkNodeList : node == NULL\n");
#endif
	return;
    }
    if (cur->type == XML_DTD_NODE) {
	xmlDocPtr doc;
	doc = cur->doc;
	if (doc != NULL) {
	    if (doc->intSubset == (xmlDtdPtr) cur)
		doc->intSubset = NULL;
	    if (doc->extSubset == (xmlDtdPtr) cur)
		doc->extSubset = NULL;
	}
    }
    if (cur->parent != NULL) {
	xmlNodePtr parent;
	xmlNodePtr n;
	parent = cur->parent;
	if (cur->type == XML_ATTRIBUTE_NODE) {
	    if (parent->properties == (xmlAttrPtr) cur)
		parent->properties = NULL;
	} else {
	    if (parent->children == cur)
		parent->children = NULL;
	    if (parent->last == cur)
		parent->last = NULL;
	}

	for (n=cur; n != NULL; n=cur->next) {
	  if (n->type == XML_ELEMENT_NODE)
	    n->doc = NULL;
	    n->parent = NULL;
	}
    }
    /*
    if (cur->next != NULL)
        cur->next->prev = cur->prev;
    */
    if (cur->prev != NULL) {
      cur->prev->next = NULL;
      cur->prev = NULL;
    }
/*     cur->next = cur->prev = NULL; */
}
#endif


