/*
  cxproc - Configurable Xml PROCessor

  Copyright (C) 2006,2007,2008 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#ifdef _MSC_VER
#include "config-win32.h"
#else
#include <unistd.h>
#include "../config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <assert.h>
#include <errno.h>

/* 
 */
#include <libxml/xmlversion.h>
#include <libxml/HTMLtree.h>
#include <libxml/parser.h>

#ifdef LIBXML_THREAD_ENABLED
#include <libxml/globals.h>
#include <libxml/threads.h>
#if HAVE_LIBPTHREAD
#include <pthread.h>
#endif
#endif

#include <iconv.h>

/* 
 */
#include "cgi-util.h"

#include <pcre.h>

/* 
 */
#include <libxslt/xslt.h>
#include <libxslt/xsltutils.h>
#include <libxslt/transform.h>
#include <libxslt/variables.h>

/* 
 */
#include "basics.h"
#include "tools.h"
#include "iconv_x.h"
#include "dom_x.h"
#include "dir_x.h"
#include "make_x.h"
#include "plain_x.h"
#include "calendar_x.h"
#include "petrinet_x.h"
#include "cxp_dtd.h"
#ifdef HAVE_LIBMYSQL
#include "sql_x.h"
#endif
#ifdef HAVE_LIBMAGICK
#include <magick/ImageMagick.h>
#include "image_x.h"
#endif


cd_collection_t cd;


/*! at init create the conversion descriptors 
 */
BOOL_T
set_encodingsUTF8(xmlChar *id_env, xmlChar *id_cgi, xmlChar *id_argv, xmlChar *id_fs, xmlChar *id_plain)
{

  /*\todo detect encoding from LANG or attributes of make element */

  if (id_env!=NULL && xmlStrlen(id_env) > 0) {

    if (xmlStrcasecmp(id_env,BAD_CAST "UTF-8")) { /* avoid 1:1 conversion */
      /*
	create the required conversion descriptors
       */
      cd.from.env = iconv_open("UTF-8",(char *)id_env);
      if (cd.from.env == (iconv_t) -1) {
	/* Something went wrong.  */
	printf_exit((errno == EINVAL),77,"Conversion from '%s' to UTF-8 not possible",id_env);
	assert(TRUE);
      }
      else {    
	printf_log(2, "Environment conversion '%s' to '%s'", id_env, "UTF-8");
 	iconv(cd.from.env, NULL, NULL, NULL, NULL);
      }
      cd.to.env = iconv_open((char *)id_env,"UTF-8");
      if (cd.to.env == (iconv_t) -1) {
	/* Something went wrong.  */
	printf_exit((errno == EINVAL),77,"conversion to '%s' to UTF-8 not possible",id_env);
	assert(TRUE);
      }
      else {    
	printf_log(2, "Environment conversion '%s' to '%s'", "UTF-8", id_env);
 	iconv(cd.to.env, NULL, NULL, NULL, NULL);
      }
    }
    else {
      printf_log(2, "No Environment conversion necessary");
      cd.from.env = NULL;
      cd.to.env = NULL;
    }
  }

  if (id_cgi!=NULL && xmlStrlen(id_cgi) > 0) {

    if (xmlStrcasecmp(id_cgi,BAD_CAST "UTF-8")) { /* avoid 1:1 conversion */
      /*
	create the required conversion descriptors
       */
      cd.from.cgi = iconv_open("UTF-8",(char *)id_cgi);
      if (cd.from.cgi == (iconv_t) -1) {
	/* Something went wrong.  */
	printf_exit((errno == EINVAL),77,"Conversion from '%s' to UTF-8 not possible",id_cgi);
	assert(TRUE);
      }
      else {    
	printf_log(2, "CGI conversion '%s' to '%s'", id_cgi, "UTF-8");
 	iconv(cd.from.cgi, NULL, NULL, NULL, NULL);
      }
      cd.to.cgi = iconv_open((char *)id_cgi,"UTF-8");
      if (cd.to.cgi == (iconv_t) -1) {
	/* Something went wrong.  */
	printf_exit((errno == EINVAL),77,"conversion to '%s' to UTF-8 not possible",id_cgi);
	assert(TRUE);
      }
      else {    
	printf_log(2, "CGI conversion '%s' to '%s'", "UTF-8", id_cgi);
 	iconv(cd.to.cgi, NULL, NULL, NULL, NULL);
      }
    }
    else {
      printf_log(2, "No CGI conversion necessary");
      cd.from.cgi = NULL;
      cd.to.cgi = NULL;
    }
  }

  if (id_plain!=NULL && xmlStrlen(id_plain) > 0) {

    if (xmlStrcasecmp(id_plain,BAD_CAST "UTF-8")) { /* avoid 1:1 conversion */
      /*
	create the required conversion descriptors
       */
      cd.from.plain = iconv_open("UTF-8",(char *)id_plain);
      if (cd.from.plain == (iconv_t) -1) {
	/* Something went wrong.  */
	printf_exit((errno == EINVAL),77,"Conversion from '%s' to UTF-8 not possible",id_plain);
	assert(TRUE);
      }
      else {    
	printf_log(2, "Plain text conversion '%s' to '%s'", id_plain, "UTF-8");
 	iconv(cd.from.plain, NULL, NULL, NULL, NULL);
      }
      cd.to.plain = iconv_open((char *)id_plain,"UTF-8");
      if (cd.to.plain == (iconv_t) -1) {
	/* Something went wrong.  */
	printf_exit((errno == EINVAL),77,"Conversion to '%s' to UTF-8 not possible",id_plain);
	assert(TRUE);
      }
      else {    
	printf_log(2, "Plain text conversion '%s' to '%s'", "UTF-8", id_plain);
 	iconv(cd.to.plain, NULL, NULL, NULL, NULL);
      }
    }
    else {
      printf_log(2, "No Plain text conversion necessary");
      cd.from.plain = NULL;
      cd.to.plain = NULL;
    }
  }

  if (id_fs!=NULL && xmlStrlen(id_fs) > 0) {

    if (xmlStrcasecmp(id_fs,BAD_CAST "UTF-8")) { /* avoid 1:1 conversion */
      /*
	create the required conversion descriptors
       */
      cd.from.fs = iconv_open("UTF-8",(char *)id_fs);
      if (cd.from.fs == (iconv_t) -1) {
	/* Something went wrong.  */
	printf_exit((errno == EINVAL),77,"Conversion from '%s' to UTF-8 not possible",id_fs);
	assert(TRUE);
      }
      else {    
	printf_log(2, "Filesystem conversion '%s' to '%s'", id_fs, "UTF-8");
 	iconv(cd.from.fs, NULL, NULL, NULL, NULL);
      }
      cd.to.fs = iconv_open((char *)id_fs,"UTF-8");
      if (cd.to.fs == (iconv_t) -1) {
	/* Something went wrong.  */
	printf_exit((errno == EINVAL),77,"Conversion to '%s' to UTF-8 not possible",id_fs);
	assert(TRUE);
      }
      else {    
	printf_log(2, "Filesystem conversion '%s' to '%s'", "UTF-8", id_fs);
 	iconv(cd.to.fs, NULL, NULL, NULL, NULL);
      }
    }
    else {
      printf_log(2, "No Filesystem conversion necessary");
      cd.from.fs = NULL;
      cd.to.fs = NULL;
    }
  }

  /*!\todo additional conversion descriptors */
  return TRUE;
}


/*! a mapper function with UTF-8 arguments and result
 */
xmlChar *
getenvUTF8(const xmlChar *ptr_name)
{
  char buffer_in[BUFFER_LENGTH];
  char buffer_out[BUFFER_LENGTH];
  char *ptr_out, *ptr_in;
  char *ptr_value;
  xmlChar *ptr_result = NULL;
  size_t int_in;
  size_t int_out;
  size_t nconv;

  if (ptr_name!= NULL && xmlStrlen(ptr_name) > 0) {

    if (cd.from.env) {
      /* convert name of environment variable */

      strncpy(buffer_in,(char *)ptr_name,BUFFER_LENGTH - 1);
      int_in = strlen(buffer_in) + 1;
      ptr_in  = buffer_in;
      int_out = BUFFER_LENGTH - 1;
      ptr_out = buffer_out;

      nconv = iconv(cd.to.env,&ptr_in,&int_in,&ptr_out,&int_out);
      iconv_error(errno,nconv);

      ptr_value = getenv((char *)buffer_out);
    }
    else {
      /* no conversion necessary */
      ptr_value = getenv((char *)ptr_name);
    }

    if (ptr_value != NULL) {
      if (cd.from.env) {
	strncpy(buffer_in,(char *)ptr_value,BUFFER_LENGTH - 1);
	int_in = strlen(buffer_in) + 1;
	ptr_in  = buffer_in;
	int_out = BUFFER_LENGTH - 1;
	ptr_out = buffer_out;

	nconv = iconv(cd.from.env,&ptr_in,&int_in,&ptr_out,&int_out);
	iconv_error(errno,nconv);
	ptr_result = xmlStrdup(BAD_CAST buffer_out);
      }
      else {
	/* no conversion necessary */
	ptr_result = xmlStrdup(BAD_CAST ptr_value);
      }
      printf_log(3, "Env '%s'='%s'", ptr_name, ptr_result);
    }
    else {
      printf_log(3, "No environment variable '%s'", ptr_name);
    }
  }

  return ptr_result;
}


/*! 
  \todo additional conversion for \& \' etc.
 */
xmlChar *
getcgiUTF8(const xmlChar *ptr_name)
{
  char buffer_in[BUFFER_LENGTH];
  char buffer_out[BUFFER_LENGTH];
  char *ptr_out, *ptr_in;
  char *ptr_value;
  xmlChar *ptr_result = NULL;
  size_t int_in;
  size_t int_out;
  size_t nconv;

  if (ptr_name!= NULL && xmlStrlen(ptr_name) > 0) {

    if (cd.from.cgi) {
      /* convert name of environment variable */

      strncpy(buffer_in,(char *)ptr_name,BUFFER_LENGTH - 1);
      int_in = strlen(buffer_in) + 1;
      ptr_in  = buffer_in;
      int_out = BUFFER_LENGTH - 1;
      ptr_out = buffer_out;

      nconv = iconv(cd.to.cgi,&ptr_in,&int_in,&ptr_out,&int_out);
      iconv_error(errno,nconv);

      ptr_value = cgi_getentrystr((char *)buffer_out);
    }
    else {
      /* no conversion necessary */
      ptr_value = cgi_getentrystr((char *)ptr_name);
    }

    if (ptr_value != NULL) {
      if (cd.from.cgi) {
	strncpy(buffer_in,(char *)ptr_value,BUFFER_LENGTH - 1);
	int_in = strlen(buffer_in) + 1;
	ptr_in  = buffer_in;
	int_out = BUFFER_LENGTH - 1;
	ptr_out = buffer_out;

	nconv = iconv(cd.from.cgi,&ptr_in,&int_in,&ptr_out,&int_out);
	iconv_error(errno,nconv);
	ptr_result = xmlStrdup(BAD_CAST buffer_out);
      }
      else {
	/* no conversion necessary */
	ptr_result = xmlStrdup(BAD_CAST ptr_value);
      }
      printf_log(2, "CGI '%s'='%s'", ptr_name, ptr_result);
    }
    else {
      printf_log(2, "No CGI variable '%s'", ptr_name);
    }
  }

  return ptr_result;
}


/*! a substitution with program arguments
 */
xmlChar *
getargvUTF8(const int int_argv)
{
  char buffer_in[BUFFER_LENGTH];
  char buffer_out[BUFFER_LENGTH];
  char *ptr_out, *ptr_in;
  char *ptr_value;
  xmlChar *ptr_result = NULL;
  size_t int_in;
  size_t int_out;
  size_t nconv;

  if (int_argv < 0 || int_argv >= argc_i) {
    printf_log(3,"Index '%i' too large (max. '%i')", int_argv, argc_i-1);
  }
  else {

    ptr_value = argv_i[int_argv];

    if (ptr_value != NULL && strlen(ptr_value) > 0) {
      if (cd.from.env) {
	strncpy(buffer_in,(char *)ptr_value,BUFFER_LENGTH - 1);
	int_in = strlen(buffer_in) + 1;
	ptr_in  = buffer_in;
	int_out = BUFFER_LENGTH - 1;
	ptr_out = buffer_out;

	nconv = iconv(cd.from.env,&ptr_in,&int_in,&ptr_out,&int_out);
	iconv_error(errno,nconv);
	ptr_result = xmlStrdup(BAD_CAST buffer_out);
      }
      else {
	/* no conversion necessary */
	ptr_result = xmlStrdup(BAD_CAST ptr_value);
      }
    }
    else {
      printf_log(2, "No valid argv '%i'", int_argv);
    }
  }

  return ptr_result;
}


xmlChar *
getcwdUTF8(void)
{
  char buffer_in[BUFFER_LENGTH];
  char buffer_out[BUFFER_LENGTH];
  char *ptr_out, *ptr_in;
  char *ptr_value;
  xmlChar *ptr_result = NULL;
  size_t int_in;
  size_t int_out;
  size_t nconv;

#ifdef _MSC_VER
  ptr_value = _getcwd((char *)buffer_in,BUFFER_LENGTH);
#else
  ptr_value = getcwd((char *)buffer_in,BUFFER_LENGTH);
#endif
  if (ptr_value != NULL && strlen(ptr_value) > 0) {
    if (cd.from.env) {
      strncpy(buffer_in,(char *)ptr_value,BUFFER_LENGTH - 1);
      int_in = strlen(buffer_in) + 1;
      ptr_in  = buffer_in;
      int_out = BUFFER_LENGTH - 1;
      ptr_out = buffer_out;

      nconv = iconv(cd.from.env,&ptr_in,&int_in,&ptr_out,&int_out);
      iconv_error(errno,nconv);
      ptr_result = xmlStrdup(BAD_CAST buffer_out);
    }
    else {
      /* no conversion necessary */
      ptr_result = xmlStrdup(BAD_CAST ptr_value);
    }
  }
  else {
    printf_log(2, "No valid working directory");
  }

  return ptr_result;
}


/* cleanup for libiconv 
 */
void
free_iconv(void) 
{

  if (cd.from.env != NULL) {
    iconv_close(cd.from.env);
  }
  if (cd.to.env != NULL) {
    iconv_close(cd.to.env);
  }
}


BOOL_T
iconv_error(int int_errno, size_t nconv)
{
  /*!\todo make iconv_error(int_errno); */
  if (int_errno==EILSEQ) {
    printf_log(1, "invalid byte sequence in the input");
    return FALSE;
  }
  else if (int_errno==E2BIG) {
    printf_log(1, "it ran out of space in the output buffer.");
    return FALSE;
  }
  else if (int_errno==EINVAL) {
    printf_log(1, "an incomplete byte sequence at the end of the input buffer.");
    return FALSE;
  }
  else if (int_errno==EBADF) {
    printf_log(1, "The cd argument is invalid.");
    return FALSE;
  }

#if 0
  if (nconv == (size_t) -1) {
    /* Not everything went right.  It might only be
       an unfinished byte sequence at the end of the
       buffer.  Or it is a real problem.  */
    if (int_errno == EINVAL)
      /* This is harmless.  Simply move the unused
	 bytes to the beginning of the buffer so that
	 they can be used in the next round.  */
      memmove (inbuf, inptr, insize);
    else {
      /* It is a real problem.  Maybe we ran out of
	 space in the output buffer or we have invalid
	 input.  In any case back the file pointer to
	 the position of the last processed byte.  */
      lseek (fd, -insize, SEEK_CUR);
      result = -1;
      break;
    }
  }
#endif
  return FALSE;
}


/*!tests the readability of a file or directory with given name 

  \param ptr_result relative or absolute path and name of file

  \return TRUE if readable

  \param path relative or absolute path and name of file
  \return true if the named file is is readable
*/
BOOL_T
type_is_readableUTF8(const xmlChar *ptr_path, const BOOL_T flag_dir, long int *ptr_size)
{
  int err;
  char buffer_in[BUFFER_LENGTH];
  char buffer_out[BUFFER_LENGTH];
  char *ptr_out, *ptr_in;
  size_t int_in;
  size_t int_out;
  size_t nconv;

#ifdef _MSC_VER
  struct _stat stbuf;
#else
  struct stat stbuf;
#endif

  if (ptr_path != NULL && xmlStrlen(ptr_path) > 0) {

    if (xmlStrcmp(ptr_path,BAD_CAST"-")==0) {	/* stdin */
      return TRUE;
    }

    if (xmlStrcasestr(ptr_path,BAD_CAST"http://")==ptr_path) { /* web access */
      return TRUE;
    }

    if (xmlStrcmp(ptr_path,BAD_CAST".")==0 || xmlStrcmp(ptr_path,BAD_CAST"..")==0) { /* relative path */
      return TRUE;
    }

    /*!\todo also other types of URI */

    if (cd.to.fs) {
      strncpy(buffer_in,(char *)ptr_path,BUFFER_LENGTH - 1);
      int_in = strlen(buffer_in) + 1;
      ptr_in  = buffer_in;
      int_out = BUFFER_LENGTH - 1;
      ptr_out = buffer_out;

      nconv = iconv(cd.to.fs,&ptr_in,&int_in,&ptr_out,&int_out);
      iconv_error(errno,nconv);
      ptr_out = buffer_out;
    }
    else {
      /* no conversion necessary */
      ptr_out = ptr_path;
    }
  }
  else {
    printf_log(2, "No valid directory path");
    return FALSE;
  }
  
#ifdef _MSC_VER
  err = _stat((const char*)ptr_out,&stbuf); 
#else
  err = stat((const char*)ptr_out,&stbuf); 
#endif
  
  if (err != -1) {
    if ((flag_dir==TRUE && S_ISDIR(stbuf.st_mode))
	||
	(flag_dir==FALSE && S_ISREG(stbuf.st_mode))) {
      /* check permissions of path */
      /*!\bug check the ownership of this directory entry */
#ifdef _MSC_VER
      if (stbuf.st_mode & _S_IREAD) { 
#else
      if (stbuf.st_mode & S_IRUSR) {
#endif
	if (ptr_size!=NULL) {
	  *ptr_size = (long int)stbuf.st_size;
	}
	return TRUE;
      } 
      else {
	printf_log(2, "%s '%s' not readable", 
		     (flag_dir==TRUE ? "Dir" : "File"), ptr_out);
      }
    } else {
      printf_log(2, "%s '%s' not regular", 
		   (flag_dir==TRUE ? "Dir" : "File"), ptr_out);
    }
    printf_log(2, "%s '%s' stat error", 
		 (flag_dir==TRUE ? "Dir" : "File"), ptr_out);
  }
  return FALSE;
}
/* end of dir_readable() */


/*! copies a single file 'ptr_from' to 'ptr_to'. There is no globbing nor
  directory handling (Use cxp:system instead).

  \param ptr_from source file name
  \param ptr_to directory or file name

  \return TRUE if successful
*/
BOOL_T
file_copyUTF8(xmlNodePtr node_copy)
{
  char buf[BUFFER_LENGTH];
  char buffer_from[BUFFER_LENGTH];
  char buffer_in[BUFFER_LENGTH];
  char buffer_to[BUFFER_LENGTH];
  char *ptri_from;
  char *ptri_to;
  char *ptri_free;
  char *ptri_out;
  FILE *ifp;
  FILE *ofp;
  int err;
  size_t int_from;
  size_t int_in;
  size_t int_to;
  size_t nconv;
  size_t nmemb;
  char *ptr_in;
  xmlChar *ptr_sep;
  xmlChar *ptr_basename;

  xmlChar *ptr_from;
  xmlChar *ptr_to;

  xmlChar *attr_to;
  xmlChar *attr_from;

  /* 
     source
   */
  attr_from = attribute_str(node_copy,BAD_CAST "from");
  if (attr_from == NULL) {
    printf_log(1,"file copy: no source attribute");
    return FALSE;;
  }

  if (xmlStrcasecmp(attribute_str(node_copy,BAD_CAST "search"),BAD_CAST "yes")==0) {
    /* search for file in DOM */
    attr_from = file_search(NULL, attr_from);
  }
  ptr_from = attr_from;

  if (ptr_from == NULL
      || xmlStrlen(ptr_from) < 1
      || xmlStrcmp(ptr_from,BAD_CAST"-")==0  /* stdin */
      || xmlStrstr(ptr_from,BAD_CAST"http://")==ptr_from /* web access */
      ) {
    /*!\todo accept stdin and http:// */
    printf_log(1,"file copy: no source file '%s' found",attribute_str(node_copy,BAD_CAST "from"));
    return FALSE;
  }

  if (cd.to.env) {
    /* 
       conversion of from path
    */
    strncpy(buffer_in,(char *)ptr_from,BUFFER_LENGTH - 1);
    int_in = strlen(buffer_in) + 1;
    ptr_in  = buffer_in;
    int_from = BUFFER_LENGTH - 1;
    ptri_from = buffer_from;

    nconv = iconv(cd.to.env,&ptr_in,&int_in,&ptri_from,&int_from);
    iconv_error(errno,nconv);
    ptri_from = buffer_from;
  }
  else {
    /* no conversion necessary */
    ptri_from = ptr_from;
  }

  if ((ifp=fopen(ptri_from, "rb")) == NULL) {
    printf_log(1,"Cant open '%s'", ptri_from);
    return FALSE;
  }
    
  /* 
     target
   */
  attr_to = attribute_str(node_copy,BAD_CAST "to");
  if (attr_to == NULL) {
    printf_log(1,"file copy: no targe specified");
    return FALSE;;
  }
  ptr_to = attr_to;

  if (xmlStrlen(ptr_to) < 1
      || xmlStrstr(ptr_to,BAD_CAST"http://")==ptr_to /* web access */
      ) {
    printf_log(2, "No valid directory path '%s'", ptr_to);
    return FALSE;
  }
  else if (xmlStrEqual(ptr_to,BAD_CAST"-")) {
    printf_log(2,"File copy from '%s' to stdout",ptr_from);
    ofp = stdout;

    if (runmode == cgi) {
      xmlChar *attr_type = attribute_str(node_copy,BAD_CAST "type");
      if (attr_type && xmlStrlen(attr_type)>5) {
	printf("Content-type: %s; charset=utf-8\n\n", attr_type);
	fflush(stdout); /*! because problems with VC++ (reverse order in stdout) */
      }
    }  
  }
  else {
    printf_log(2,"File copy from '%s' to '%s'",ptr_from,ptr_to);
  
    /*!\todo also other types of URI */

    if (cd.to.env) {
      /* 
	 conversion of to path
      */
      strncpy(buffer_in,(char *)ptr_to,BUFFER_LENGTH - 1);
      int_in = strlen(buffer_in) + 1;
      ptr_in  = buffer_in;
      int_to = BUFFER_LENGTH - 1;
      ptri_to = buffer_to;

      nconv = iconv(cd.to.env,&ptr_in,&int_in,&ptri_to,&int_to);
      iconv_error(errno,nconv);
      ptri_to = buffer_to;
    }
    else {
      /* no conversion necessary */
      ptri_to   = ptr_to;
    }

    if (ptr_to[xmlStrlen(ptr_to)-1]==(xmlChar)'/') {
      /* there is a trailing slash in ptri_to, this a pure directory name */
#ifdef _MSC_VER
#define MODE_DIR_CREATE (0)
#else
#define MODE_DIR_CREATE (S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)
#endif
      /*!\todo use umask for MODE_DIR_CREATE */
      mkdir_pathUTF8(ptr_to,MODE_DIR_CREATE);
    }

    if (dir_is_readableUTF8(BAD_CAST ptri_to)) {
      /* this directory exists already, find the pure file name in ptri_from */
      xmlChar *ptr_release;
      for (ptr_basename=ptr_sep=ptri_from; ptr_sep!=NULL && *ptr_sep!=(xmlChar)'\0'; ptr_sep++) {
	/* */
	if ((*ptr_sep==(xmlChar)'/' || *ptr_sep==(xmlChar)'\\') && *(ptr_sep+1)!=(xmlChar)'\0') {
	  ptr_basename = ptr_sep + 1;
	}
      }
      ptri_to = xmlStrncatNew(BAD_CAST ptri_to,BAD_CAST "/",2);
      ptr_release = ptri_to;
      ptri_to = xmlStrncatNew(ptr_release,ptr_basename,xmlStrlen(ptr_basename));
      xmlFree(ptr_release);
    }
    else if (file_is_readableUTF8(BAD_CAST ptri_to)) {
      /* there is a file already, find the pure file name in ptri_from */
    }
    else {
      /* ptri_to consists of directory and filename */
      xmlChar *ptr_release;
      for (ptr_basename=ptr_sep=BAD_CAST ptri_to; ptr_sep!=NULL && *ptr_sep!=(xmlChar)'\0'; ptr_sep++) {
	/* */
	if ((*ptr_sep==(xmlChar)'/' || *ptr_sep==(xmlChar)'\\') && *(ptr_sep+1)!=(xmlChar)'\0') {
	  ptr_basename = ptr_sep;
	}
      }
      ptr_release = xmlStrndup(ptri_to,((char *)ptr_basename - ptri_to));
      mkdir_pathUTF8(ptr_release,MODE_DIR_CREATE);
      xmlFree(ptr_release);
    }

    if ((ofp=fopen(ptri_to, "wb")) == NULL) {
      printf_log(1,"Can't create '%s'", ptri_to);
      return FALSE;
    }
  }

  if ((ifp=fopen(ptri_from, "rb")) == NULL) {
    if (ofp != stdout) fclose(ofp);
    printf_log(1,"Cant open '%s'", ptri_from);
    return FALSE;
  }

  while ((nmemb=fread(buf, 1, sizeof(buf), ifp)) > 0) {
    if (fwrite(buf, 1, nmemb, ofp) != nmemb) {
      printf_log(1,"fwrite failed");
      fclose(ifp);
      if (ofp != stdout) fclose(ofp);
      return FALSE;
    }
  }

  fclose(ifp);
  if (ofp != stdout) fclose(ofp);

  /*  xmlFree(ptri_to); */

  return TRUE;
}
	

/*! creates a given path string as directories in filesystem

  \param path relative or absolute path and name of directory
  \param mode argument for mkdir()
  \return error code of mkdir()
 */
BOOL_T
mkdir_pathUTF8(const xmlChar *ptr_path, const int mode)
{
  xmlChar *ptr, *ptr_s, *ptr_start, sep_prev, *ptr_path_norm;
  BOOL_T flag_sep;		/* flag for a found directory separator */
  int e = 0;
  char buffer_in[BUFFER_LENGTH];
  char buffer_path[BUFFER_LENGTH];
  char *ptri_path;
  char *ptr_in;
  size_t int_in;
  size_t int_path;
  size_t nconv;

  if (isreadonly()) {
    /* we are running in readonly mode */
    printf_log(1,"Cant create directories in readonly mode");
    return FALSE;
  }

  if (ptr_path == NULL
      || xmlStrlen(ptr_path) < 1
      || xmlStrcmp(ptr_path,BAD_CAST"-")==0  /* stdin */
      || xmlStrstr(ptr_path,BAD_CAST"http://")==ptr_path /* web access */
      || (ptr_path_norm = file_normalize(ptr_path,NULL)) == NULL
      ) {
    printf_log(2, "No valid directory path '%s'", ptr_path);
    return FALSE;
  }

  /*!\todo also other types of URI */

  if (cd.to.fs) {
    /* 
       conversion of from path
     */
    strncpy(buffer_in,(char *)ptr_path_norm,BUFFER_LENGTH - 1);
    int_in = strlen(buffer_in) + 1;
    ptr_in  = buffer_in;
    int_path = BUFFER_LENGTH - 1;
    ptri_path = buffer_path;

    nconv = iconv(cd.to.fs,&ptr_in,&int_in,&ptri_path,&int_path);
    iconv_error(errno,nconv);
  }
  else {
    strncpy(buffer_path,(char *)ptr_path_norm,BUFFER_LENGTH - 1);
  }
  xmlFree(ptr_path_norm);

  ptr = BAD_CAST buffer_path;

  if (isalpha(ptr[0]) && ptr[1]==':' && (ptr[2]=='\\' || ptr[2]=='/')) {
    /* skip leading drive char */
    ptr += 2;
  }
#ifdef _MSC_VER
  ptr_start = BAD_CAST buffer_path;
#else
  ptr_start = BAD_CAST ptr; 		/* ignore windows drives */
#endif
  if (*ptr=='\\' || *ptr=='/') { /* ignore leading slash */
    ptr++;
  }

  for ( ; ptr && *ptr; ptr++) {
    /* search for the next path separator */
    if ((flag_sep = ((ptr_s=(unsigned char*)strchr((const char*)ptr,'/'))
		     || (ptr_s=(unsigned char*)strchr((const char*)ptr,'\\'))))
	|| (ptr_s=(unsigned char*)buffer_path+strlen((const char*)buffer_path))) {
      
      if (flag_sep) {
	sep_prev = *ptr_s;
	*ptr_s = '\0';
      }
      
      if (dir_is_readableUTF8(ptr_start)) {
	/* OK */
      }
      else {
	printf_log(2,"MKDIR '%s'", ptr_start);
#ifdef _MSC_VER
	e = _mkdir((const char*)ptr_start);
#else
	e = mkdir((const char*)ptr_start,mode);
#endif
	printf_exit(e,45,"Cant create NAME_DIR '%s'", ptr_start);
      }

      if (flag_sep) {
	*ptr_s = sep_prev;
	ptr = ptr_s;
      }
      else {
	return TRUE;
      }
    }
  }
  assert(TRUE);
  return FALSE;
}


/*! test a node if \@view is set to "yes" and call the named file or directory
*/
BOOL_T
test_and_viewUTF8(xmlNodePtr n)
{
  xmlChar *attr_name;
  xmlChar *attr_view;
  xmlChar buffer_call[BUFFER_LENGTH];
  char buffer_in[BUFFER_LENGTH];
  char buffer_out[BUFFER_LENGTH];
  char *ptr_out, *ptr_in;
  char *ptr_value;
  xmlChar *ptr_result = NULL;
  size_t int_in;
  size_t int_out;
  size_t nconv;


  if ((attr_view = attribute_str(n,BAD_CAST "view")) != NULL
       && xmlStrcasecmp(attr_view,BAD_CAST "yes")==0
      && runmode != cgi
      && (((attr_name = attribute_str(n,BAD_CAST "name")) != NULL
	   && xmlStrcasecmp(attr_name, BAD_CAST "-") != 0)
	  || (attr_name = attribute_str(n,BAD_CAST "chdir")) != NULL
	  || (attr_name = attribute_str(n,BAD_CAST "mkdir")) != NULL
	  )) {
    /* 
       asynchronous call for viewing
    */
    unsigned char buffer_path[BUFFER_LENGTH];
    xmlChar *attr_path;
    xmlChar *path_call = file_normalize(attr_name,NULL);

    /*!\todo use an environment variable for format string */

    if ((attr_path = attribute_str(n,BAD_CAST "path")) != NULL
	&& xmlStrlen(attr_path) > 0) {
      /* append the anchor/path to the filename */
      xmlChar *ptr_release = path_call;
      path_call = xmlStrncatNew(path_call,BAD_CAST "#",2);
      xmlFree(ptr_release);
      ptr_release = path_call;
      path_call = xmlStrncatNew(path_call,attr_path,xmlStrlen(attr_path));
      xmlFree(ptr_release);
      printf_log(3,"Concat Path '%s'", path_call);
    }

#ifdef _MSC_VER
    xmlStrPrintf(buffer_call,
		 BUFFER_LENGTH,
		 BAD_CAST "explorer.exe \"%s\" & ",
		 path_call);
#else
    xmlStrPrintf(buffer_call,
		 sizeof(buffer_call),
		 "firefox -remote 'openFile(%s, new-tab)'",
		 path_call);
#endif
    xmlFree(path_call);

    printf_log(2,"Call '%s' for VIEW", buffer_call);
    /*!\bug shell in system() encode '#' to '%23' */


    if (xmlStrlen(buffer_call) > 0) {
      xmlChar *ptr_call = buffer_call;
      if (cd.to.env) {
	/* convert name of environment variable */
	
	strncpy(buffer_in,(char *)ptr_call,BUFFER_LENGTH - 1);
	int_in = strlen(buffer_in) + 1;
	ptr_in  = buffer_in;
	int_out = BUFFER_LENGTH - 1;
	ptr_out = buffer_out;
	
	nconv = iconv(cd.to.env,&ptr_in,&int_in,&ptr_out,&int_out);
	iconv_error(errno,nconv);
      }
      else {
	/* no conversion necessary */
	strncpy(buffer_out,(char *)buffer_call,BUFFER_LENGTH - 1);
      }
      system(buffer_out);
    }
  }
  return TRUE;
}
/* end of test_and_view() */


/*! read a whole file into an allocated buffer and convert to UTF-8 encoding

\param file_name

\return pointer to read buffer or NULL
*/
xmlChar *
fgetUTF8(xmlChar *file_name)
{
  xmlChar *ptr_result = NULL;
  xmlChar *str_file;
  long int size_file;
  FILE *file_input;

  str_file = file_normalize(file_name,NULL);
  if (type_is_readableUTF8(str_file,FALSE,&size_file)
      && size_file > 0
      && (file_input = fopen((char *)str_file,"r"))!=NULL) {
    /* OK */
    unsigned char *str_input;
    int input_length;

    str_input = xmlMalloc(size_file+1);
    input_length = fread((void *)(str_input),sizeof(char),size_file,file_input);
    if (ferror(file_input)) {
      /*!\todo why is input_length != size_file with VC++ */
      printf_log(1,"Read error '%s'", str_file);
    }
    else {
      str_input[input_length] = '\0';
      if (xmlCheckUTF8(BAD_CAST str_input)) {
	/* is UTF-8 encoding already */
	ptr_result = str_input;
      }
      else {
	/* dynamic buffer */
	xmlChar *str_output = xmlMalloc(size_file * 12 / 10);
	long int outlen;
	long int inlen = (long int)strlen((const char*)str_input);
	long int l = isolat1ToUTF8(str_output,&outlen,BAD_CAST str_input,&inlen);
	if (l < 1) {
	  printf_log(1,"Problem with content");
	}
	else {
	  str_output[l] = (xmlChar)'\0';
	  ptr_result = str_output;
	}
	xmlFree(str_input);
      }
    }
  }
  else {
    ptr_result = NULL;
  }
  if (str_file) xmlFree(str_file);

  return ptr_result;
}


/*! get a text paragraph from file

\param *mode set to '1' if marked as preformated
\param *mode_type of reading mode (par, line or csv)
\param stream

accepts mixed encoding UTF-8/ISO-8859-1 input data

\todo use plain/\@encoding and iconv() and return an UTF-8 result string

\return pointer to read buffer or NULL at end of file
*/
xmlChar *
fgetparUTF8(int *mode, int *mode_type, int size, FILE *stream)
{
  char *ptr_collect;		/*! pointer to collection buffer */
  char *ptr_in;		/*! pointer to input buffer for fgets() */
  char *ptr_read;		/*! pointer to read buffer for fgets() */
  int i, j;
  int i_line;			/*! counter for lines */
  int length_collect;		/*! counter for collected string length */
  int size_collect;		/*! size of collection buffer */
  xmlChar *ptr_result;
  BOOL_T flag_skip;

  *mode = 0;	     /* default: no special mode */

  ptr_in = NULL;
  ptr_collect = NULL;
  length_collect = 0;
  size_collect = 0;
  ptr_result = NULL;
  for ( flag_skip=FALSE, i_line=0; ; i_line++) {

    if (ptr_collect == NULL || size_collect - length_collect < BUFFER_LENGTH) {
      /*
	inrease the size of buffer
      */
      char *ptr_new;
      size_collect += 2 * BUFFER_LENGTH * sizeof(char);
      /* if (ptr_collect) printf_log(2,"Resize text input buffer to '%i' Byte", size_collect); */
      ptr_new = (char *) xmlRealloc((void *)ptr_collect,size_collect);
      printf_exit(ptr_new==NULL,77,"Not enough memory");
      /* if (ptr_new != ptr_collect) xmlFree(ptr_collect); */
      ptr_collect = ptr_new;
      ptr_in = ptr_collect + length_collect;
    }

    ptr_read = fgets(ptr_in,(size_collect - length_collect - 3),stream);

    if (ptr_read==NULL) {
      if (i_line > 0) {
	/* there was no empty line at file end */
        length_collect = chomp(ptr_collect);
      }
      break;
    }
    else if (strstr(ptr_in,"#begin_of_skip")==ptr_in) {
      /* handle skip markup in input file */
      flag_skip = TRUE;
    }
    else if (strstr(ptr_in,"#end_of_skip")==ptr_in) {
      i_line = -1;         /* reset line counter */
      length_collect = 0;
      flag_skip = FALSE;
    }
    else if (flag_skip==TRUE) {
      /* skip this marked block */
    }
    else if (strstr(ptr_in,"#begin_of_pre")==ptr_in) {
      /* handle preformated markup */
      *mode = 1;
    }
    else if (strstr(ptr_in,"#end_of_pre")==ptr_in) {
      *ptr_in = '\0';
      break;
    }
    else if (*mode == 1) {
      /* append */
      length_collect += strlen(ptr_in);
      ptr_in = &(ptr_collect[length_collect]);
    }
    else {
      /* 
	 normal input line
      */
      int l = chomp((unsigned char *)ptr_in);
      if (l > 0) {
	if (*mode_type==1 || *mode_type==3) {
	  /* return line oriented results */
	  length_collect = chomp(ptr_collect);
	  break;
	}
        ptr_in += l + 1;
	*ptr_in = ' ';
	ptr_in++;
	*ptr_in = '\0';
	length_collect = strlen(ptr_collect);
      }
      else {
	/* empty line */
	if (*mode_type == 3) {
	  /* empty lines in csv mode are table separators */
	  return BAD_CAST ptr_collect;
	} 
	else if (i_line < 1) {
	  /* ignore first emtpy line, its not an paragraph */
	  i_line = -1;         /* reset line counter */
	  length_collect = 0;
	}
	else {
	  length_collect = chomp(ptr_collect);
	  break;
	}
      }
    }
  }

  /*
    remove all control chars and redundant spaces
  */
  if (*mode==1) {
    /* this par is preformatted */
  }
  else if (*mode_type==3) {
    /* this par is comma separated */
  }
  else if (ptr_collect) {
    for (ptr_in=ptr_collect, i=0, j=0; ptr_in[i]; i++) {
      if (isspace(ptr_in[i])) {
	if (j && isspace(ptr_in[j-1])) {
	}
	else {
	  ptr_in[j++] = ' ';
	}
      }
      else {
	ptr_in[j++] = ptr_in[i];
      }
    }
    if (j && isspace(ptr_in[j-1])) {
      j--;	/* cut last space */
    }
    ptr_in[j] = '\0';
  }

  /* 
     do encoding conversions
  */
  if (length_collect > 0) {
    /*!\todo detect encoding from very first three bytes and set_encodingsUTF8() */
    if (cd.from.plain && !xmlCheckUTF8(ptr_collect)) {
      size_t int_out;
      size_t int_in;       
      size_t nconv;
      char *ptr_out;
      char *ptr_in;

      int_in = length_collect + 2;
      int_out = length_collect * 2;
      ptr_result = xmlMalloc(int_out * sizeof(char));
      printf_exit(ptr_result==NULL,78,"Not enough memory");

      ptr_out = (char *) ptr_result;
      ptr_in  = ptr_collect;
      nconv = iconv(cd.from.plain,&ptr_in,&int_in,&ptr_out,&int_out);
      iconv_error(errno,nconv);
      xmlFree(ptr_collect);
    }
    else {
      /* no conversion necessary */
      ptr_result = BAD_CAST ptr_collect;
    }
  }
  else {
    xmlFree(ptr_collect);
    printf_log(3, "Empty line");
  }

  return ptr_result;
}
/* end of fgetparUTF8() */


#ifdef TESTCODE

/*! test code for this module 
*/
int
test_module_iconv(void)
{
    xmlChar *ptr;
    int n_ok;
    int i;

    n_ok=0;
    i=0;

    if (TRUE) {
      /*
       */
      FILE *in;
      int mode = 0;
      int mode_type = 2;
      xmlChar *ptr_in;

      i++;
      if ((in = fopen("test-pie-3.txt","r"))) {
	while ((ptr_in = fgetparUTF8(&mode,&mode_type,BUFFER_LENGTH-1,in))) {
	  fprintf(stderr,"%s\n\n",ptr_in);
	  xmlFree(ptr_in);
	}
	fclose(in);
      }
      else {
	printf_log(1,"Error opening file: %i",i);
      }
    }

    /*
    */
    i++;
#ifdef _MSC_VER
    if (mkdir_pathUTF8("c:/tmp/aaaa/ooo/kkk",(0))==TRUE) {
#else
    if (mkdir_pathUTF8("/tmp/aaaa/ooo/kkk",(S_IRUSR | S_IWUSR | S_IXUSR))==TRUE) {
#endif
	n_ok++;
    }
    else {
	printf_log(1,"Error mkdir_pathUTF8(): %i",i);
    }

    /*
    */
    i++;
    if (file_copyUTF8(BAD_CAST "Validate.xsl",BAD_CAST "/tmp")==TRUE) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_copyUTF8(): %i",i);
    }

    /*
    */
    i++;
    if (file_copyUTF8(file_search(NULL,BAD_CAST "Validate.xsl"),BAD_CAST "/tmp/deep/dir/structure/")==TRUE) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_copyUTF8(): %i",i);
    }

    /*
    */
    i++;
    if (file_copyUTF8(file_search(NULL,BAD_CAST "pie.css"),BAD_CAST "/tmp")==TRUE) {
	n_ok++;
    }
    else {
	printf_log(1,"Error file_copyUTF8(): %i",i);
    }

    printf_log(1,"Test: %i/%i OK in '%s'", n_ok,i,__FILE__);

    return (n_ok - i);
}
#endif
