/*
  cxproc - Configurable Xml PROCessor

  Copyright (C) 2006,2007,2008 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#ifdef _MSC_VER
#include "config-win32.h"
#else
#include <unistd.h>
#include "../config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* 
 */
#include <libxml/parser.h>

#include "basics.h"
#include "cxp_dtd.h"
#include "plain_x.h"
#include "dom_x.h"
#include "tools.h"

#ifdef HAVE_LIBMAGICK
#include <magick/ImageMagick.h>
#include "image_x.h"

#define ThrowWandException(wand) \
{ \
  char \
    *description; \
 \
  ExceptionType \
    severity; \
 \
  description=MagickGetException(wand,&severity); \
  (void) fprintf(stderr,"%s %s %lu %s\n",GetMagickModule(),description); \
  description=(char *) MagickRelinquishMemory(description); \
  exit(-1); \
}

#endif


/*!
 */
#ifdef HAVE_LIBMAGICK
Image *
process_image(xmlNodePtr node_path)
{
  ExceptionInfo *exception;
  ImageInfo *image_info, *write_info;
  Image *image, *orient_image;

  MagickBooleanType status;

  xmlChar *attr_src;
  xmlChar *attr_to;

  printf_log(1,"Processing IMAGE");

  exception=AcquireExceptionInfo();
  image_info=AcquireImageInfo();

  attr_src = attribute_str(node_path,BAD_CAST "src");
  attr_to  = attribute_str(node_path,BAD_CAST "to");

  /*!\todo allow annotation of image (s. ImageMagick-6.4.3/coders/label.c) */

  if ((attr_src != NULL && xmlStrlen(attr_src) > 0)) {
    /*
      Read an image file.
    */
    printf_log(2,"Read IMAGE '%s'",attr_src);
    CopyMagickString(image_info->filename,(char *)attr_src,MaxTextExtent);
    image=ReadImage(image_info,exception);
    CatchException(exception);
    /*!\todo Error handling */

    DestroyExceptionInfo(exception);
    DestroyImageInfo(image_info);

    if (image == NULL) {
      printf_log(1,"Read Error");
      return NULL;
    }

    return image;
  }
  else if ((attr_to != NULL && xmlStrlen(attr_to) > 0)) {

    xmlNodePtr node_plain;
    if ((node_plain = child_element(node_path,NAME_PLAIN))) {
      /* there is a simple comment text */

      xmlChar *result_str = NULL;
      FILE *file_comment;
      xmlChar *name_file = xmlStrdup(attr_to);
      xmlChar *p;

      printf_log(2,"Comment IMAGE '%s'",attr_to);

      /* derive name of comment file from image file */
      for ( p = name_file + xmlStrlen(name_file); p > name_file && *p != (xmlChar)'.'; p-- ) {
	;
      }
      *(p+1) = (xmlChar)'\0';
      result_str = xmlStrcat(name_file,BAD_CAST "txt");

      if (result_str != NULL && (file_comment = fopen((char *)result_str,"w")) != NULL) {
 	char *r = (char *)process_plain(node_plain);
	/*\todo remove line breaks etc. normalize_space() ? */

	if (fputs(r,file_comment) == EOF) {
	  printf_log(1,"Write error file '%s'", result_str);
	}
	else {
	  printf_log(2,"Wrote comment file '%s'", result_str);
	}
	fclose(file_comment);
	xmlFree(r);
      }
      else {
	printf_log(1,"Error opening comment file '%s'", result_str);
      }
      xmlFree(name_file);
      DestroyExceptionInfo(exception);
      DestroyImageInfo(image_info);

      return NULL;
    }
    else {
      /*
	Write the image then destroy it.
      */
      xmlChar *attr_frame;
      unsigned long width_frame, height_frame;
      float scale;

      image = process_image(child_element(node_path,NAME_IMAGE));
      if (image == NULL) {
	return NULL;
      }

      height_frame = 0;
      width_frame  = 0;
      attr_frame = attribute_str(node_path,BAD_CAST "frame");
      if (attr_frame) {
	char *ptr_end;
	width_frame = (unsigned long) strtol((const char *)attr_frame, &ptr_end, 10);
	if (*ptr_end=='x') {
	  ptr_end++;
	  height_frame = (unsigned long) strtol((const char *)ptr_end, &ptr_end, 10);
	  printf_log(2,"Frame for IMAGE '%ix%i'",width_frame,height_frame);
	}
	else {
	}
      }

#if 0
      /* s. magick/constitute.c:513 */
      value=GetImageProperty(image,"exif:Orientation");
      if (value != (char *) NULL)
	{
	  image->orientation=(OrientationType) atol(value);
	  /* (void) DeleteImageProperty(image,"exif:Orientation"); */
	}
#endif
      /*!\bug in ImageMagick-6.3.7: "exif:Orientation" is NULL */

      scale = 1.1;
      if (image->orientation == RightTopOrientation
	  || image->orientation == LeftBottomOrientation)	{
	/* this image will be rotated 90 or 270 deg later
	 */
	if (image->columns > image->rows) {
	  /* landscape */
	  if (height_frame > 0 && image->columns > height_frame) {
	    scale = (float)height_frame / (float)image->columns;
	  }
	}
	else {
	  /* portrait */
	  if (width_frame > 0 && image->rows > width_frame) {
	    scale = (float)width_frame / (float)image->rows;
	  }   
	}
      }
      else {
	if (image->columns > image->rows) {
	  /* landscape */
	  if (width_frame > 0 && image->columns > width_frame) {
	    scale = (float)width_frame / (float)image->columns;
	  }
	}
	else {
	  /* portrait */
	  if (height_frame > 0 && image->rows > height_frame) {
	    scale = (float)height_frame / (float)image->rows;
	  }   
	}
      }

      if (scale < 1.0) {
	Image
	  *resize_image;
	printf_log(1,"Scale IMAGE '%.2f'",scale);
	resize_image = ResizeImage(image,
				   (unsigned long) ceil((double)image->columns * scale),
				   (unsigned long) ceil((double)image->rows * scale),
				   BoxFilter,
				   1.0,
				   exception);

	if (resize_image) {
	  DestroyImage(image);
	  image = resize_image;
	}
      }

      /* MagickCommentImage(magick_wand,"Tescht"); */
      /*
	DrawComment(magick_wand,"Tescht");
	DrawRender(magick_wand);
      */

      printf_log(2,"Write IMAGE to '%s'",attr_to);

      /* s. wand/mogrify.c:351 */
      orient_image=NewImageList();
      switch (image->orientation)
	{
	case TopRightOrientation:
	  {
	    orient_image=FlopImage(image,exception);
	    break;
	  }
	case BottomRightOrientation:
	  {
	    orient_image=RotateImage(image,180.0,exception);
	    break;
	  }
	case BottomLeftOrientation:
	  {
	    orient_image=FlipImage(image,exception);
	    break;
	  }
	case LeftTopOrientation:
	  {
	    orient_image=TransposeImage(image,exception);
	    break;
	  }
	case RightTopOrientation:
	  {
	    orient_image=RotateImage(image,90.0,exception);
	    break;
	  }
	case RightBottomOrientation:
	  {
	    orient_image=TransverseImage(image,exception);
	    break;
	  }
	case LeftBottomOrientation:
	  {
	    orient_image=RotateImage(image,270.0,exception);
	    break;
	  }
	default:
	  printf_log(2,"No rotation operation neccessary");
	  break;
	}

      if (attr_to[0]==(xmlChar)'-' && attr_to[1]==(xmlChar)'\0') {
	/* write to stdout */
    
	if (runmode == cgi) {
	  printf("Content-type: image/%s\n\n", image->magick);
	}

	write_info=CloneImageInfo(image_info);
	SetImageInfoFile(write_info,stdout);
	write_info->adjoin=MagickTrue;

	if (orient_image) {

	  status=WriteImage(write_info,orient_image);
	  write_info=DestroyImageInfo(write_info);
	  if (status == MagickFalse)
	    InheritException(exception,&orient_image->exception);
	  DestroyImage(orient_image);
	}
	else {

	  status=WriteImage(write_info,image);
	  write_info=DestroyImageInfo(write_info);
	  if (status == MagickFalse)
	    InheritException(exception,&orient_image->exception);
	}
      }
      else {
	if (orient_image) {
	  WriteImages(image_info,orient_image,(char *)attr_to,exception);
	  DestroyImage(orient_image);
	}
	else {
	  WriteImages(image_info,image,(char *)attr_to,exception);
	}
      }

      DestroyImageInfo(image_info);
      DestroyImage(image);

      exception=DestroyExceptionInfo(exception);
      MagickCoreTerminus();
    }

  }
  else {
    printf_log(1,"No valid attributes in '%s'",node_path->name);
  }
  return NULL;
}
#endif

