/*
  cxproc - Configurable Xml PROCessor

  Copyright (C) 2006,2007,2008 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#ifdef _MSC_VER
#include "config-win32.h"
#else
#include <unistd.h>
#include "../config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <assert.h>
#include <errno.h>

/* 
 */
#include <libxml/xmlversion.h>
#include <libxml/HTMLtree.h>
#include <libxml/parser.h>

#ifdef LIBXML_THREAD_ENABLED
#include <libxml/globals.h>
#include <libxml/threads.h>
#if HAVE_LIBPTHREAD
#include <pthread.h>
#endif
#endif

/* 
 */
#include "cgi-util.h"

#include <pcre.h>

/* 
 */
#include <libxslt/xslt.h>
#include <libxslt/xsltutils.h>
#include <libxslt/transform.h>
#include <libxslt/variables.h>

#include <iconv.h>

/* 
 */
#include "basics.h"
#include "iconv_x.h"
#include "tools.h"
#include "dom_x.h"
#include "dir_x.h"
#include "make_x.h"
#include "plain_x.h"
#include "calendar_x.h"
#include "petrinet_x.h"
#include "cxp_dtd.h"
#include "sql_x.h"
#ifdef HAVE_LIBMAGICK
#include <magick/ImageMagick.h>
#include "image_x.h"
#endif

#define RE_EACH "([^ \\,\\;]+)"

static pcre *re_each  = NULL;


static char **
process_param(xmlNodePtr n);

static  BOOL_T
modify_param(xmlDocPtr result, char **param);

static xmlDocPtr
process_each(xmlNodePtr n);

static void
substitute_each(xmlNodePtr node, xmlChar *name, xmlChar *value);

static BOOL_T
process_subst(xmlNodePtr n);

static BOOL_T
substitute_subst(const xmlNodePtr node, const xmlChar *ptr_from, const xmlChar *ptr_to);

static BOOL_T
substitute_substdom(xmlNodePtr node, const xmlChar *ptr_from, xmlNodePtr node_new);

static xmlDocPtr
process_info(xmlNodePtr node_info);

static void
usage (void);


/* for global access
 */
char **envp_i = NULL;
char **argv_i = NULL;
int    argc_i = -1;

xmlNsPtr ns_cxp = NULL;
const xmlChar *ptr_cxp = BAD_CAST "http://www.tenbusch.info/cxproc";

runmode_t runmode;

static xmlChar *str_filename_cxp = NULL;

/*!\todo merge with today structs in calendar_x.c */
static xmlChar *today_year = NULL;
static xmlChar *today_mon = NULL;
static xmlChar *today_week = NULL;
static xmlChar *today_yday = NULL;
static xmlChar *today_mday = NULL;
static xmlChar *today_hour = NULL;
static xmlChar *today_min = NULL;
static xmlChar *today_sec = NULL;



/*! release allocated memory of static variables
*/
void
free_make(void)
{
  /*!\todo release other static variables */
  if (str_filename_cxp) {
    xmlFree(str_filename_cxp);
  }
}
/* end of free_make() */


/*! prints a short usage message
 */
void
usage (void)
{
  fprintf(stderr,"%s - a Configurable Xml PROCessor (version %s)\n\n", argv_i[0], VERSION);

  fprintf(stderr,"Usage:\n");

  fprintf(stderr,"    plain mode: %s [pie2html.xsl] < test.txt\n", argv_i[0]);

  fprintf(stderr,"      dir mode: %s .\n", argv_i[0]);

  fprintf(stderr," calendar mode: %s 2007[..2008]\n", argv_i[0]);

  fprintf(stderr,"   config mode: %s config.cxp\n", argv_i[0]);

  fprintf(stderr,"     test mode: %s -t [config.cxp|2008|c:\\temp]\n", argv_i[0]);

  fprintf(stderr,"     info mode: %s -i\n", argv_i[0]);

  fprintf(stderr,"\n");

  fprintf(stderr," PATHXSL=/usr/local/share/cxproc\n\n");

  fprintf(stderr," CXP_LOG=2\n\n");

  fprintf(stderr,"Documentations, Copyrights, Examples see %s\n\n", ptr_cxp);

  /*!\todo complete the help message */
}


/*! process the main() arguments into a config DOM

  \param argc argument from main()
  \param argv argument from main()
  \return a config DOM according argv
*/
xmlDocPtr
process_argv(int argc, char **argv, char **envp)
{
  xmlChar *pathfilename_config = NULL;
  xmlChar *ptr;
  xmlChar *ptr_search_env;
  xmlChar *ptr_search_path;
  xmlChar *ptr_search_pwd;
  xmlDocPtr doc_result = NULL;
  xmlNodePtr node_dir;
  xmlNodePtr node_make;
  xmlNodePtr node_pie;
  xmlNodePtr node_plain;
  xmlNodePtr node_xml;
  xmlNodePtr node_xsl;

  argc_i = argc;
  argv_i = argv;
  envp_i = envp;

  ptr_search_pwd = getcwdUTF8();
  assert(ptr_search_pwd!=NULL);

  printf_log(1,"Its me '%s' %s in '%s'", argv[0], VERSION, ptr_search_pwd);

  doc_result = xmlNewDoc(BAD_CAST "1.0");
  node_make = xmlNewDocNode(doc_result, NULL, NAME_MAKE, NULL); 
  xmlDocSetRootElement(doc_result,node_make);
  doc_result->encoding = xmlStrdup(BAD_CAST "UTF-8"); /* according to conversion in append_input() */

  ns_cxp = xmlNewNs(node_make, ptr_cxp, BAD_CAST "cxp");
  xmlSetNs(node_make,ns_cxp);

  /* 
    some tricks to set the default search path
  */
  ptr_search_env = ptr_search_path = NULL;

  /* 1.) use environment variable "PATHXSL"
  */
  if ((ptr = getenvUTF8(BAD_CAST "PATHXSL"))) {
    /*!\todo check some conventions like drive + directory: searchpath must contain one directory name (not "c:\") */
    ptr_search_env = xmlStrdup(ptr);
    xmlFree(ptr);
    printf_log(2,"Use value '%s' of environment variable 'PATHXSL'", ptr_search_env);
  }
  else {
#if 1
    int length_dir;
    xmlChar buffer_conv[BUFFER_LENGTH];
    xmlChar *ptr_path, *ptr_sep;

    /* 2.) find "cxproc" in argv[0]
     */
    if ((ptr = (xmlChar *) xmlStrcasestr(BAD_CAST argv[0], BAD_CAST"cxproc"))) {
      length_dir = ptr - BAD_CAST argv[0] + 6;
      assert(length_dir > 0 && length_dir < BUFFER_LENGTH);
      strncpy((char *)buffer_conv,argv[0],length_dir);
      buffer_conv[length_dir] = '\0';
      if (dir_is_readableUTF8(buffer_conv)==TRUE) {
	ptr_search_path = buffer_conv;
	printf_log(2,"Use found directory '%s' in '%s'", ptr_search_path, argv[0]);
      }
    }
    else {
      /* 3.) find dirname of argv[0] 
       */
      for (ptr=BAD_CAST argv[0] + strlen(argv[0]); ptr != BAD_CAST argv[0]; ptr--) {
	if (*ptr == '\\') {
	  length_dir = ptr - BAD_CAST argv[0];
	  assert(length_dir > 0 && length_dir < BUFFER_LENGTH);
	  strncpy((char *)buffer_conv,argv[0],length_dir);
	  buffer_conv[length_dir] = '\0';
	  if (dir_is_readableUTF8(buffer_conv)==TRUE) {
	    ptr_search_path = buffer_conv;
	    printf_log(2,"Use prefix directory '%s' of '%s'", ptr_search_path, argv[0]);
	  }
	}
      }
      /* 4.) find "cxproc" in %PATH 
       */
      if (!ptr_search_path && (ptr_path = getenvUTF8((const xmlChar *) "PATH")) && (ptr= (xmlChar *) xmlStrcasestr((const xmlChar *)ptr_path,BAD_CAST"cxproc"))) {
	for (ptr_sep=ptr; ptr_sep >= ptr_path; ptr_sep--) {
	  if (*ptr_sep == ';' || ptr_sep == ptr_path) {
	    length_dir = (ptr + 6) - ptr_sep;
	    assert(length_dir > 0 && length_dir < BUFFER_LENGTH);
	    strncpy((char *)buffer_conv,(const char *)((*ptr_sep == ';') ? ptr_sep+1 : ptr_sep),length_dir);
	    buffer_conv[length_dir] = '\0';
	    if (dir_is_readableUTF8(buffer_conv)) {
	      ptr_search_path = buffer_conv;
	      printf_log(2,"Use value '%s' in environment variable 'PATH'", ptr_search_path);
	    }
	  }
	}
      }
    }
#else
    /* 2.) find "cxproc" in default installation directory
     */
    /*!\todo use configured value from "../config.h" or -D in Makefile.am ? */
    if (ptr_search_path==NULL &&
	dir_is_readableUTF8(BAD_CAST "/usr/local/share/cxproc")==TRUE) {
      ptr_search_path = BAD_CAST "/usr/local/share/cxproc";
    }
    if (ptr_search_path==NULL
	&& dir_is_readableUTF8(BAD_CAST "/usr/share/cxproc")==TRUE) {
      ptr_search_path = BAD_CAST "/usr/share/cxproc";
    }
#endif
  }

  if (ptr_search_env) {
    /* use dir from PATHXSL recursive only */
    xmlSetProp(node_make, BAD_CAST "pathxsl", ptr_search_env);
    xmlFree(ptr_search_env);
  }
  else if (ptr_search_path) {
#if 0
    if (xmlStrstr(ptr_search_pwd,ptr_search_path)) {
      /* ptr_search_pwd contains ptr_search */
      xmlSetProp(node_make, BAD_CAST "pathxsl", ptr_search_path);
    }
    else {
      /* add ptr_search_pwd and ptr_search_path */
      xmlChar *str_pathxsl = xmlStrdup(ptr_search_pwd);
      str_pathxsl = xmlStrcat(str_pathxsl,BAD_CAST ";");
      str_pathxsl = xmlStrcat(str_pathxsl,ptr_search_path);
      xmlSetProp(node_make, BAD_CAST "pathxsl", str_pathxsl);
    }
#else    
    xmlSetProp(node_make, BAD_CAST "pathxsl", ptr_search_path);
#endif
  }
  xmlFree(ptr_search_pwd);

  /* early search path because check for CXP files before calling */
  xmlSetProp(node_make, BAD_CAST "verbosity", BAD_CAST "1");
  make_set_searchpath(node_make);

#ifdef TESTCODE
  printf_log(1,"");
  printf_log(1,"TESTCODE");
  if (test_module_dir() != 0 && test_module_iconv() != 0) {
    exit(EXIT_FAILURE);
  }
  printf_log(1,"");
#endif

  if (argc > 1) {
    /*! build internal DOMs 
     */
    xmlChar *ptr_argv_1 = getargvUTF8(1);
    xmlChar *ptr_argv_1_norm;
    xmlChar *ptr_argv_1_ext;
    int j = 1;

    if (xmlStrcasecmp(ptr_argv_1,BAD_CAST"-?")==0) {
      usage();
      return NULL;
    }
    else if (xmlStrcasecmp(ptr_argv_1,BAD_CAST"-p")==0) {
      /* pause with prompt and dummy input */
      char buffer[10];
      printf_log(1,"PAUSE, Type <ENTER> to continue");
      fgets(buffer,sizeof(buffer),stdin);
      j++;
      ptr_argv_1 = getargvUTF8(j);
    }
    else if (xmlStrcasecmp(ptr_argv_1,BAD_CAST"-t")==0) {
      xmlNodePtr node_copy = xmlCopyNode(node_make,1);
      
      node_xml = xmlNewChild(node_make, NULL, NAME_XML, NULL);
      xmlSetProp(node_xml, BAD_CAST "name", BAD_CAST "-");
      xmlSetProp(node_xml, BAD_CAST "eval", BAD_CAST "no");

      node_make = xmlAddChild(node_xml,node_copy);

      j++;
      ptr_argv_1 = getargvUTF8(j);
    }
    else if (xmlStrcasecmp(ptr_argv_1,BAD_CAST"-i")==0) {
      node_xml = xmlNewChild(node_make, NULL, NAME_XML, NULL);
      xmlSetProp(node_xml, BAD_CAST "name", BAD_CAST "-");
      xmlNewChild(node_xml, NULL, NAME_INFO, NULL);
      return doc_result;
    }
    ptr_argv_1_norm = file_normalize(ptr_argv_1,NULL);
    ptr_argv_1_ext  = file_extension(ptr_argv_1);
    if (dir_is_readableUTF8(ptr_argv_1_norm)) {
      /*! argv[1] is a valid name of a directory
       */
      int i;

      node_xml = xmlNewChild(node_make, NULL, NAME_XML, NULL);
      xmlSetProp(node_xml, BAD_CAST "name", BAD_CAST "-");

      node_dir = xmlNewChild(node_xml, NULL, NAME_DIR, NULL);
/*       xmlSetProp(node_dir, BAD_CAST "verbosity", BAD_CAST "4"); */
      for (i=j; i< argc; i++) {
	xmlChar *argv_ii = file_normalize(getargvUTF8(i),NULL);

	if (xmlStrcasecmp(file_extension(argv_ii),BAD_CAST"xsl")==0) {
	  node_xsl = xmlNewChild(node_xml, NULL, NAME_XSL, NULL);
	  xmlSetProp(node_xsl, BAD_CAST "name", BAD_CAST argv_ii);
	}
	else if (dir_is_readableUTF8(argv_ii)==TRUE) {
	  printf_log(1,"Try DIR '%s'", argv_ii);
	  node_pie = xmlNewChild(node_dir, NULL, NAME_DIR, NULL);
	  xmlSetProp(node_pie, BAD_CAST "name", BAD_CAST argv_ii);
	}
	else {
	}
      }
    }
    else if ((xmlStrcasecmp(ptr_argv_1_ext,BAD_CAST "cxp")==0)
	     && 
	     ((file_is_readableUTF8((pathfilename_config = ptr_argv_1)))==TRUE
	      || 
	      (file_is_readableUTF8((pathfilename_config = ptr_argv_1_norm)))==TRUE
	      || 
	      (pathfilename_config = file_search(NULL,ptr_argv_1)) != NULL)) {
      /*
	argv[1] is a valid name of a config XML
      */
	  node_xml = xmlNewChild(node_make, NULL, NAME_XML, NULL);
	  xmlSetProp(node_xml, BAD_CAST "name", pathfilename_config);
	  xmlSetProp(node_xml, BAD_CAST "eval", BAD_CAST "yes");
      }
      else if (ptr_argv_1 != NULL && atoi((char *)ptr_argv_1) > 1900) {
	/*
	  argv[1] is a valid year number, calendar
	*/
	xmlNodePtr node_calendar;

	node_xml = xmlNewChild(node_make, NULL, NAME_XML, NULL);
	xmlSetProp(node_xml, BAD_CAST "name", BAD_CAST "-");
	node_calendar = xmlNewChild(node_xml, NULL, NAME_CALENDAR, NULL);
	xmlSetProp(node_calendar, BAD_CAST "year", ptr_argv_1);
      }
      else {
	/*
	  Plain text filter for stdin and XSL names in argv[]
	*/
	int j;
	node_xml = xmlNewChild(node_make, NULL, NAME_XML, NULL);
	xmlSetProp(node_xml, BAD_CAST "name", BAD_CAST "-");

	node_pie = xmlNewChild(node_xml, NULL, NAME_PIE, NULL);
	xmlSetProp(node_pie, BAD_CAST "url", BAD_CAST "yes");
	xmlSetProp(node_pie, BAD_CAST "figure", BAD_CAST "yes");

	node_plain = xmlNewChild(node_pie, NULL, NAME_PLAIN, NULL);
	xmlSetProp(node_plain, BAD_CAST "name", BAD_CAST "-");

	/* collect the named XSL 
	 */
	for (j=1; j<argc; j++) {
	  xmlChar *ptr_argv_j;

	  ptr_argv_j = getargvUTF8(j);
	  if (xmlStrcasecmp(file_extension(ptr_argv_j),BAD_CAST "xsl")==0) {
	    node_xsl = xmlNewChild(node_xml, NULL, NAME_XSL, NULL);
	    xmlSetProp(node_xsl, BAD_CAST "name", ptr_argv_j);
	  }
	  else {
	    /*!\todo handle additional types (XML, cxp, ...) */
	    printf_log(1,"Ignoring argument '%s'", ptr_argv_j);
	  }
	}
/* 	make_set_searchpath(NULL); */
      }
    if (ptr_argv_1)      xmlFree(ptr_argv_1);
    if (ptr_argv_1_norm) xmlFree(ptr_argv_1_norm);
    if (ptr_argv_1_ext)  xmlFree(ptr_argv_1_ext);
  }
  else {
    /*! without arguments work as a simple plain text to XML converter
     */
    node_xml = xmlNewChild(node_make, NULL, NAME_XML, NULL);
    xmlSetProp(node_xml, BAD_CAST "name", BAD_CAST "-");

    node_pie = xmlNewChild(node_xml, NULL, NAME_PIE, NULL);

    node_plain = xmlNewChild(node_pie, NULL, NAME_PLAIN, NULL);
    xmlSetProp(node_plain, BAD_CAST "name", BAD_CAST "-");
  }

  return doc_result;
}
/* end of process_argv() */


/*! process the CGI options into a config DOM

  \return a config DOM according CGI request
*/
xmlDocPtr
process_cgi(int argc, char **argv, char **envp)
{
  xmlChar buffer_cxp[BUFFER_LENGTH];
  xmlChar *cxp_str, *ptr_prefix;
  xmlChar *pathfilename_cxp;
  xmlChar *ptr_pathxsl;
  xmlDocPtr doc_result = NULL;
  xmlNodePtr node_make;
  xmlNodePtr node_plain;
  xmlNodePtr node_xml;

  argc_i = argc;
  argv_i = argv;
  envp_i = envp;

  printf_log(1,"Its me '%s' %s", argv[0], VERSION);

#if 0
  /* use Apache autorization instead */
  printf_exit((xmlStrcmp(getenvUTF8(BAD_CAST "REMOTE_ADDR"),BAD_CAST "127.0.0.1")!=0),32,"Local delivery allowed only");
#endif

  doc_result = xmlNewDoc(BAD_CAST "1.0");
  node_make = xmlNewDocNode(doc_result, NULL, NAME_MAKE, NULL); 
  xmlDocSetRootElement(doc_result,node_make);

  ns_cxp = xmlNewNs(node_make, ptr_cxp, BAD_CAST "cxp");
  xmlSetNs(node_make,ns_cxp);

  ptr_pathxsl = getenvUTF8(BAD_CAST "PATHXSL");
  printf_exit((ptr_pathxsl==NULL || xmlStrlen(ptr_pathxsl) < 1),
	      33,
	      "Variable 'PATHXSL' not set");
  /* use dir from PATHXSL recursive only */
  xmlSetProp(node_make, BAD_CAST "pathxsl", BAD_CAST ptr_pathxsl);

  /* early search path because check for CXP files before calling 
   */
  xmlSetProp(node_make, BAD_CAST "verbosity", BAD_CAST "1");
  make_set_searchpath(node_make); /* redundant to process_make() */
  /*!\todo delete properties to avoid a redundant later call of make_set_searchpath() */
/*   xmlFreePropList(node_make->properties); */

  /* default is readonly mode in CGI */
  xmlSetProp(node_make, BAD_CAST "readonly", BAD_CAST "yes");

  /* common path prefix for all CGI requests 
   */
  ptr_prefix = getenvUTF8(BAD_CAST "CXP_ROOT");
  printf_exit((ptr_prefix==NULL || xmlStrlen(ptr_prefix) < 1),
	      34,
	      "Variable 'CXP_ROOT' not set");

  cxp_str = getcgiUTF8(BAD_CAST"cxp");
  if (cgi_num_entries > 0 && cxp_str!=NULL && xmlStrlen(cxp_str) > 0) {
    xmlStrPrintf(buffer_cxp,BUFFER_LENGTH,BAD_CAST"Cgi%s.cxp",cxp_str);
  }
  else {
    /* xmlStrPrintf(buffer_cxp,BUFFER_LENGTH,BAD_CAST"Cgi%s.cxp",BAD_CAST"info"); */
    xmlStrPrintf(buffer_cxp,BUFFER_LENGTH,BAD_CAST"CgiDefault.cxp");
  }
  printf_log(2,"Search for '%s'",buffer_cxp);
  pathfilename_cxp = file_search(NULL,buffer_cxp);
  if (pathfilename_cxp) {
    printf_log(2,"Found '%s'",pathfilename_cxp);
    node_xml = xmlNewChild(node_make, NULL, NAME_XML, NULL);
    xmlSetProp(node_xml, BAD_CAST "name", pathfilename_cxp);
    xmlSetProp(node_xml, BAD_CAST "eval", BAD_CAST "yes");
  } 
  else {
    node_plain = xmlNewChild(node_make, NULL, NAME_PLAIN, BAD_CAST "\n\n\nThis file extension is not yet supported!");
    xmlSetProp(node_plain, BAD_CAST "name", BAD_CAST "-");
  }

  return doc_result;
}
/* end of process_cgi() */


/*! substitutes each occurance of 'name' in XML_ATTRIBUTE_NODE with value
 */
void
substitute_each(xmlNodePtr node, xmlChar *name, xmlChar *value)
{
  while (node) {
    if (node->type == XML_ELEMENT_NODE) {
      substitute_each((xmlNodePtr)node->properties, name, value);
      substitute_each(node->children, name, value);
    }
    else if (node->type == XML_TEXT_NODE) {
      /*!\todo substitution in text nodes */
    }
    else if (node->type == XML_ATTRIBUTE_NODE) {
      xmlChar *ptr_a;
      if (node->children!=NULL
	  && node->children->type == XML_TEXT_NODE
	  && node->children->content!=NULL
	  && (ptr_a = str_replace(node->children->content,name,value))!=NULL) {
	printf_log(3,"Pattern '%s' found in '%s' '%s' to '%s'", name, node->name, node->children->content, ptr_a);
	xmlNodeSetContent(node->children,ptr_a);
	xmlFree(ptr_a);
      }
    }
    else {
      /* ignore */
    }
    substitute_each(node->next, name, value);
    node = node->next;
  }
}
/* end of substitute_each() */


/*! iteration loop

build and returns a tree of MAKE including all single steps as node copies
*/
xmlDocPtr
process_each(xmlNodePtr n)
{
  xmlChar *attr_name;
  xmlChar *attr_thread;
  xmlChar *ptr_i;
  xmlChar *ptr_values;
  xmlDocPtr doc_result = NULL;
  xmlNodePtr node_cur;
  xmlNodePtr node_from;
  xmlNodePtr node_make;

  if (re_each==NULL) {
    int erroffset;
    const char *error;
    /* compile regexp first */
    printf_log(2,"Initialize Iterator regexp '%s'", RE_EACH);
    re_each = pcre_compile((const char*)RE_EACH,0,&error,&erroffset,NULL);
    /*!\todo error handling */
  }

  if (n!=NULL && n->type == XML_ELEMENT_NODE
      && (IS_NODE_EACH(n))
      && (attr_name = attribute_str(n,BAD_CAST "name"))!=NULL
      && (xmlStrlen(attr_name) > 0)
      && (node_from = child_element(n,NAME_FROM))!=NULL ) {
    if (node_from->children) {
      /*
	FROM element
      */
      int i;
      xmlChar *ptr_substr;
          
      doc_result = xmlNewDoc(BAD_CAST "1.0");
      node_make = xmlNewDocNode(doc_result, NULL, NAME_MAKE, NULL); 
      if ((attr_thread = attribute_str(n,BAD_CAST "thread"))) {
	xmlSetProp(node_make,BAD_CAST "thread",attr_thread);
      }

      xmlDocSetRootElement(doc_result,node_make);
      /* doc_result->encoding = xmlStrdup(BAD_CAST "ISO-8859-1"); */
      doc_result->encoding = xmlStrdup(BAD_CAST "UTF-8"); /* according to conversion in append_input() */

      /* with simple text node */
      ptr_values = process_plain(node_from);
      if (ptr_values == NULL) {
	printf_log(1,"No iteration keys for '%s'", attr_name);
	return NULL;
      }
      printf_log(2,"Iteration with '%s' over '%s'", attr_name, ptr_values);
      /* copy next value for iterator */

      for (ptr_substr=ptr_values, i=0; ptr_substr; i++) {
	int rc;
	int ovector[30];

	rc = pcre_exec(
		       re_each,        /* result of pcre_compile() */
		       NULL,           /* we didn't study the pattern */
		       (const char*)ptr_substr,  /* the subject string */
		       strlen((const char*)ptr_substr),             /* the length of the subject string */
		       0,              /* start at offset 0 in the subject */
		       0,              /* default options */
		       ovector,        /* vector of integers for substring information */
		       30);            /* number of elements (NOT size in bytes) */
	if (rc > -1) {
	  printf_log(3,"each %i..%i in '%s'", ovector[0],ovector[1],ptr_substr);
	  ptr_i = xmlStrndup(ptr_substr + ovector[0], ovector[1] - ovector[0]);
	  printf_exit(ptr_i==NULL,45,"Cant duplicate iterator");

	  printf_log(2,"Step '%s' with '%s'", attr_name, ptr_i);
	  for (node_cur = node_from->next; node_cur; node_cur = node_cur->next) {
	    /* copy all element nodes after FROM with childs */
	    if (node_cur->type == XML_ELEMENT_NODE) {
	      xmlNodePtr node_copy = xmlCopyNode(node_cur,1);
	      if (node_copy) {
		substitute_each(node_copy, attr_name, ptr_i);
		xmlAddChildList(node_make,node_copy);
	      }
	    }
	  }
	  xmlFree(ptr_i);
	  if (*(ptr_substr + ovector[1]) == '\0') {
	    /* there are no trailing chars */
	    ptr_substr = NULL;
	  }
	  else {
	    ptr_substr += ovector[1];
	  }
	}
	else {
	  ptr_substr = NULL;
	}
      }
      xmlFree(ptr_values);
      /* #error xmlFree() for xmlStrdup() ; xmlMemFree() for xmlMemMalloc() */
    }
    else {
      printf_log(1,"No valid iteration with '%s'", attr_name);
    }
  }
  else {
    printf_log(1,"No valid iteration EACH");
  }
  /*!\todo dump result to tmp */
  return doc_result;
}
/* end of process_each() */


/*! process the XML element and his childs
*/
xmlDocPtr
process_xml(xmlNodePtr n)
{
  BOOL_T mode_cache  = FALSE;
  char *filename_found = NULL;
  int options = XSLT_PARSE_OPTIONS;
  xmlChar *attr_cache;
  xmlChar *attr_eval;
  xmlChar *attr_file;
  xmlChar *attr_xpath;
  xmlChar *attr_search;
  xmlDocPtr doc_result = NULL;
  xmlDocPtr result_tmp = NULL; /* preliminary result DOM */

  assert(n);
  /*  */
  if (IS_NODE_XML(n) || IS_NODE_XHTML(n)) {

    /* attr_file   = file_normalize(attribute_str(n,BAD_CAST "name"),NULL); */
    attr_file   = attribute_str(n,BAD_CAST "name");
    attr_search = attribute_str(n,BAD_CAST "search");
    attr_eval   = attribute_str(n,BAD_CAST "eval");
    attr_cache  = attribute_str(n,BAD_CAST "cache");
    attr_xpath  = attribute_str(n,BAD_CAST "xpath");

    if (attr_xpath && (IS_NODE_XML(n->parent) || IS_NODE_XHTML(n->parent))) {
      /* add a xsl:stylesheet child to n->parent, !! side effect !! */
      xmlNsPtr ns_xsl;
      xmlNodePtr node_new_xsl;
      xmlNodePtr node_new;

      node_new_xsl = xmlNewChild(n->parent, NULL, BAD_CAST "stylesheet", NULL);
      /* create local namespace for XSL */
      ns_xsl = xmlNewNs(node_new_xsl,XSLT_NAMESPACE,BAD_CAST "xsl");
      xmlSetNs(node_new_xsl,ns_xsl);
      xmlSetProp(node_new_xsl, BAD_CAST "version", BAD_CAST "1.0");

      node_new = xmlNewChild(node_new_xsl,ns_xsl,BAD_CAST"output",NULL);
      xmlSetProp(node_new, BAD_CAST "method", BAD_CAST "xml");

      node_new = xmlNewChild(node_new_xsl,ns_xsl,BAD_CAST"template",NULL);
      xmlSetProp(node_new, BAD_CAST "match", BAD_CAST "/");
      /* create a root element for multiple selection results */
      node_new = xmlNewChild(node_new,ns_xsl,BAD_CAST"element",NULL);
      xmlSetProp(node_new, BAD_CAST "name", BAD_CAST"{name(child::*)}");
      node_new = xmlNewChild(node_new,ns_xsl,BAD_CAST"copy-of",NULL);
      xmlSetProp(node_new, BAD_CAST "select", attr_xpath);
    }

    if (xmlStrcasecmp(attr_eval,BAD_CAST "no")==0) {
      printf_log(3,"No evaluation");
      doc_result = xmlNewDoc(BAD_CAST "1.0");
      if (doc_result) {
	xmlNodePtr node_to_copy;
	xmlNodePtr node_copy = NULL;

	for (node_to_copy=n->children; node_to_copy; node_to_copy=node_to_copy->next) {
	  if (node_to_copy->type == XML_ELEMENT_NODE) {
	    /* only the very first element node is copied */
	    node_copy = xmlDocCopyNodeList(doc_result,node_to_copy);
	    break;
	  }
	}

	if (node_copy) {
	  xmlDocSetRootElement(doc_result, node_copy);
	  save_file(doc_result,NULL,n);
	}
	else {
	  printf_log(1,"DOM not usable");
	}
      }
    }
    else if (n->children) {
      /*
	this is a XML instruction
      */
      xmlNodePtr c;

#if 0
/*
every child of make node has to fit into the following pattern


name attribute

*/
if (
c->children && ((c->children->type == XML_ELEMENT_NODE && (node_process=c->children))
             || (c->children->next && c->children->next->type == XML_ELEMENT_NODE && (node_process=c->children->next)))

) {
}
else {
continue;
}
#error TODO
#endif

      if ((c = child_element(n,NAME_INFO))) {
	result_tmp = process_info(c);
      }
      else if ((c = child_element(n,NAME_XML))!=NULL || (c = child_element(n,NAME_XHTML))!=NULL) {
	result_tmp = process_xml(c);
      }
      else if ((c = child_element(n,NAME_DIR))) {
	result_tmp = process_dir(c);
      }
      else if ((c = child_element(n,NAME_PIE))) {
	result_tmp = process_pie(c);
      }
      else if ((c = child_element(n,NAME_PATHTABLE))!=NULL || (c = child_element(n,NAME_PATHNET))!=NULL) {
	result_tmp = process_path(c);
      }
      else if ((c = child_element(n,NAME_CALENDAR))) {
	result_tmp = process_calendar(c);
      }
      else if ((c = child_element(n,NAME_SQL))) {
	result_tmp = process_sql(c);
      }
      else {
	printf_log(1,"Ignoring '%s'", n->name);
      }
      /*!\todo process_subst(n); */
      doc_result = process_xsl(result_tmp,n);
      if (doc_result != result_tmp) {
	xmlFreeDoc(result_tmp);
      }

      if (xmlStrcasecmp(attr_eval,BAD_CAST "yes")==0) {
	xmlNodePtr node_root = xmlDocGetRootElement(doc_result);
	xmlDocPtr doc_release = doc_result;
	printf_log(1,"XML evaluation");
	/* recursive call for evaluation */
	doc_result = process_xml(node_root);
	xmlFreeDoc(doc_release);
      }
    }
    else if (attr_file) {
      /*
	this is a source XML
      */
      xmlChar *ptr_ext = file_extension(attr_file);
      xmlChar *attr_file_norm = file_normalize(attr_file,NULL);

      if (xmlStrcasecmp(attr_cache,BAD_CAST "yes")==0) {
	mode_cache = TRUE;
      }

      /*!\todo validate="yes" for XML validation on demand */

      if ((file_is_readableUTF8(attr_file_norm)==TRUE && (filename_found = (char *)attr_file_norm)!=NULL)
	       ||
	       (xmlStrcasecmp(attr_search,BAD_CAST "yes")==0
		&& (filename_found = (char *)file_search(NULL,attr_file))!=NULL)) {
	/*
	  this is an XML file (to include)
	*/

	if (xmlStrcasecmp(attribute_str(n,BAD_CAST "include"),BAD_CAST "yes")==0) {
	  /* this is a simple include instruction, dont change file context */
	  printf_log(1,"Include XML DOM from '%s'", filename_found);
	}
	else {
	  printf_log(1,"Reading XML '%s'", filename_found);
	  if (str_filename_cxp) {
	    /* release former string */
	    xmlFree(str_filename_cxp);
	  }
	  str_filename_cxp = xmlStrdup(BAD_CAST filename_found);
	}

        if ((doc_result = get_dom(BAD_CAST filename_found))) {
	  /* there is a cached DOM */
	  doc_result = xmlCopyDoc(doc_result,1); /* copy this DOM, ready for later transformations */
        }
        else if (IS_NODE_XML(n)) {
	  options = XML_PARSE_NOENT | XML_PARSE_NOWARNING | XML_PARSE_NOCDATA;
	  doc_result = xmlReadFile((char *)filename_found, NULL, options);
	}
	else {
	  /* this is a source XHTML */
	  doc_result = htmlReadFile((char *)filename_found, NULL, 0);
	}

	if (mode_cache==TRUE) {
	  set_dom(xmlCopyDoc(doc_result,1),BAD_CAST filename_found); /* ready for later transformations */
	}
      }
      else {
	/*!\bug 'https://', use sscanf_next_url() ? */
	printf_log(1,"Not readable '%s'", attr_file);
      }
      xmlFree(attr_file_norm);
      if (ptr_ext) xmlFree(ptr_ext);
    }
    else {
      printf_log(1,"No valid make instruction");
    }
  }
  return doc_result;
}
/* end of process_xml() */


/*! 
 */
static void *
process_thread(void *node_private)
{
  const xmlNodePtr node_thread = (xmlNodePtr) node_private;

  process_make(node_thread);

  return NULL;
}
/* end of process_thread() */


/*! process the childs of top element MAKE (non-recursive!)

\todo Validate against DTD if readable locally
\todo configurable Processing-Instructions?
 */
BOOL_T
process_make(xmlNodePtr n)
{
  xmlNodePtr c;
  xmlNodePtr node_xsl;
  xmlChar *attr_name;
  xmlChar *attr_thread;
#if HAVE_LIBPTHREAD
  BOOL_T flag_thread = TRUE;
  int n_threads, i_threads;
  pthread_t *pool = NULL;
#endif

  assert(n);

  if (IS_NODE_MAKE(n) || IS_NODE_THREAD(n)) {
    /*
      MAKE level, there is no result DOM
    */
    if ((attr_name = attribute_str(n,BAD_CAST "readonly"))) {
      if (xmlStrcasecmp(attr_name,BAD_CAST "yes")==0) {
	/* set a global readonly mode, no files are written, file names
	   are only tags for later access */
	set_readonly(TRUE);
      }
      else {
	/* reset value */
	set_readonly(FALSE);
      }
    }
    if ((attr_name = attribute_str(n,BAD_CAST "log"))) {
      set_level_log(atoi((char *)attr_name));
    }
    /*!\todo set the max number of threads via attribute THREADS="11" */
#if HAVE_LIBPTHREAD
    if ((attr_thread = attribute_str(n,BAD_CAST "thread"))!=NULL 
	&& xmlStrcasecmp(attr_thread,BAD_CAST "0")==0) {
      flag_thread = FALSE;
      printf_log(1,"Threads disabled");
    }
#endif
    make_set_searchpath(n);
    /*!\todo substitute format strings in text nodes */
    process_subst(n);

#if HAVE_LIBPTHREAD
    i_threads = 0;
    n_threads = number_of_child(n,NAME_THREAD);
    if (flag_thread==TRUE && n_threads > 0) {
      printf_log(1,"Using %i threads", n_threads);
      pool = (pthread_t *)xmlMemMalloc(n_threads * sizeof(pthread_t));
      printf_exit(pool == NULL,55,"no mem");
    }
#endif

    for (c=n->children; c; c=c->next) {

#if 0
      if (number_of_child(c,NULL) < 1) {
	/* ignore element nodes without node childs */
      }
      else
#error cxp:xml eval="yes"
#endif
      if (IS_NODE_EACH(c)) {
	xmlDocPtr doc_make;
	if (attr_thread) {
	  xmlSetProp(c,BAD_CAST "thread",attr_thread);
	}
	doc_make = process_each(c);
	if (doc_make) {
	  process_make(xmlDocGetRootElement(doc_make));
	  xmlFreeDoc(doc_make);
	}
      }
      else if (IS_NODE_MAKE(c)) {
	if (attr_thread) {
	  xmlSetProp(c,BAD_CAST "thread",attr_thread);
	}
	process_make(c);
      }
      else if (IS_NODE_THREAD(c)) {
#if HAVE_LIBPTHREAD
	/* create a new thread */
	int ret;
	if (flag_thread==TRUE) {
	  /*!\todo check if exeeds */
	  assert(i_threads>-1 && i_threads<n_threads);
	  printf_log(2,"Create thread %i", i_threads);
	  ret = pthread_create(&pool[i_threads], NULL, process_thread, (void *)c);
	  printf_exit(ret != 0,55,"pthread_create %i", i_threads);
	  i_threads++;
	}
	else {
	  process_make(c);
	}
#else
	/* simple recursive call if no threads are available */
	process_make(c);
#endif
      }
      else if (IS_NODE_XML(c) || IS_NODE_XHTML(c)) {
	xmlDocPtr r_transformed, r;
	xmlNodePtr node_root;

	r = process_xml(c);
	if (r) {
	  if (xmlStrcasecmp(attribute_str(c,BAD_CAST "eval"),BAD_CAST "yes")==0) {
	    node_root = xmlDocGetRootElement(r);
	    /* root element must be MAKE */
	    if (IS_NODE_MAKE(node_root)) {
	      make_set_searchpath(node_root);
	      process_subst(node_root);
	      if ((node_xsl = child_element(node_root,NAME_XSL))) {
		printf_log(1,"Self transforming");
		r_transformed = process_xsl(r,node_root);
		if (r_transformed == NULL) {
		  printf_log(1,"No successful self evaluating");
		}
		else if (r_transformed != r) {
		  xmlFreeDoc(r);
		  r = r_transformed;
		  /* xmlSaveFormatFileEnc("debug-config-evaled.cxp", r, NULL, 1); */
		}		  
	      }
	      printf_log(1,"Self evaluating");
	      node_root = xmlDocGetRootElement(r);
	      if (node_root == NULL) {
		printf_log(1,"Empty DOM");
	      }
	      else {
		process_make(xmlDocGetRootElement(r));
	      }
	    }
	    else {
	      printf_log(1,"Cant EVAL non MAKE elements");
	    }
	  }
	  /* we are not interested in any result at MAKE level */
	  xmlFreeDoc(r); 
	}
	test_and_viewUTF8(c);
      }
      else if (IS_NODE_IMAGE(c)) {
#ifdef HAVE_LIBMAGICK
	process_image(c);
#else
	printf_log(1,"Cant process images (compiled without this option)");
#endif
      }
      else if (IS_NODE_PLAIN(c)) {
	xmlChar *r;
	r = process_plain(c);
	if (r != NULL) {
	  xmlFree(r);
	}
	test_and_viewUTF8(c);
      }
      else if (IS_NODE_FILECOPY(c)) {
	file_copyUTF8(c);
      }
      else if (IS_NODE_SYSTEM(c)) {
	xmlChar *r;
	/*!\todo move code to process_system() */

	r = attribute_str(c,BAD_CAST "message");
	if (r != NULL) {
	  /*!\todo convert r to isolat1? */
	  fputs((const char*) r,stderr);
	  fputs("\n",stderr);
	}

	r = attribute_str(c,BAD_CAST "call");
	/*!\todo restricted to existing shell script files? */
	if (r != NULL) {
	  /* this is a system shell call */
	  /*!\todo convert r to isolat1? */
	  /*!\todo set environment with all env childs */
	  system((const char*) r);
	}
	else {
	  if ((r = attribute_str(c,BAD_CAST "chdir"))) {
	    /* this is a system chdir call */
	    /*!\todo convert r to isolat1? */
	    int e;
	    printf_log(1,"CD to '%s'", r);
#ifdef _MSC_VER
	    e = _chdir((const char*) r);
#else
	    e = chdir((const char*) r);
#endif
	    /*!\todo check some conventions like root directory */
	    printf_exit(e,45,"Cant change to '%s'",r);
	  }
	  else if ((r = attribute_str(c,BAD_CAST "mkdir"))) {
	    /* this is a system mkdir call */
	    /*!\todo convert r to isolat1? */
#ifdef _MSC_VER
#define MODE_DIR_CREATE (0)
#else
#define MODE_DIR_CREATE (S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH)
#endif
	    /*!\todo use umask for MODE_DIR_CREATE */
	    mkdir_pathUTF8(r,MODE_DIR_CREATE);
	    test_and_viewUTF8(c);
	  }
	}
      }
    }
#if HAVE_LIBPTHREAD
    if (flag_thread==TRUE && n_threads > 0) {
      /* join all created threads of this MAKE element */
      int i;
      assert(i_threads==n_threads);
      printf_log(1,"Joining %i threads",n_threads);
      for (i=0; i<n_threads; i++) { 
	int ret;
	ret = pthread_join(pool[i], NULL);
	printf_exit(ret != 0,56,"pthread_join %i", i);
      }
      xmlMemFree(pool);
    }
#endif
    return TRUE;
  }
  else {
    printf_log(1,"Unusable Element '%s'", n->name);
    return FALSE;
  }
}
/* end of process_make() */


/*! apply all sibling PARAM elements, non recursive!!
*/
char **
process_param(xmlNodePtr n)
{
  xmlNodePtr c_param;
  xmlNodePtr node_plain;
  char **result = NULL;
  char *attr_select;
  index_t n_v;
  index_t i;

  if (n == NULL) {
    return result;
  }

/*   dump_dom("dump.dot",n,9); */

  n_v = number_of(n->children, BAD_CAST "variable", 0);
  if (n_v > 0) {
    result = (char **)xmlMemMalloc((n_v * 2 + 1) * sizeof(char *));
    result[0] = NULL;		/* end markup */
    for (i=0, c_param=n->children; result!=NULL && c_param!=NULL; c_param=c_param->next) {
      /* all PARAM siblings */
      if (IS_NODE_VARIABLE(c_param)) {
	/* this is an PARAM */
	char *str_name;

	str_name = (char *)attribute_str(c_param,BAD_CAST "name");
	if (str_name!=NULL && strlen(str_name) > 0) {
	  char *str_value;
	  printf_log(3,"VARIABLE '%s'", str_name);
	  str_value = NULL;
	  if ((node_plain = child_element(c_param,NAME_PLAIN))) {
	    /* there are child nodes (cxp:plain only!!), read as value */
	    str_value = (char *)process_plain(node_plain);
	  }
	  else if ((attr_select = (char *)attribute_str(c_param,BAD_CAST "select")) && xmlStrlen(BAD_CAST attr_select) > 0) {
	    /* use attribute value */
	    str_value = attr_select;
	  }
	  else if (c_param->children!=NULL
		   && c_param->children->content!=NULL
		   && xmlStrlen(c_param->children->content) > 0) {
	    /* use text node */
	    str_value = (char *)c_param->children->content;
	  }
	  else {
	    str_value = NULL;
	    printf_log(1,"Ignoring empty VARIABLE '%s'", str_name);
	  }

	  if (str_value) {
	    result[i]   = str_name;
	    result[i+1] = str_value;
	    i += 2;
	  }
	}
      }
      result[i] = NULL;		/* end markup */
    }
  }
  return result;
}
/* end of process_param() */


/*! global XSL variables only (between RootElement and first xsl:template element)
*/
BOOL_T
modify_param(xmlDocPtr result, char **param)
{
  xmlNodePtr node_c;
  xmlNodePtr node_root = xmlDocGetRootElement(result);
  index_t i;

  for (i=0; param[i]; i+=2) {	/* all name/value pairs */
    for (node_c=node_root->children; node_c; node_c=node_c->next) {
      if (node_c->type == XML_ELEMENT_NODE
	  && xmlStrcasecmp(node_c->ns->prefix,BAD_CAST "xsl")==0) {
	if (xmlStrcasecmp(node_c->name,BAD_CAST "variable")==0) {
	  xmlChar *attr_name = attribute_str(node_c,BAD_CAST "name");
	  if (attr_name!=NULL && xmlStrcasecmp(attr_name,BAD_CAST param[i])==0) {
	    /* variable exists already */
	    printf_log(2,"Modifying VARIABLE '%s' to '%s'", attr_name, param[i+1]);
	    xmlSetProp(node_c, BAD_CAST "select", BAD_CAST param[i+1]);
	    break;
	  }
	} else if (xmlStrcasecmp(node_c->name,BAD_CAST "template")==0) {
	  /* the first xsl:template element is reached, insert this new variable */
	  xmlNodePtr node_variable;
	  printf_log(2,"Inserting VARIABLE '%s' with value '%s'", param[i], param[i+1]);
	  node_variable = xmlNewNode(node_c->ns,BAD_CAST "variable");
	  xmlSetProp(node_variable, BAD_CAST "name", BAD_CAST param[i]);
	  xmlSetProp(node_variable, BAD_CAST "select", BAD_CAST param[i+1]);
	  /* insert the xsl:variable after xsl:import and xsl:output
	     -> before first xsl:template
	  */
	  xmlAddPrevSibling(node_c,node_variable);
	  break;
	}
      }
    }
  }
  return TRUE;
}
/* end of modify_param() */


/*! apply all sibling XSL elements of node_parent to sibling_result, non recursive!!

  returns sibling_result if no transformations are done, else a new created xmlDocPtr

  \todo use DOM caching for XSL
*/
xmlDocPtr
process_xsl(const xmlDocPtr sibling_result, const xmlNodePtr node_parent)
{
  BOOL_T mode_cache  = FALSE;
  char **param;
  xmlChar *attr_cache;
  xmlChar *filename;
  xmlChar *filename_xsl = NULL;
  xmlDocPtr doc_tmp;	/* preliminary result DOM */
  xmlDocPtr doc_xsl = NULL;
  xmlDocPtr result = sibling_result;
  xmlNodePtr c;
  xmlNodePtr c_copy;
  xmlNodePtr c_xsl;
  xsltStylesheetPtr xsl = NULL;

  if (sibling_result == NULL || node_parent == NULL) {
    printf_log(1,"Empty result");
    return NULL;
  }

  for (c_xsl=node_parent->children; result!=NULL && c_xsl!=NULL; c_xsl=c_xsl->next) {
    /* all XSL siblings */
    if (c_xsl->type == XML_ELEMENT_NODE) {

      if (xmlStrcasecmp(c_xsl->name,BAD_CAST "stylesheet")==0
	  && c_xsl->ns!=NULL
	  && c_xsl->ns->prefix!=NULL
	  && xmlStrcasecmp(c_xsl->ns->prefix,BAD_CAST "xsl")==0) {
	/* this is a xsl:stylesheet tree */
	doc_xsl = xmlNewDoc(BAD_CAST "1.0");
	if (doc_xsl) {
	  printf_log(1,"Applying MAKE internal xsl:stylesheet");
	  c_copy = xmlDocCopyNodeList(doc_xsl,c_xsl);
	  if (c_copy == NULL) {
	    printf_log(1,"xsl:stylesheet not usable");
	    continue;
	  }
	  xmlDocSetRootElement(doc_xsl, c_copy);
	}
      }
      else if (IS_NODE_XSL(c_xsl)) {
	xmlChar *attr_file = attribute_str(c_xsl,BAD_CAST "name");
        xmlChar *attr_file_norm = file_normalize(attr_file,NULL);
        char *filename_found = NULL;

	if ((attr_cache = attribute_str(c_xsl,BAD_CAST "cache"))!=NULL
	    && xmlStrcasecmp(attr_cache,BAD_CAST "yes")==0) {
	  mode_cache = TRUE;
	}

	if ((c = child_element(c_xsl,NAME_XML))) {
	  /* the XSL DOM has to be transformed first */
	  doc_xsl = process_xml(c);
	}
	else if (attr_file!=NULL) {
	  xmlChar *ptr_release = NULL;
	  if ((file_is_readableUTF8(attr_file_norm)==TRUE && (filename_found = (char *)attr_file_norm)!=NULL)
	      ||
	      (((filename_found = (char *)file_search(NULL,attr_file))!=NULL) && (ptr_release = BAD_CAST filename_found)!=NULL)) {

	    int options = XML_PARSE_NOENT | XML_PARSE_NOWARNING | XML_PARSE_NOCDATA;

	    printf_log(2,"Applying '%s'", filename_found);
	    if ((doc_xsl = get_dom(BAD_CAST filename_found))) {
	      /* there is a cached DOM */
	      doc_xsl = xmlCopyDoc(doc_xsl,1); /* copy this DOM, ready for later transformations */
	    }
	    else if ((doc_xsl = xmlReadFile((char *)filename_found, NULL, options))) {
	      if (mode_cache==TRUE) {
		set_dom(xmlCopyDoc(doc_xsl,1),BAD_CAST filename_found); /* ready for later transformations */
	      }
	    }
	    else {
	      printf_log(1,"Not found '%s'", attr_file);
	      return NULL;
	    }
	    if (ptr_release) xmlFree(ptr_release);
	  }
	  else {
	    /*!\bug 'https://', use sscanf_next_url() ? */
	    printf_log(1,"Not readable '%s'", attr_file);
	  }
	  xmlFree(attr_file_norm);
	}
	else {
	  printf_log(1,"No file attribute");
	  return NULL;
	}

	if (doc_xsl == NULL) {
	  printf_log(1,"No DOM");
	  return NULL;
	}
      }
      else {
	/* other tag */
	continue;
      }

      param = process_param(c_xsl);
      if (param) {
	/*!\todo handle param[] as Global Vars ? (but see comment at libxslt/variables.c:947) */
	modify_param(doc_xsl,param);
      }

      if (xsl) {
	/* release former stylesheets */
	xsltFreeStylesheet(xsl);
      }
      xsl = xsltParseStylesheetDoc(doc_xsl);
      /*!\bug xsltParseStylesheetDoc() doesnt set xsl->omitXmlDeclaration correct */
      if (xsl==NULL) {
	xmlFreeDoc(doc_xsl);
	printf_exit(TRUE,44,"Cant parse this Stylesheet");
      }

      /*!\bug check rules
	- 'xsl->method' must be xml or html,
	exception: at last sibling 'text' is allowed if parent is PLAIN 
      */

      /* apply the stylesheet, check the result
       */
      doc_tmp = xsltApplyStylesheet(xsl, result, (const char **)param);

      if (param)
	xmlMemFree(param);

      if (result != sibling_result) {
	/* release the previous result DOM */
	xmlFreeDoc(result);
	result = NULL;
      }
      if (doc_tmp != NULL) {
	if ((xsl->method == NULL) || (xmlStrEqual(xsl->method, BAD_CAST "text")==1)) {
	  /* dump the result into a buffer and return 
	   */
	  xmlChar *result_ptr;
	  int l;
	  xsltSaveResultToString(&result_ptr,&l,doc_tmp,xsl);
	  xsltFreeStylesheet(xsl);
	  xmlFreeDoc(doc_tmp);
	  if (l > 0) {
	    return (xmlDocPtr) result_ptr; /*!\todo not nice!! */
	  }
	  return NULL;
	  /* return after first text output, not reachable */
	}
	else {
	  /* DOM result doc_tmp */
	  if (xmlDocGetRootElement(doc_tmp) == NULL) {
	    printf_log(1,"No content with this Stylesheet");
	    xmlFreeDoc(doc_tmp);
	  }
	  else {
	    result = doc_tmp;
	  }
	}
      }
      else {
	printf_log(1,"No result with Stylesheet '%s'",filename_xsl);
	xsltFreeStylesheet(xsl);
	return NULL;
      }
    }
  }

  if (result!=NULL && (filename = attribute_str(node_parent,BAD_CAST "name"))!=NULL) {
    save_file(result,NULL,node_parent);
  }
  if (xsl) {
    /* release former stylesheets */
    xsltFreeStylesheet(xsl);
  }

  return result;
}
/* end of process_xsl() */


/*! apply all substitution 

\todo substitutions in text nodes also
*/
BOOL_T
substitute_subst(const xmlNodePtr node, const xmlChar *ptr_from, const xmlChar *ptr_to)
{
    xmlChar *str_new;

    if (node==NULL) {
	return FALSE;
    }

    if (node->type == XML_TEXT_NODE) {
	if (node->content) {
	    /*!\todo reduce multiple xmlNodeSetContent() */
	    str_new = str_replace(node->content,ptr_from,ptr_to);
	    if(str_new) {
		printf_log(3,"Old content '%s' new content '%s'", node->content, str_new);
		xmlNodeSetContent(node,str_new);
		xmlFree(str_new);
	    }
	}
    }
    else {
	assert(node->type == XML_ELEMENT_NODE 
	    || node->type == XML_ATTRIBUTE_NODE
	    || node->type == XML_TEXT_NODE
	    || node->type == XML_COMMENT_NODE );
	substitute_subst((xmlNode *)node->properties,ptr_from,ptr_to);
	substitute_subst(node->children,ptr_from,ptr_to);
    }
    substitute_subst(node->next,ptr_from,ptr_to);

    return TRUE;
}
/* end of substitute_subst() */


/*! apply all substitution 

\todo substitutions in text nodes also
*/
BOOL_T
substitute_substdom(xmlNodePtr node, const xmlChar *ptr_from, xmlNodePtr node_new)
{
  xmlChar *str_new;

  if (node==NULL || node_new==NULL) {
    return FALSE;
  }

  if (node->type == XML_TEXT_NODE) {
    if (node->content) {
      /*!\todo reduce multiple xmlNodeSetContent() */
      str_new = (xmlChar *) xmlStrstr(node->content,ptr_from);
      if(str_new) {
	xmlReplaceNode(node,xmlCopyNodeList(node_new));
	node = node_new;
      }
    }
  }
  else {
    assert(node->type == XML_ELEMENT_NODE 
	   || node->type == XML_ATTRIBUTE_NODE
	   || node->type == XML_TEXT_NODE
	   || node->type == XML_COMMENT_NODE );
    substitute_substdom(node->children,ptr_from,node_new);
  }
  substitute_substdom(node->next,ptr_from,node_new);

  return TRUE;
}
/* end of substitute_substdom() */


/*! 
 */
void
free_subst(void)
{
  if (today_year) {
    /* release strings */
    xmlFree(today_year);
    xmlFree(today_mon);
    xmlFree(today_week);
    xmlFree(today_yday);
    xmlFree(today_hour);
    xmlFree(today_min);
    xmlFree(today_sec);
  }
}

/*! process the childs of top element MAKE (non-recursive!)
 */
BOOL_T
process_subst(xmlNodePtr n)
{
  xmlNodePtr c;

  if (today_year==NULL) {
    /* initialize */
    struct tm today;
    xmlChar buffer[BUFFER_LENGTH];

    /* mit Systemzeit initialisieren */
    time(&system_zeit);
    memcpy(&today, localtime(&system_zeit), sizeof(struct tm));

    xmlStrPrintf(buffer,BUFFER_LENGTH, BAD_CAST "%i", today.tm_year + 1900);
    today_year = xmlStrdup(buffer);

    xmlStrPrintf(buffer,BUFFER_LENGTH, BAD_CAST "%02i", today.tm_mon + 1);
    today_mon = xmlStrdup(buffer);

    xmlStrPrintf(buffer,BUFFER_LENGTH, BAD_CAST "%02i", week_in_year(today.tm_mday, today.tm_mon + 1, today.tm_year + 1900));
    today_week = xmlStrdup(buffer);

    xmlStrPrintf(buffer,BUFFER_LENGTH, BAD_CAST "%02i", today.tm_yday + 1);
    today_yday = xmlStrdup(buffer);

    xmlStrPrintf(buffer,BUFFER_LENGTH, BAD_CAST "%02i", today.tm_mday);
    today_mday = xmlStrdup(buffer);

    xmlStrPrintf(buffer,BUFFER_LENGTH, BAD_CAST "%02i", today.tm_hour);
    today_hour = xmlStrdup(buffer);

    xmlStrPrintf(buffer,BUFFER_LENGTH, BAD_CAST "%02i", today.tm_min);
    today_min = xmlStrdup(buffer);

    xmlStrPrintf(buffer,BUFFER_LENGTH, BAD_CAST "%02i", today.tm_sec);
    today_sec = xmlStrdup(buffer);
  }

  if (IS_NODE_MAKE(n) || IS_NODE_THREAD(n)) {
    for (c=n->children; c; c=c->next) {
      if (IS_NODE_SUBST(c)) {
	xmlNodePtr node_subst = c;
	xmlChar *ptr_name;
	/*!\todo use text node content also */

	if ((ptr_name = attribute_str(node_subst, BAD_CAST "string"))) {
	  xmlNodePtr node_root;
	  xmlNodePtr node_xml;
	  xmlNodePtr node_plain;
	  xmlDocPtr doc_result;
	  xmlChar *ptr_value;
	  xmlChar *str_argv;
	  xmlChar *str_cgi;
	  xmlChar *str_default;
	  xmlChar *str_dir;
	  xmlChar *str_file;
	  xmlChar *str_filename;
	  xmlChar *str_env;
	  xmlChar *str_name;
	  xmlChar *str_now;
	  xmlChar *str_to;

	  str_name  = xmlStrdup(ptr_name);

	  if ((ptr_value = attribute_str(node_subst, BAD_CAST "default"))) {
	    /* detect a possible default value */
	    str_default = xmlStrdup(ptr_value);
	  }
	  else {
	    str_default = NULL;
	  }

	  if ((ptr_value = attribute_str(node_subst, BAD_CAST "to"))) {
	    /* this is a substitution with a simple string */
	    str_to = xmlStrdup(ptr_value);
	  }
	  else if ((node_plain = child_element(node_subst, NAME_PLAIN))) {
	    /* there is a plain children element */
	    str_to = process_plain(node_plain);
	  }
	  else {
	    str_to = NULL;
	  }

	  if ((ptr_value = attribute_str(node_subst, BAD_CAST "dir"))) {
	    /* 
	       this is a substitution with current directories
	    */
	    if (xmlStrcasecmp(ptr_value, BAD_CAST "pwd")==0) {
	      str_dir = getcwdUTF8();
	    } 
	    else if (xmlStrcasecmp(ptr_value, BAD_CAST "ofFile")==0) {
	      if (str_filename_cxp) {
                str_dir = file_basedir(str_filename_cxp);
       	      }
       	      else {
       	        printf_log(1, "No filename registered");
		str_dir = NULL;
	      }
	    } 
	    else {
	      /*!\todo dir attributes "ofCurrentFile" or "ofCurrentWork" */
	      str_dir = NULL;
	    }
	  }
	  else {
	    str_dir = NULL;
	  }

	  if ((ptr_value = attribute_str(node_subst, BAD_CAST "file"))
	      && xmlStrlen(ptr_value) > 0) {
	    /* 
	       this is a substitution with file content
	    */
	    if (xmlStrEqual(ptr_value,BAD_CAST".") && str_filename_cxp) {
	      str_file = str_filename_cxp;
	    }
	    else {
	      str_file = ptr_value;
	    }
	  }
	  else {
	    str_file = NULL;
	  }

	  if ((ptr_value = attribute_str(node_subst, BAD_CAST "filename"))
	      && xmlStrlen(ptr_value) > 0) {
	    /* 
	       this is a substitution with file content
	    */
	    if (xmlStrEqual(ptr_value,BAD_CAST".") && str_filename_cxp) {
	      str_filename = str_filename_cxp;
	    }
	    else {
	      str_filename = ptr_value;
	    }
	  }
	  else {
	    str_filename = NULL;
	  }

	  if ((ptr_value = attribute_str(node_subst, BAD_CAST "now"))) {
	    /* 
	       this is a substitution with current time values
	    */
	    if (xmlStrcasecmp(ptr_value, BAD_CAST "Year")==0) {
	      str_now = xmlStrdup(today_year);
	    } 
	    else if (xmlStrcasecmp(ptr_value, BAD_CAST "NumOfMonth")==0) {
	      str_now = xmlStrdup(today_mon);
	    } 
	    else if (xmlStrcasecmp(ptr_value, BAD_CAST "NumOfWeek")==0) {
	      str_now = xmlStrdup(today_week);
	    } 
	    else if (xmlStrcasecmp(ptr_value, BAD_CAST "NameOfMonth")==0) {
	      /*\bug substitute NameOfMonth */
	      str_now = xmlStrdup(today_mon);
	    } 
	    else if (xmlStrcasecmp(ptr_value, BAD_CAST "Day")==0) {
	      str_now = xmlStrdup(today_yday);
	    } 
	    else if (xmlStrcasecmp(ptr_value, BAD_CAST "DayOfMonth")==0) {
	      str_now = xmlStrdup(today_mday);
	    } 
	    else if (xmlStrcasecmp(ptr_value, BAD_CAST "Hour")==0) {
	      str_now = xmlStrdup(today_hour);
	    } 
	    else if (xmlStrcasecmp(ptr_value, BAD_CAST "Minute")==0) {
	      str_now = xmlStrdup(today_min);
	    } 
	    else if (xmlStrcasecmp(ptr_value, BAD_CAST "Second")==0) {
	      str_now = xmlStrdup(today_sec);
	    } 
	    else {
	      str_now = NULL;
	    }
	  }
	  else {
	    str_now = NULL;
	  }

	  if ((ptr_value = attribute_str(node_subst, BAD_CAST "argv"))) {
	    /* this is a substitution with program arguments */
	    str_argv = getargvUTF8(atoi((char *)ptr_value));
	  }
	  else {
	    str_argv = NULL;
	  }

	  if ((ptr_value = attribute_str(node_subst, BAD_CAST "cgi"))) {
	    str_cgi = getcgiUTF8(ptr_value);
	  }
	  else {
	    str_cgi = NULL;
	  }

	  if ((ptr_value = attribute_str(node_subst, BAD_CAST "env"))!=NULL) {
	    /* this is a substitution with Environment variables */
	    str_env = getenvUTF8(ptr_value);
	  }
	  else {
	    str_env = NULL;
	  }

	  /*!\todo substitute searchpath="pie.css" to whole filename */

	  if (str_to) {
	    printf_log(2,"Substituting '%s' to '%s'",str_name,str_to);
	    substitute_subst(node_subst->next,str_name,str_to);
	  }
	  else if (str_argv) {
	    printf_log(2,"Substituting Argv '%s' to '%s'",str_name,str_argv);
	    substitute_subst(node_subst->next,str_name,str_argv);
	  }
	  else if (str_cgi) {
#ifdef _MSC_VER
	    xmlChar *ptr_a, *ptr_b;
	    for (ptr_a=ptr_b=str_cgi;ptr_a && *ptr_a != '\0'; ptr_a++) {
	      if (*ptr_a == '\r') {
		}
	      else {
		*ptr_b = *ptr_a;
		ptr_b++;
		}
	      }
	    *ptr_b = '\0';
#endif	    
	    printf_log(2,"Substituting CGI '%s' to '%s'",str_name,str_cgi);
	    substitute_subst(node_subst->next,str_name,str_cgi);
	  }
	  else if (str_env) {
	    printf_log(2,"Substituting Env '%s' to '%s'",str_name,str_env);
	    substitute_subst(node_subst->next,str_name,str_env);
	  }
	  else if (str_now) {
	    printf_log(2,"Substituting Now '%s' to '%s'",str_name,str_now);
	    substitute_subst(node_subst->next,str_name,str_now);
	  }
	  else if (str_dir) {
	    printf_log(2,"Substituting Dir '%s' to '%s'",str_name,str_dir);
	    substitute_subst(node_subst->next,str_name,str_dir);
	  }
	  else if (str_filename) {
	    printf_log(2,"Substituting Filename '%s' to '%s'",str_name,str_filename);
	    substitute_subst(node_subst->next,str_name,str_filename);
	  }
	  else if (str_file) {
	    printf_log(2,"Substituting File '%s' to content of '%s'",str_name,str_file);
	    substitute_subst(node_subst->next,str_name,fgetUTF8(str_file));
	  }
	  else if ((node_xml = child_element(node_subst,NAME_XML))
		   && (doc_result = process_xml(node_xml))
		   && (node_root = xmlDocGetRootElement(doc_result))) {
	    printf_log(2,"Substituting '%s' to DOM",str_name);
	    substitute_substdom(node_subst->next,str_name,node_root);
	    xmlFreeDoc(doc_result);
	  }
	  else if (str_default) {
	    printf_log(3, "Using default value '%s'", str_default);
	    substitute_subst(node_subst->next,str_name,str_default);
	  }
	  else {
	    /* ?? */
	  }

	  if (str_now) xmlFree(str_now);
	  if (str_name) xmlFree(str_name);
	}
      }
    }
  }
  return TRUE;
}


/*! process the 
 */
xmlDocPtr
process_info(xmlNodePtr node_info)
{
    xmlDocPtr result;
    xmlNodePtr node_root;
    xmlNodePtr node_pwd;
    int i;

    result = xmlNewDoc(BAD_CAST "1.0");
    node_root = xmlNewDocNode(result, NULL, NAME_INFO, NULL); 
    xmlDocSetRootElement(result, node_root);
    ns_cxp = xmlNewNs(node_root, ptr_cxp, BAD_CAST "cxp");
    xmlSetNs(node_root,ns_cxp);

    /*!\todo system time */
    
    /* working dir */
    node_pwd = xmlNewChild(node_root, NULL, BAD_CAST"pwd", NULL); 
    xmlSetProp(node_pwd,BAD_CAST "select",getcwdUTF8());

    /* arguments */
    for ( i = 0; i < argc_i; i++ ) {
        xmlChar str_index[BUFFER_LENGTH];
	xmlNodePtr node_argv;
	node_argv = xmlNewChild(node_root, NULL, BAD_CAST"arg", NULL); 
	xmlStrPrintf(str_index,BUFFER_LENGTH-1,BAD_CAST "%i",i);
	xmlSetProp(node_argv,BAD_CAST "name",str_index);
	xmlSetProp(node_argv,BAD_CAST "select",getargvUTF8(i));
    }

    if (runmode == cgi) {
      /* CGI environment */
      for (i = 0; i < cgi_num_entries; i++) {
	xmlNodePtr node_env;
	xmlChar *ptr_name;
	/*!\todo improve encoding */
	ptr_name = xmlStrdup(BAD_CAST cgi_entries[i].name);
	node_env = xmlNewChild(node_root, NULL, BAD_CAST"cgi", NULL); 
	xmlSetProp(node_env,BAD_CAST "name",ptr_name);
	xmlSetProp(node_env,BAD_CAST "select",BAD_CAST cgi_entries[i].val);
      }
    }
      
    /* environment */
    for ( i = 0; envp_i[i] != NULL; i++ ) {
	xmlNodePtr node_env;
	xmlChar *ptr_sep;
	if ((ptr_sep = (xmlChar *)strchr(envp_i[i],(int)'='))) {
	    xmlChar *ptr_name;
	    /*!\todo improve encoding */
	    ptr_name = xmlStrndup(BAD_CAST envp_i[i],(ptr_sep - (xmlChar *)envp_i[i]));
	    node_env = xmlNewChild(node_root, NULL, NAME_ENV, NULL); 
	    xmlSetProp(node_env,BAD_CAST "name",ptr_name);
	    xmlSetProp(node_env,BAD_CAST "select",getenvUTF8(ptr_name));
	}
    }

    return result;
}
