/* 
   cxproc - Configurable Xml PROCessor
   
   Copyright (C) 2006,2007,2008 by Alexander Tenbusch

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

*/

/*! mysql frontend
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <assert.h>
#include <errno.h>
#include <limits.h>

#include <libxml/tree.h>

/* 
 */
#include "basics.h"
#include "tools.h"
#include "dom_x.h"
#include "plain_x.h"
#include "make_x.h"
#include "calendar_x.h"
#include "cxp_dtd.h"
#ifdef _MSC_VER
#else
#include "../config.h"
#if HAVE_LIBMYSQL
#include <mysql/mysql.h>
#endif
#endif


/*! process the required 
 */
xmlDocPtr
process_sql(xmlNodePtr node_make_sql)
{
  xmlDocPtr result = NULL;

#if HAVE_LIBMYSQL
  xmlNodePtr node_meta;
  xmlNodePtr node_sql;
  xmlNodePtr node_result;
  xmlChar *str_query;
  xmlChar *str_db;

  MYSQL *connection;
  MYSQL mysql;

  int i;
  int j;

  MYSQL_ROW row;
  MYSQL_RES *result_sql;	/* Ergebnis der Anfrage */
  MYSQL_FIELD *fields, *field;

  int error;
  int n_rows;
  int n_fields;

  /*!\todo support for sqlite ??? */

  /*!\todo allow INSERT and UPDATE queries with values from an input DOM */

  printf_log(1,"Processing SQL");

  /*! create DOM
   */
  result = xmlNewDoc(BAD_CAST "1.0");
  node_sql = xmlNewDocNode(result, NULL, BAD_CAST "sql", NULL); 
  xmlDocSetRootElement(result, node_sql);

  node_meta = xmlNewChild(node_sql, NULL, BAD_CAST "meta", NULL);
  xmlAddChild(node_meta,xmlCopyNode(node_make_sql,1));

  /* 
     detect the necessary connection informations
   */
  str_db = attribute_str(node_make_sql,BAD_CAST "db");
  if (str_db == NULL) {
    printf_log(1,"No db attribute");
    return NULL;    
  }
  printf_log(1,"Database '%s'", str_db);

  str_query = node_make_sql->children->content;
  /*!\todo process_plain() here */
  if (str_query == NULL || xmlStrlen(str_query) < 5) {
    printf_log(1,"No usable query text node");
    return NULL;    
  }
  printf_log(1,"Query '%s'", str_query);

  /* 
     connect to the specified database
   */
  mysql_init(&mysql);

  /*!\todo additional attributes host,user,passwd */

  connection = mysql_real_connect( &mysql,
				   "",
				   NULL,
				   NULL,
				   str_db,
				   0,
				   0,
				   0);

  if (connection == NULL) {
    printf_log(1,"Error '%s' for connection", mysql_error(&mysql));
    return NULL;
  }

  error = mysql_query(connection, str_query);
  if (error) {
    printf_log(1,"Error result '%i' for query", error);
    return NULL;
  }
  result_sql = mysql_store_result(connection);

  n_rows = mysql_num_rows(result_sql);
  if (n_rows < 1) {
    return NULL;
  }

  /*!\todo write a DTD section into DOM to validate the result */

  node_result = xmlNewChild(node_sql, NULL, BAD_CAST "result", NULL);

  n_fields = mysql_num_fields(result_sql);
  fields = mysql_fetch_fields(result_sql);

  for (i=0; (row = mysql_fetch_row(result_sql)) != NULL; i++ ) {
    xmlNodePtr node_entry;

    assert(i<n_rows);

    node_entry = xmlNewChild(node_result, NULL, BAD_CAST "entry", NULL);
    if (node_entry==NULL) {
      return NULL;
    }

    for (j=0; j<n_fields; j++) {
      int outlen = BUFFER_LENGTH-1;
      int inlen;
      xmlChar *ptr_date;
      unsigned char buffer_out[BUFFER_LENGTH];
      char *value = row[j];
      if (value) {
	inlen = strlen((char *)value);
	if (inlen > 0) {

	  if (xmlCheckUTF8(value)) {
	    ptr_date = value;
	  }
	  else {
	    int l = isolat1ToUTF8(buffer_out,&outlen,value,&inlen);
	    if (l < 1) {
	      printf_log(1,"Problem with entry");
	      continue;
	    }
	    buffer_out[l] = '\0';
	    ptr_date = buffer_out;
	  }

	  field = &(fields[j]);
	  if ( field->type == MYSQL_TYPE_STRING 
	       || field->type == MYSQL_TYPE_VAR_STRING
	       || field->type == MYSQL_TYPE_DATE
	       || field->type == MYSQL_TYPE_BLOB) {
	    xmlNewChild(node_entry, NULL, BAD_CAST field->name, ptr_date);
	  }
	  else {
	    xmlSetProp(node_entry, BAD_CAST field->name, ptr_date);
	  }
	}
      }
    }
  }

  mysql_free_result(result_sql);

  mysql_close(&mysql);
#else
  printf_log(1,"Compiled without SQL, ignoring");  
#endif

  return result;
}

