/*
  cxproc - Configurable Xml PROCessor
   
  Copyright (C) 2006,2007,2008 by Alexander Tenbusch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License 
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <libxml/parser.h>

#include "basics.h"
#include "tools.h"

/*! internal level for debug messages */
static int level_set = -1;

/*! internal flag for readonly mode */
static BOOL_T mode_readonly = FALSE;

time_t system_zeit = (time_t)0;


/*! replace string at needle with string at replace in string at haystack

  \return pointer to the new concatenated string or NULL if not contained
*/
xmlChar *
str_replace(xmlChar *haystack, xmlChar *needle, xmlChar *replace)
{
  int l_needle;
  xmlChar *found;

#if 0
  assert(haystack != NULL);
  assert(needle != NULL);
#else
  if (haystack==NULL || needle==NULL)
    return NULL;
#endif

  if (replace == NULL) {
    replace = xmlStrdup(BAD_CAST "");
  }

  if (xmlStrlen(haystack) < 1) {
    return NULL;
  }
  l_needle = xmlStrlen(needle);

  assert(l_needle > 0);

  found = (xmlChar *) xmlStrstr(haystack,needle);
  if (found) {
    xmlChar *postfix, *prefix;
    xmlChar *ptr_release;

    postfix = str_replace(&found[l_needle],needle,replace);
    if (postfix == NULL) {
      postfix = xmlStrdup(&found[l_needle]);
    }
    if (found - haystack > 0) {
      prefix = xmlStrndup(haystack,(found - haystack));
      ptr_release = prefix;
      prefix = xmlStrncatNew(ptr_release,replace,xmlStrlen(replace));
      xmlFree(ptr_release);
      ptr_release = prefix;
      prefix = xmlStrncatNew(prefix,postfix,xmlStrlen(postfix));
      xmlFree(ptr_release);
    }
    else {
      /* found points to the start of haystack */
      prefix = xmlStrncatNew(replace,postfix,xmlStrlen(postfix));
    }
    xmlFree(postfix);
    return prefix;
  }
  return NULL;
}
/* end of str_replace() */


/*! set the internal readonly flag
*/
void
set_readonly(BOOL_T mode)
{
  printf_log(1,"Set mode to '%s'", (mode==TRUE ? "readonly" : "readwrite"));
  mode_readonly = mode;
}
/* end of set_readonly() */


/*! returns the internal readonly flag
*/
BOOL_T
isreadonly(void)
{
  return mode_readonly;
}
/* end of isreadonly() */


/*! set the internal debug level

0 - no log messages at all, error messages only
1 - minimum log level
*/
void
set_level_log(int level)
{
#define LEVEL_MAX 4

  if (level > 0) {
    if (level != level_set) {
      if (level > LEVEL_MAX) {
	level = LEVEL_MAX;
      }
      printf_log(0,"Change log level from '%i' to '%i'", level_set, level);
      level_set = level;
    }
  }
  else {
    /* printf_log(0,"Minimize log level"); */
    level_set = 0;
  }
}
/* end of set_level_log() */


/*! fprintf's log messages

\todo use xmlStrPrintf()
 */
void
_printf__log(int level, char *fmt, ...)
{
  int time_delta;
  va_list ap;

  if (level > level_set)
    return;

  va_start(ap,fmt);
#if 1
  if (system_zeit==0) {
    /* initialize first time */
    time(&system_zeit);
    time_delta = 0;
  }
  else {
    time_t system_zeit_1;
    time(&system_zeit_1);
    time_delta = (int) difftime(system_zeit_1,system_zeit);
  }
  fprintf(stderr,"(%is) ",time_delta);
/*   fprintf(stderr,"(%is/%i) ",time_delta,level); */
#elif 1
  fprintf(stderr,"(%i/%i) ",(int)clock()/CLOCKS_PER_SEC,level);
#else
  fprintf(stderr,"(%i) ",level);
#endif

  vfprintf(stderr, fmt, ap);
  fprintf(stderr,"\n");
  va_end(ap);
#if 0
#ifdef DEBUG
#elif !defined(WIN32) 
  fprintf(stderr,".");
#endif
#endif
}


/*! fprintf's error messages and exit(exit_code)

  \param flag is the result of an exit test criteria
  \param exit_code is the return code of the whole program
  \param fmt is the printf format string for exit message
  \param ... printf format arguments
 */
void
printf_exit(BOOL_T flag, int exit_code, char *fmt, ...)
{
  if (flag==TRUE) {
    va_list ap;

    va_start(ap,fmt);
    fprintf(stderr,"(E%i) ", exit_code);
    vfprintf(stderr, fmt, ap);
    fprintf(stderr,"\n");
    va_end(ap);

    exit(exit_code);
  }
}


/*! this fuction is registered for program exit atexit()
*/
void
exit_debug(void)
{
#if 0
   char buffer[10];
    fgets(buffer,sizeof(buffer),stdin);
#endif

#ifdef DEBUG
  printf_log(1,"This is my valid exit()");
#endif
}
/* end of exit_debug() */



